/* xscreensaver, Copyright (c) 1992, 1995, 1996, 1997, 1998, 2001
 *  Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* Simulation of a pair of quasi-gravitational fields, maybe sorta kinda
   a little like the strong and weak electromagnetic forces.  Derived from
   a Lispm screensaver by John Pezaris <pz@mit.edu>.  Mouse control and
   viscosity added by "Philip Edward Cutone, III" <pc2d+@andrew.cmu.edu>.

   John sez:

       The simulation started out as a purely accurate gravitational
       simulation, but, with constant simulation step size, I quickly
       realized the field being simulated while grossly gravitational
       was, in fact, non-conservative.  It also had the rather annoying
       behavior of dealing very badly with colliding orbs.  Therefore,
       I implemented a negative-gravity region (with two thresholds; as
       I read your code, you only implemented one) to prevent orbs from
       every coming too close together, and added a viscosity factor if
       the speed of any orb got too fast.  This provides a nice stable
       system with interesting behavior.

       I had experimented with a number of fields including the van der
       Waals force (very interesting orbiting behavior) and 1/r^3
       gravity (not as interesting as 1/r^2).  An even normal viscosity
       (rather than the thresholded version to bleed excess energy) is
       also not interesting.  The 1/r^2, -1/r^2, -10/r^2 thresholds
       proved not only robust but also interesting -- the orbs never
       collided and the threshold viscosity fixed the
       non-conservational problem.

   Philip sez:
       > An even normal viscosity (rather than the thresholded version to
       > bleed excess energy) is also not interesting.

       unless you make about 200 points.... set the viscosity to about .8
       and drag the mouse through it.   it makes a nice wave which travels
       through the field.

   And (always the troublemaker) Joe Keane <jgk@jgk.org> sez:

       Despite what John sez, the field being simulated is always
       conservative.  The real problem is that it uses a simple hack,
       computing acceleration *based only on the starting position*,
       instead of a real differential equation solver.  Thus you'll
       always have energy coming out of nowhere, although it's most
       blatant when balls get close together.  If it were done right,
       you wouldn't need viscosity or artificial limits on how close
       the balls can get.

   Matt <straitm@carleton.edu> sez:

       Added a switch to remove the walls.

       Added a switch to make the threshold viscosity optional.  If
       nomaxspeed is specified, then balls going really fast do not
       recieve special treatment.

       I've made tail mode prettier by eliminating the first erase line
       that drew from the upper left corner to the starting position of
       each point.

       Made the balls in modes other than "balls" bounce exactly at the
       walls.  (Because the graphics for different modes are drawn
       differently with respect to the "actual" position of the point,
       they used to be able to run somewhat past the walls, or bounce
       before hitting them.)

       Added an option to output each ball's speed in the form of a bar
       graph drawn on the same window as the balls.  If only x or y is
       selected, they will be represented on the appropriate axis down
       the center of the window.  If both are selected, they will both
       be displayed along the diagonal such that the x and y bars for
       each point start at the same place.  If speed is selected, the
       speed will be displayed down the left side.  */


/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 09/03
 */

/* plan9-related stuff */
#include "plan9.h"

struct ball {
	double x, y;
	double vx, vy;
	double dx, dy;
	double mass;
	int size;
	int pixel_index;
	int hue;
};

static struct ball *balls;
static double *x_vels;
static double *y_vels;
static double *speeds;
static int npoints;
static int threshold;
static int delay;
static int global_size;
static int segments;
static Bool glow_p;
static Bool orbit_p;
static Bool walls_p;
static Bool maxspeed_p;
static Bool cbounce_p;
static int color_shift;
static XPoint *point_stack;
static int point_stack_size, point_stack_fp;
static int ncolors;
Bool no_erase_yet; /* for tail mode fix */

/*flip mods for mouse interaction*/
static Bool mouse_p;
int mouse_x, mouse_y, mouse_mass, root_x, root_y;
static double viscosity;

static enum object_mode {
	ball_mode, line_mode, polygon_mode, 
	tail_mode
} mode;

static enum graph_mode {
	graph_none, graph_x, graph_y, graph_both, graph_speed
} graph_mode;

/* The normal (and max) width for a graph bar */
#define BAR_SIZE 11
#define MAX_SIZE 16


static void
init_balls (void)
{
	int i;
	int xlim, ylim, midx, midy, r, vx, vy;
	double th;

	xlim = Dx(screen->r);
	ylim = Dy(screen->r);
	midx = xlim/2;
	midy = ylim/2;
	walls_p = 1;

	r = MIN(xlim/2, ylim/2) - 50;

	vx = 0;
	vy = 0;

	npoints = 3 + (random () % 5);
	balls = (struct ball *) malloc (npoints * sizeof (struct ball));

	no_erase_yet = 1; /* for tail mode fix */

	segments = 500;

	threshold = 100;

	delay = 1;

	global_size = 0;

	glow_p = 0;

	orbit_p = 0;

	maxspeed_p = 1;

	cbounce_p = 1;

	color_shift = 5;
	color = colors[random() % 256];

	viscosity = 1;

	/*switch(random() % 2 ) {
	case 0:	mode = ball_mode; break;
	case 1:	mode = tail_mode; break;
	}
	*/

	/* this one looks the best */
	mode = ball_mode;

	if (mode != ball_mode && mode != tail_mode) glow_p = False;

	if (mode == polygon_mode && npoints < 3)
		mode = line_mode;

	ncolors = 256;


	if (mode != ball_mode)
	{
		int size = (segments ? segments : 1);
		point_stack_size = size * (npoints + 1);
		point_stack = (XPoint *) calloc (point_stack_size, sizeof (XPoint));
		point_stack_fp = 0;
	}

#define rand_size() MIN (MAX_SIZE, 8 + (random () % (MAX_SIZE - 9)))

	if (orbit_p && !global_size)
		/* To orbit, all objects must be the same mass, or the math gets really hairy... */
		global_size = rand_size ();

	th = frand ()*2*M_PI;
	for (i = 0; i < npoints; i++)
	{
		int new_size = (global_size ? global_size : rand_size ());
		balls [i].dx = 0;
		balls [i].dy = 0;
		balls [i].size = new_size;
		balls [i].mass = (new_size * new_size * 10);
		balls [i].x = midx + r * cos (i * ((M_PI+M_PI) / npoints) + th);
		balls [i].y = midy + r * sin (i * ((M_PI+M_PI) / npoints) + th);
		if (! orbit_p)
		{
			balls [i].vx = vx ? vx : ((6.0 - (random () % 11)) / 8.0);
			balls [i].vy = vy ? vy : ((6.0 - (random () % 11)) / 8.0);
		}

		balls [i].pixel_index = random() % ncolors;
	}


	if (orbit_p)
	{
		double a = 0;
		double v;
		double v_mult = 0.9;
		if (v_mult == 0.0) v_mult = 1.0;

		for (i = 1; i < npoints; i++)
		{
			double _2ipi_n = (2 * i * M_PI / npoints);
			double x = r * cos (_2ipi_n);
			double y = r * sin (_2ipi_n);
			double distx = r - x;
			double dist2 = (distx * distx) + (y * y);
			double dist = sqrt (dist2);
			double a1 = ((balls[i].mass / dist2) *
				((dist < threshold) ? -1.0 : 1.0) *
				(distx / dist));
			a += a1;
		}
		if (a < 0.0)
		{
			fprint (2, "domain error: forces on balls too great\n");
			threadexitsall("domain error");
		}
		v = sqrt (a * r) * v_mult;
		for (i = 0; i < npoints; i++)
		{
			double k = ((2 * i * M_PI / npoints) + th);
			balls [i].vx = -v * sin (k);
			balls [i].vy =  v * cos (k);
		}
	}


	draw(img, img->r, display->black, nil, ZP);
}

static void
compute_force (int i, double *dx_ret, double *dy_ret)
{
	int j;
	double x_dist, y_dist, dist, dist2;
	*dx_ret = 0;
	*dy_ret = 0;
	for (j = 0; j < npoints; j++)
	{
		if (i == j) continue;
		x_dist = balls [j].x - balls [i].x;
		y_dist = balls [j].y - balls [i].y;
		dist2 = (x_dist * x_dist) + (y_dist * y_dist);
		dist = sqrt (dist2);

		if (dist > 0.1) /* the balls are not overlapping */
		{
			double new_acc = ((balls[j].mass / dist2) *
				((dist < threshold) ? -1.0 : 1.0));
			double new_acc_dist = new_acc / dist;
			*dx_ret += new_acc_dist * x_dist;
			*dy_ret += new_acc_dist * y_dist;
		}
		else
		{		/* the balls are overlapping; move randomly */
			*dx_ret += (frand ()*10.0 - 5.0);
			*dy_ret += (frand ()*10.0 - 5.0);
		}
	}
}

int tick = 500;
int fg_index = 0;

static void
run_balls (int total_ticks)
{
	int last_point_stack_fp = point_stack_fp;
	static int xlim, ylim;

	int i, radius = global_size/2;
	if(global_size == 0)
		radius = (MAX_SIZE / 3);

	if (tick++ == 500)
	{
		tick = 0;
		xlim = Dx(screen->r);
		ylim = Dy(screen->r);
	}

	/* compute the force of attraction/repulsion among all balls */
	for (i = 0; i < npoints; i++)
		compute_force (i, &balls[i].dx, &balls[i].dy);

	/* move the balls according to the forces now in effect */
	for (i = 0; i < npoints; i++)
	{
		double old_x = balls[i].x;
		double old_y = balls[i].y;
		double new_x, new_y;
		int size = balls[i].size;
		balls[i].vx += balls[i].dx;
		balls[i].vy += balls[i].dy;

		if (balls[i].vx > 10 && maxspeed_p)
		{
			balls[i].vx *= 0.9;
			balls[i].dx = 0;
		}
		else if (viscosity != 1)
		{
			balls[i].vx *= viscosity;
		}

		if (balls[i].vy > 10 && maxspeed_p)
		{
			balls[i].vy *= 0.9;
			balls[i].dy = 0;
		}
		else if (viscosity != 1)
		{
			balls[i].vy *= viscosity;
		}

		balls[i].x += balls[i].vx;
		balls[i].y += balls[i].vy;


		/* bounce off the walls if desired
	 note: a ball is actually its upper left corner */
		if(walls_p)
		{
			if(cbounce_p)  /* with correct bouncing */
			{
				/* so long as it's out of range, keep bouncing */

				while( (balls[i].x >= (xlim - balls[i].size)) ||
					(balls[i].y >= (ylim - balls[i].size)) ||
					(balls[i].x <= 0) ||
					(balls[i].y <= 0) )
				{
					if (balls[i].x >= (xlim - balls[i].size))
					{
						balls[i].x = (2*(xlim - balls[i].size) - balls[i].x);
						balls[i].vx = -balls[i].vx;
					}
					if (balls[i].y >= (ylim - balls[i].size))
					{
						balls[i].y = (2*(ylim - balls[i].size) - balls[i].y);
						balls[i].vy = -balls[i].vy;
					}
					if (balls[i].x <= 0)
					{
						balls[i].x = -balls[i].x;
						balls[i].vx = -balls[i].vx;
					}
					if (balls[i].y <= 0)
					{
						balls[i].y = -balls[i].y;
						balls[i].vy = -balls[i].vy;
					}
				}
			}
		}
		new_x = balls[i].x;
		new_y = balls[i].y;


		if (mode == ball_mode)
		{
			//XFillArc (dpy, window, erase_gc, (int) old_x, (int) old_y,
			//    size, size, 0, 360*64);
			fillellipse(img, Pt(old_x, old_y), 
				size, size, display->black, ZP);

			//XFillArc (dpy, window, draw_gc,  (int) new_x, (int) new_y,
			//    size, size, 0, 360*64);
			fillellipse(img, Pt(new_x, new_y), 
				size, size, colors[balls[i].pixel_index], ZP);


		}
		else
		{
			point_stack [point_stack_fp].x = new_x;
			point_stack [point_stack_fp].y = new_y;
			point_stack_fp++;
		}
	}

	/* draw the lines or polygons after computing all points */
	if (mode != ball_mode)
	{
		point_stack [point_stack_fp].x = balls [0].x; /* close the polygon */
		point_stack [point_stack_fp].y = balls [0].y;
		point_stack_fp++;
		if (point_stack_fp == point_stack_size)
			point_stack_fp = 0;
		else if (point_stack_fp > point_stack_size) /* better be aligned */
			abort ();
		{
			static int tick = 0;
			if (tick++ == color_shift)
			{
				fg_index = (fg_index + 1) % ncolors;
				tick = 0;
				//XSetForeground (dpy, draw_gc, colors[fg_index]);
				color = colors[fg_index];
			}
		}
	}

	switch (mode)
	{
	case ball_mode:
		break;
	case tail_mode:
		{
			for (i = 0; i < npoints; i++)
			{
				int index = point_stack_fp + i;
				int next_index = (index + (npoints + 1)) % point_stack_size;
				if(no_erase_yet == 1)
				{
					if(total_ticks >= segments)
					{
						no_erase_yet = 0;
						//XDrawLine (dpy, window, erase_gc,
						//    point_stack [index].x + radius,
						//    point_stack [index].y + radius,
						//    point_stack [next_index].x + radius,
						//    point_stack [next_index].y + radius);
						line(screen, addpt(screen->r.min, Pt(point_stack [index].x + radius, point_stack [index].y + radius)), 
							addpt(screen->r.min, Pt(point_stack [next_index].x + radius, point_stack [next_index].y + radius)), 
							Endsquare, Endsquare, 0, display->black, ZP);

					}
				}
				else
				{
					//XDrawLine (dpy, window, erase_gc,
					//    point_stack [index].x + radius,
					//    point_stack [index].y + radius,
					//    point_stack [next_index].x + radius,
					//    point_stack [next_index].y + radius);
					line(screen, addpt(screen->r.min, Pt(point_stack [index].x + radius, point_stack [index].y + radius)), 
						addpt(screen->r.min, Pt(point_stack [next_index].x + radius, point_stack [next_index].y + radius)), 
						Endsquare, Endsquare, 0, display->black, ZP);
				}
				index = last_point_stack_fp + i;
				next_index = (index - (npoints + 1)) % point_stack_size;
				if (next_index < 0) next_index += point_stack_size;
				if (point_stack [next_index].x == 0 &&
					point_stack [next_index].y == 0)
					continue;
				//XDrawLine (dpy, window, draw_gc,
				//    point_stack [index].x + radius,
				//    point_stack [index].y + radius,
				//    point_stack [next_index].x + radius,
				//    point_stack [next_index].y + radius);
				line(img, addpt(screen->r.min, Pt(point_stack [index].x + radius, point_stack [index].y + radius)), 
					addpt(screen->r.min, Pt(point_stack [next_index].x + radius, point_stack [next_index].y + radius)), 
					Endsquare, Endsquare, 0, color, ZP);
			}
		}
		break;
	default:
		abort ();
	}

	draw(screen, screen->r, img, nil, ZP);
	frames++;
	flushimage(display, 1);
}


void
screenhack (void)
{
	/* for tail mode fix */
	int total_ticks = 0;

	init_balls ();
	while (1)
	{
		total_ticks++;
		run_balls (total_ticks);
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		if (delay)
			sleep (delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	//fprint(2, "resize not supported...\n");
	//threadexitsall("resize not supported");
	tick = 500;

	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

}

void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "attraction") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			cmap2rgba(i<<2));
	}

	einit(Emouse);
	eresized(0);
	screenhack();
}
