/* xscreensaver, Copyright (c) 2005, 2006 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 * Boxfit -- fills space with a gradient of growing boxes or circles.
 *
 * Written by jwz, 21-Feb-2005.
 *
 * Inspired by http://www.levitated.net/daily/levBoxFitting.html
 */

/*
 * ported to Plan 9 by mirtchovski@gmail.com, 05/07
 */

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"

#define ALIVE   1
#define CHANGED 2
#define UNDEAD  4

typedef struct {
	unsigned long fill_color;
	short x, y, w, h;
	char flags;
} box;

typedef struct {
	Image *fg_color, *bg_color;
	int border_size;
	int spacing;
	int inc;

	Bool circles_p;
	Bool growing_p;
	Bool color_horiz_p;

	int box_count;
	int boxes_size;
	int nboxes;
	box *boxes;

	XImage *image;
	int ncolors;
	XColor *colors;
	int delay;
	int countdown;

	Bool done_once;

} state;

state st;

static void
reset_boxes (state *st)
{
	st->nboxes = 0;
	st->growing_p = True;
	st->color_horiz_p = nrand(2);

	st->circles_p = nrand(30) <10; /* we favor squares since they render faster */

	smooth(colors);
	draw(screen, screen->r, display->black, nil, ZP);
}


static void
reshape_boxes (state *st)
{
	int i;
	for (i = 0; i < st->nboxes; i++)
	{
		box *b = &st->boxes[i];
		b->flags |= CHANGED;
	}
}

void
boxfit_init (state *st)
{

	st->delay = 20000;

	st->ncolors = 256;

	st->inc = nrand(5)+1;
	st->spacing = nrand(2)+1;
	st->border_size = 1;
	st->fg_color = display->white;
	st->bg_color = display->black;

	st->box_count = nrand(50)+25;

	st->nboxes = 0;
	st->boxes_size = st->box_count * 2;
	st->boxes = (box *) calloc (st->boxes_size, sizeof(*st->boxes));

	reset_boxes (st);

	reshape_boxes (st);
}



static Bool
boxes_overlap_p (box *a, box *b, int pad)
{
	/* Two rectangles overlap if the max of the tops is less than the
     min of the bottoms and the max of the lefts is less than the min
     of the rights.
   */
# undef MAX
# undef MIN
# define MAX(A,B) ((A)>(B)?(A):(B))
# define MIN(A,B) ((A)<(B)?(A):(B))

	int maxleft  = MAX(a->x - pad, b->x);
	int maxtop   = MAX(a->y - pad, b->y);
	int minright = MIN(a->x + a->w + pad + pad - 1, b->x + b->w);
	int minbot   = MIN(a->y + a->h + pad + pad - 1, b->y + b->h);
	return (maxtop < minbot && maxleft < minright);
}


static Bool
circles_overlap_p (box *a, box *b, int pad)
{
	int ar = a->w/2;	/* radius */
	int br = b->w/2;
	int ax = a->x + ar;	/* center */
	int ay = a->y + ar;
	int bx = b->x + br;
	int by = b->y + br;
	int d2 = (((bx - ax) * (bx - ax)) +	/* distance between centers squared */
	((by - ay) * (by - ay)));
	int r2 = ((ar + br + pad) *		/* sum of radii squared */
	(ar + br + pad));
	return (d2 < r2);
}


static Bool
box_collides_p (state *st, box *a, int pad)
{
	int i;

	/* collide with wall */
	if (a->x - pad < 0 ||
		a->y - pad < 0 ||
		a->x + a->w + pad + pad >= Dx(screen->r) ||
		a->y + a->h + pad + pad >= Dy(screen->r))
		return True;

	/* collide with another box */
	for (i = 0; i < st->nboxes; i++)
	{
		box *b = &st->boxes[i];
		if (a != b &&
			(st->circles_p
			? circles_overlap_p (a, b, pad)
			: boxes_overlap_p (a, b, pad)))
			return True;
	}

	return False;
}


static unsigned int
grow_boxes (state *st)
{
	int inc2 = st->inc + st->spacing + st->border_size;
	int i;
	int live_count = 0;

	/* check box collisions, and grow if none.
   */
	for (i = 0; i < st->nboxes; i++)
	{
		box *a = &st->boxes[i];
		if (!(a->flags & ALIVE)) continue;

		if (box_collides_p (st, a, inc2))
		{
			a->flags &= ~ALIVE;
			continue;
		}

		live_count++;
		a->x -= st->inc;
		a->y -= st->inc;
		a->w += st->inc + st->inc;
		a->h += st->inc + st->inc;
		a->flags |= CHANGED;
	}

	/* Add more boxes.
   */
	while (live_count < st->box_count)
	{
		box *a;
		st->nboxes++;
		if (st->boxes_size <= st->nboxes)
		{
			st->boxes_size = (st->boxes_size * 1.2) + st->nboxes;
			st->boxes = (box *)
				realloc (st->boxes, st->boxes_size * sizeof(*st->boxes));
			if (! st->boxes)
			{
				fprint (2, "%s: out of memory (%d boxes)\n", argv0, st->boxes_size);
				exits ("memory");
			}
		}

		a = &st->boxes[st->nboxes-1];
		a->flags = CHANGED;

		for (i = 0; i < 100; i++)
		{
			a->x = inc2 + (lrand() % (Dx(screen->r)  - inc2));
			a->y = inc2 + (lrand() % (Dy(screen->r) - inc2));
			a->w = 0;
			a->h = 0;

			if (! box_collides_p (st, a, inc2))
			{
				a->flags |= ALIVE;
				live_count++;
				break;
			}
		}

		if (! (a->flags & ALIVE) ||	/* too many retries; */
		st->nboxes > 65535)		/* that's about 1MB of box structs. */
		{
			st->nboxes--;			/* go into "fade out" mode now. */
			st->growing_p = False;
			return 2000000; /* make customizable... */
		}

		{
			int n = (st->color_horiz_p
				? (a->x * st->ncolors / Dx(screen->r))
				: (a->y * st->ncolors / Dy(screen->r)));
			a->fill_color   = n % st->ncolors;
		}
	}

	return st->delay;
}


static unsigned int
shrink_boxes (state *st)
{
	int i;
	int remaining = 0;

	for (i = 0; i < st->nboxes; i++)
	{
		box *a = &st->boxes[i];

		if (a->w <= 0 || a->h <= 0) continue;

		a->x += st->inc;
		a->y += st->inc;
		a->w -= st->inc + st->inc;
		a->h -= st->inc + st->inc;
		a->flags |= CHANGED;
		if (a->w < 0) a->w = 0;
		if (a->h < 0) a->h = 0;

		if (a->w > 0 && a->h > 0)
			remaining++;
	}

	if (remaining == 0) {
		reset_boxes (st);
		return 1000000;
	} else {
		return st->delay;
	}
}


static void
draw_boxes (state *st)
{
	Point p;
	Rectangle r;
	Image *bd;
	int i;

	for (i = 0; i < st->nboxes; i++)
	{
		box *b = &st->boxes[i];

		if (b->flags & UNDEAD) continue;
		if (! (b->flags & CHANGED)) continue;
		b->flags &= ~CHANGED;

		if (!st->growing_p)
		{
			int margin = st->inc + st->border_size;

			if(st->circles_p) {
				p = Pt(b->x+b->w/2, b->y+b->h/2);
				fillellipse(screen, 
					addpt(p, screen->r.min), b->w/2+margin, b->h/2+margin,
					st->bg_color, ZP);
			} else {
				r = Rect(b->x-margin, 
					b->y-margin, 
					b->x+margin+b->w, 
					b->y+margin+b->h);
				draw(screen, rectaddpt(r, screen->r.min),
					st->bg_color, nil, ZP);
			}

			if (b->w <= 0 || b->h <= 0)
				b->flags |= UNDEAD;   /* really very dead now */
		}

		if (b->w <= 0 || b->h <= 0) continue;

		bd = colors [(b->fill_color + st->ncolors/2)% st->ncolors];

		p = Pt(b->x+b->w/2, b->y+b->h/2);
		r = Rect(b->x, b->y, b->x+b->w, b->y+b->h);

		if(st->circles_p) {
			fillellipse(screen, addpt(p, screen->r.min), b->w/2, b->h/2, colors[b->fill_color], ZP);
			ellipse(screen, addpt(p, screen->r.min), b->w/2, b->h/2, st->border_size, bd, ZP);
		} else {
			draw(screen, rectaddpt(r, screen->r.min), colors[b->fill_color], nil, ZP);
			border(screen, rectaddpt(r, screen->r.min), st->border_size, bd, ZP);
		}
	}
}


static unsigned long
boxfit_draw (void *closure)
{
	state *st = (state *) closure;
	int delay;

	if (st->countdown > 0)
	{
		st->countdown -= st->delay;
		if (st->countdown <= 0)
		{
			st->countdown = 0;
			draw(screen, screen->r, st->bg_color, nil, ZP);
		}
		return st->delay;
	}

	if (st->growing_p) {
		draw_boxes (st);
		delay = grow_boxes (st);
	} else {
		delay = shrink_boxes (st);
		draw_boxes (st);
	}
	return delay;
}

static void
boxfit_reshape (void *closure)
{
	state *st = (state *) closure;
	reshape_boxes(st);
}

static Bool
boxfit_event (void *closure)
{
	state *st = (state *) closure;
	st->growing_p = !st->growing_p;
	return True;
}


void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

	draw(screen, screen->r, display->black, nil, ZP);
	boxfit_reshape(&st);
}


void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "boxfit") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);
	eresized(0);

	boxfit_init(&st);

	while(1) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4) {
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(0);
			} else if(m.buttons & (1|2)) {
				boxfit_event(&st);
			}
		}
		frames++;
		boxfit_draw(&st);
		flushimage(display, 1);
		sleep(30);
	}
}









