/* cloudlife by Don Marti <dmarti@zgp.org>
 *
 * Based on Conway's Life, but with one rule change to make it a better
 * screensaver: cells have a max age.
 *
 * When a cell exceeds the max age, it counts as 3 for populating the next
 * generation.  This makes long-lived formations explode instead of just 
 * sitting there burning a hole in your screen.
 *
 * Cloudlife only draws one pixel of each cell per tick, whether the cell is
 * alive or dead.  So gliders look like little comets.

 * 20 May 2003 -- now includes color cycling and a man page.

 * Based on several examples from the hacks directory of: 
 
 * xscreensaver, Copyright (c) 1997, 1998, 2002 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 1/04
 */


#include "plan9.h"

#include "smooth.h"

#define MAX_WIDTH 32767

struct field {
	unsigned int height;
	unsigned int width;
	unsigned int max_age;
	unsigned int cell_size;
	unsigned char *cells;
	unsigned char *new_cells;
};

static void 
*xrealloc(void *p, uint size)
{
	void *ret;
	if ((ret = realloc(p, size)) == NULL) {
		threadexitsall("out of memory");
	}
	return ret;
}

struct field 
*init_field(void)
{
	struct field *f = xrealloc(NULL, sizeof(struct field));
	f->height = 0;
	f->width = 0;
	f->cell_size = 3;
	f->max_age = 64;
	f->cells = NULL;
	f->new_cells = NULL;
	return f;
}

void 
resize_field(struct field * f, unsigned int w, unsigned int h)
{
	f->width = w;
	f->height = h;

	f->cells = xrealloc(f->cells,
		w * sizeof(unsigned char) * 
		h * sizeof(unsigned char));
	f->new_cells =
		xrealloc(f->new_cells,
		w * sizeof(unsigned char) * h * sizeof(unsigned char));
}

unsigned char 
*cell_at(struct field * f, unsigned int x, unsigned int y)
{
	return (f->cells + x * sizeof(unsigned char) + 
		y * f->width * sizeof(unsigned char));
}

unsigned char 
*new_cell_at(struct field * f, unsigned int x, unsigned int y)
{
	return (f->new_cells + x * sizeof(unsigned char) + 
		y * f->width * sizeof(unsigned char));
}

static void
draw_field(int fgc, struct field * f)
{
	unsigned int x, y;
	unsigned int rx, ry;	/* random amount to offset the dot */
	unsigned int size = 1 << f->cell_size;
	unsigned int mask = size - 1;
	static XPoint fg_point;
	static XPoint bg_point;

	/* columns 0 and width-1 are off screen and not drawn. */
	for (y = 1; y < f->height - 1; y++) {

		/* rows 0 and height-1 are off screen and not drawn. */
		for (x = 1; x < f->width - 1; x++) {
			rx = random();
			ry = rx >> f->cell_size;
			rx &= mask;
			ry &= mask;

			if (*cell_at(f, x, y)) {
				Rectangle r;
				fg_point.x = (short) x *size - rx - 1;
				fg_point.y = (short) y *size - ry - 1;
				r = Rpt(fg_point, addpt(fg_point, Pt(1,1)));
				draw(img, r, colors[fgc], nil, ZP);
			}
		}

		//		XDrawPoints(dpy, window, fgc, fg_points, fg_count,
		//		    CoordModeOrigin);
		//		XDrawPoints(dpy, window, bgc, bg_points, bg_count,
		//		    CoordModeOrigin);
	}
}

unsigned int 
cell_value(unsigned char c, unsigned int age)
{
	if (!c) {
		return 0;
	} else if (c > age) {
		return (3);
	} else {
		return (1);
	}
}

unsigned int 
is_alive(struct field * f, unsigned int x, unsigned int y)
{
	unsigned int count;
	unsigned int i, j;
	unsigned char *p;

	count = 0;

	for (i = x - 1; i <= x + 1; i++) {
		for (j = y - 1; j <= y + 1; j++) {
			if (y != j || x != i) {
				count += cell_value(*cell_at(f, i, j), f->max_age);
			}
		}
	}

	p = cell_at(f, x, y);
	if (*p) {
		if (count == 2 || count == 3) {
			return ((*p) + 1);
		} else {
			return (0);
		}
	} else {
		if (count == 3) {
			return (1);
		} else {
			return (0);
		}
	}
}

unsigned int 
do_tick(struct field * f)
{
	unsigned int x, y;
	unsigned int count = 0;
	for (x = 1; x < f->width - 1; x++) {
		for (y = 1; y < f->height - 1; y++) {
			count += *new_cell_at(f, x, y) = is_alive(f, x, y);
		}
	}
	memcpy(f->cells, f->new_cells, f->width * sizeof(unsigned char) *
		f->height * sizeof(unsigned char));
	return count;
}


unsigned int 
random_cell(unsigned int p)
{
	int r = random() & 0xff;

	if (r < p) {
		return (1);
	} else {
		return (0);
	}
}

void 
populate_field(struct field * f, unsigned int p)
{
	unsigned int x, y;

	for (x = 0; x < f->width; x++) {
		for (y = 0; y < f->height; y++) {
			*cell_at(f, x, y) = random_cell(p);
		}
	}
}

void 
populate_edges(struct field * f, unsigned int p)
{
	unsigned int i;

	for (i = f->width; i--;) {
		*cell_at(f, i, 0) = random_cell(p);
		*cell_at(f, i, f->height - 1) = random_cell(p);
	}

	for (i = f->height; i--;) {
		*cell_at(f, f->width - 1, i) = random_cell(p);
		*cell_at(f, 0, i) = random_cell(p);
	}
}

static int resize = 0;

void screenhack(void)
{
	struct field *f = init_field();

	unsigned int cycles = 0;
	unsigned int colorindex = 0;  /* which color in the colormap are we on */
	unsigned int colortimer = 0;  /* when this reaches 0, cycle to next color */

	int cycle_delay;
	int cycle_colors;
	int ncolors;
	int density;

	int fgc = 0;

	cycle_delay = 25;
	cycle_colors = 2;
	ncolors = 256;
	density = (30 % 100 * 256)/100;


	while (1) {
		if(resize) {
			f = init_field();
			resize = 0;
		}

		if (cycle_colors) {
			if (colortimer == 0) {
				colortimer = cycle_colors;
				if( colorindex == 0 )
					colorindex = ncolors;
				colorindex--;
				fgc = colorindex;
			}
			colortimer--;
		}

		if (f->height != Dy(screen->r) / (1 << f->cell_size) + 2 ||
			f->width != Dx(screen->r) / (1 << f->cell_size) + 2) {

			resize_field(f, Dx(screen->r) / (1 << f->cell_size) + 2,
				Dy(screen->r) / (1 << f->cell_size) + 2);
			populate_field(f, density);
		}

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(0);
		}

		draw_field(fgc, f);

		if (do_tick(f) < (f->height + f->width) / 4) {
			populate_field(f, density);
		}

		if (cycles % (f->max_age /2) == 0) {
			populate_edges(f, density);
			do_tick(f);
			populate_edges(f, 0);
		}

		draw(screen, screen->r, img, nil, ZP);
		flushimage(display, 1);
		frames++;

		if(!(cycles++ % 1000))
			draw(img, img->r, display->black, nil, ZP);

		if (cycle_delay)
			sleep(cycle_delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	resize = 1;

	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

	smooth(colors);

}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0)*getpid());


	if(initdraw(nil, nil, "cloudlife") < 0)
		sysfatal("initdraw failed: %r");


	einit(Emouse);
	eresized(0);

	screenhack();
}
