/* critical -- Self-organizing-criticality display hack for XScreenSaver
 * Copyright (C) 1998, 1999, 2000 Martin Pool <mbp@humbug.org.au>
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation.  No representations are made
 * about the suitability of this software for any purpose.  It is
 * provided "as is" without express or implied warranty.
 *
 * See `critical.man' for more information.
 *
 * Revision history:
 * 13 Nov 1998: Initial version, Martin Pool <mbp@humbug.org.au>
 * 08 Feb 2000: Change to keeping and erasing a trail, <mbp>
 *
 * It would be nice to draw curvy shapes rather than just straight
 * lines, but X11 doesn't have spline primitives (?) so we'd have to
 * do all the work ourselves  */

/*
  * Ported to Plan 9 by andrey@lanl.gov, 07/02
  */


/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"

typedef struct {
	int width, height;		/* in cells */
	unsigned short *cells;
} CriticalModel;

typedef struct {
	int trail;			/* length of trail */
	int cell_size;
} CriticalSettings;

CriticalSettings	settings;
CriticalModel * model_allocate (int w, int h);
void model_initialize (CriticalModel *model);

XPoint		*history; /* in cell coords */
CriticalModel		*model;
int			model_w, model_h;

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	fprint(2, "resize not supported...\n");
	threadexitsall("resize not supported");

}



int
clip (int low, int val, int high)
{
	if (val < low)
		return low;
	else if (val > high)
		return high;
	else
		return val;
}


/* Allocate an return a new simulation model datastructure.
 */

CriticalModel *
model_allocate (int model_w, int model_h)
{
	CriticalModel		*model;

	model = malloc (sizeof (CriticalModel));
	if (!model)
		return 0;

	model->width = model_w;
	model->height = model_h;

	model->cells = malloc (sizeof (int) * model_w * model_h);
	if (!model->cells)
		return 0;

	return model;
}



/* Initialize the data model underlying the hack.

   For the self-organizing criticality hack, this consists of a 2d
   array full of random integers.

   I've considered storing the data as (say) a binary tree within a 2d
   array, to make finding the highest value faster at the expense of
   storage space: searching the whole array on each iteration seems a
   little inefficient.  However, the screensaver doesn't seem to take
   up many cycles as it is: presumably the search is pretty quick
   compared to the sleeps.  The current version uses less than 1% of
   the CPU time of an AMD K6-233.  Many machines running X11 at this
   point in time seem to be memory-limited, not CPU-limited.

   The root of all evil, and all that.
*/


void
model_initialize (CriticalModel *model)
{
	int i;

	for (i = model->width * model->height; i >= 0; i--)
	{
		model->cells[i] = (unsigned short) random ();
	}
}


/* Move one step forward in the criticality simulation.

   This function locates and returns in (TOP_X, TOP_Y) the location of
   the highest-valued cell in the model.  It also replaces that cell
   and it's eight nearest neighbours with new random values.
   Neighbours that fall off the edge of the model are simply
   ignored. */
static void
model_step (CriticalModel *model, XPoint *ptop)
{
	int			x, y, i;
	int			dx, dy;
	unsigned short	top_value;
	int			top_x = 0, top_y = 0;

	/* Find the top cell */
	top_value = 0;
	i = 0;
	for (y = 0; y < model->height; y++)
		for (x = 0; x < model->width; x++)
		{
			if (model->cells[i] >= top_value)
			{
				top_value = model->cells[i];
				top_x = x;
				top_y = y;
			}
			i++;
		}

	/* Replace it and its neighbours with new random values */
	for (dy = -1; dy <= 1; dy++)
	{
		int y = top_y + dy;
		if (y < 0  ||  y >= model->height)
			continue;

		for (dx = -1; dx <= 1; dx++)
		{
			int x = top_x + dx;
			if (x < 0  ||  x >= model->width)
				continue;

			model->cells[y * model->width + x] = (unsigned short) random();
		}
	}

	ptop->x = top_x;
	ptop->y = top_y;
}



/* Draw one step of the hack.  Positions are cell coordinates. */
static void
draw_step (CriticalSettings *settings,
int pos, XPoint *history, Image *col)
{
	int cell_size = settings->cell_size;
	int half = cell_size/2;
	int old_pos = (pos + settings->trail - 1) % settings->trail;

	pos = pos % settings->trail;

	//  XDrawLine (dpy, window, gc, 
	//	     history[pos].x * cell_size + half,
	//	     history[pos].y * cell_size + half,
	//	     history[old_pos].x * cell_size + half,
	//	     history[old_pos].y * cell_size + half);
	line(screen, Pt(screen->r.min.x + history[pos].x * cell_size + half,
		screen->r.min.y +history[pos].y * cell_size + half),
		Pt(screen->r.min.x +history[old_pos].x * cell_size + half,
		screen->r.min.y +history[old_pos].y * cell_size + half),
		Endsquare, Endsquare, 0, col, ZP);
	frames++;
	flushimage(display, 1);
}


/* Display a self-organizing criticality screen hack.  The program
   runs indefinately on the root window. */
void
screenhack (void)
{
	int			n_colors = 256;
	int			lines_per_color = 10;
	int			i_color = 0;
	long		delay_usecs;
	int			batchcount;
	int			pos = 0;
	int			wrapped = 0;

	int			n_restart, i_restart;

	batchcount = 1500;
	if (batchcount < 5)
		batchcount = 5;

	/* For the moment the model size is just fixed -- making it vary
     with the screen size just makes the hack boring on large
     screens. */
	model_w = 80;
	settings.cell_size = Dx(screen->r) / model_w;
	model_h = Dy(screen->r) / settings.cell_size;

	/* Construct the initial model state. */

	settings.trail = clip(2, 50, 1000);

	history = malloc (sizeof history[0] * settings.trail);
	if (!history)
	{
		fprint (2, "critical: "
			"couldn't allocate trail history of %d cells\n",
			settings.trail);
		return;
	}

	model = model_allocate (model_w, model_h);
	if (!model)
	{
		fprint (2, "critical: error preparing the model\n");
		return;
	}


	delay_usecs = 10;
	n_restart = 8;

	/* xscreensaver will kill or stop us when the user does something
   * that deserves attention. */
	i_restart = 0;

	while (1) {
		int i_batch;

		if (i_restart == 0)
		{
			/* Time to start a new simulation, this one has probably got
	   to be a bit boring. */
			draw(screen, screen->r, display->black, nil, ZP);
			model_initialize (model);
			pos = 1;
			wrapped = 0;
		}

		for (i_batch = batchcount; i_batch; i_batch--)
		{
			/* Set color */
			if ((i_batch % lines_per_color) == 0)
			{
				i_color = (i_color + 1) % n_colors;
			}

			model_step (model, &history[pos]);

			draw_step (&settings,  pos, history, colors[i_color]);

			/* we use the history as a ring buffer, but don't start erasing until
	   we've wrapped around once. */
			if (++pos >= settings.trail)
			{
				pos -= settings.trail;
				wrapped = 1;
			}

			if (wrapped)
			{
				draw_step (&settings, pos+1, history, display->black);
			}

			if(ecanmouse()) {
				m = emouse();
				if(m.buttons&4)
					if(emenuhit(3, &m, &menu) == 0) {
						threadexitsall(0);
					}
			}

			sleep( delay_usecs);

		}

		i_restart = (i_restart + 1) % n_restart;
	}
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "critical") < 0)
		sysfatal("initdraw failed: %r");


	smooth(colors);
	einit(Emouse);

	screenhack();
}
