/* demon --- David Griffeath's cellular automata */
/*-
 * Copyright (c) 1995 by David Bagley.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * 01-Nov-2000: Allocation checks
 * 10-May-1997: Compatible with xscreensaver
 * 16-Apr-1997: -neighbors 3, 9 (not sound mathematically), 12, and 8 added
 * 30-May-1996: Ron Hitchens <ron@idiom.com>
 *            Fixed memory management that caused leaks
 * 14-Apr-1996: -neighbors 6 runtime-time option added
 * 21-Aug-1995: Coded from A.K. Dewdney's "Computer Recreations", Scientific
 *              American Magazine" Aug 1989 pp 102-105.  Also very similar
 *              to hodgepodge machine described in A.K. Dewdney's "Computer
 *              Recreations", Scientific American Magazine" Aug 1988
 *              pp 104-107.  Also used life.c as a guide.
 */

/*-
 * A cellular universe of 4 phases debris, droplets, defects, and demons.
 */



/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/03
 */

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"


static XPoint hexagonUnit[6] =
{
	{1, -1},
	{1, 1},
	{0, 2},
	{-1, 1},
	{-1, -1},
	{0, -2}
};


static XPoint triangleUnit[2][3] =
{
	{
	{
		1, 0},
		{-1, -1},
		{-1, 1	}
	},
	{
		{
		-1, 0},
		{1, -1},
		{1, 1	}
}
};

/*-
  Grid     Number of Neighbors
  ----     ------------------
  Square   4 or 8
  Hexagon  6
  Triangle 3, 9, or 12
*/

static int  neighbors = 0;


#define REDRAWSTEP 2000		/* How many cells to draw per cycle */
#define MINSTATES 2
#define MINGRIDSIZE 24
#define MINSIZE 4
#define NEIGHBORKINDS 6

/* Singly linked list */
typedef struct _CellList {
	XPoint      pt;
	struct _CellList *next;
} CellList;

typedef struct {
	int         generation;
	int         xs, ys;
	int         xb, yb;
	int         nrows, ncols;
	int         width, height;
	int         states;
	int         state;
	int         redrawing, redrawpos;
	int        *ncells;
	CellList  **cellList;
	unsigned char *oldcell, *newcell;
	int         neighbors;
	int         init_bits;
	union {
		XPoint      hexagon[6];
		XPoint      triangle[2][3];
	} shape;
} demonstruct;

static char plots[2][NEIGHBORKINDS] =
{
	{3, 4, 6, 8, 9, 12},
	{12, 16, 18, 20, 22, 24}	/* Number of states */
};



static demonstruct demons;

static void
drawcell(int col, int row, unsigned char state)
{
	demonstruct *dp = &demons;
	int i;
	Point p;
	Image *colr;

	if (!state) {
		//XSetForeground(MI_DISPLAY(mi), MI_GC(mi), MI_BLACK_PIXEL(mi));
		//gc = MI_GC(mi);
		colr = display->black;
	} else {
		//XSetForeground(MI_DISPLAY(mi), MI_GC(mi),
		//    MI_PIXEL(mi, (((int) state - 1) * MI_NPIXELS(mi) /
		//    (dp->states - 1)) % MI_NPIXELS(mi)));
		//gc = MI_GC(mi);
		colr = colors[(((int) state - 1) * 256 / (dp->states - 1)) % 256];
	}

	if (dp->neighbors == 6) {
		int         ccol = 2 * col + !(row & 1), crow = 2 * row;

		p.x = dp->xb + ccol * dp->xs + screen->r.min.x;
		p.y = dp->yb + crow * dp->ys + screen->r.min.y;

		for (i = 0; i < 6; i++) {
			dp->shape.hexagon[i].x = p.x + (dp->xs - 1) * hexagonUnit[i].x;
			dp->shape.hexagon[i].y = p.y + ((dp->ys - 1) * hexagonUnit[i].y / 2) * 4 / 3;
		}

		//XFillPolygon(MI_DISPLAY(mi), MI_WINDOW(mi), gc,
		//    dp->shape.hexagon, 6, Convex, CoordModePrevious);
		fillpoly(screen, dp->shape.hexagon, 6, 1, 
			colr, screen->r.min);
	} else if (dp->neighbors == 4 || dp->neighbors == 8) {
		//XFillRectangle(MI_DISPLAY(mi), MI_WINDOW(mi), gc,
		//    dp->xb + dp->xs * col, dp->yb + dp->ys * row,
		//    dp->xs - (dp->xs > 3), dp->ys - (dp->ys > 3));
		draw(screen, rectaddpt(
			Rect(dp->xb + dp->xs * col, dp->yb + dp->ys * row, 
			dp->xb + dp->xs * col+dp->xs - (dp->xs > 3), 
			dp->yb + dp->ys * row + dp->ys - (dp->ys > 3)), 
			screen->r.min), colr, nil, ZP);

	} else {		/* TRI */
		int         orient = (col + row) % 2;	/* O left 1 right */

		p.x = dp->xb + col * dp->xs + screen->r.min.x;
		p.y = dp->yb + row * dp->ys + screen->r.min.y;

		for (i = 0; i < 6; i++) {
			dp->shape.triangle[orient][i].x = p.x + (dp->xs - 2) * triangleUnit[orient][i].x;
			dp->shape.triangle[orient][i].y = p.y + (dp->ys - 2) * triangleUnit[orient][i].y;
		}

		if (dp->xs <= 3 || dp->ys <= 3)
			//XDrawPoint(MI_DISPLAY(mi), MI_WINDOW(mi), gc,
			//    ((orient) ? -1 : 1) + dp->shape.triangle[orient][0].x,
			//    dp->shape.triangle[orient][0].y);
			draw(screen, rectaddpt(
				Rect(((orient) ? -1 : 1) + dp->shape.triangle[orient][0].x, 
				dp->shape.triangle[orient][0].y,
				((orient) ? -1 : 1) + dp->shape.triangle[orient][0].x+1, 
				dp->shape.triangle[orient][0].y+1
				), 
				screen->r.min), colr, nil, ZP);
		else {
			//			if (orient)
			//				dp->shape.triangle[orient][0].x += (dp->xs / 2 - 1);
			//			else
			//				dp->shape.triangle[orient][0].x -= (dp->xs / 2 - 1);
			//XFillPolygon(MI_DISPLAY(mi), MI_WINDOW(mi), gc,
			//    dp->shape.triangle[orient], 3, Convex, CoordModePrevious);
			fillpoly(screen, dp->shape.triangle[orient], 3, 1, 
				colr, screen->r.min);

		}
	}
}

static Bool
addtolist(int col, int row, unsigned char state)
{
	demonstruct *dp = &demons;
	CellList   *current;

	current = dp->cellList[state];
	if ((dp->cellList[state] = (CellList *)
		malloc(sizeof (CellList))) == NULL) {
		return False;
	}
	dp->cellList[state]->pt.x = col;
	dp->cellList[state]->pt.y = row;
	dp->cellList[state]->next = current;
	dp->ncells[state]++;
	return True;
}


static void
free_state(demonstruct * dp, int state)
{
	CellList   *current;

	while (dp->cellList[state]) {
		current = dp->cellList[state];
		dp->cellList[state] = dp->cellList[state]->next;
		free((void *) current);
	}
	dp->cellList[state] = (CellList *) NULL;
	if (dp->ncells != NULL)
		dp->ncells[state] = 0;
}


static void
free_list(demonstruct * dp)
{
	int         state;

	for (state = 0; state < dp->states; state++)
		free_state(dp, state);
	free((void *) dp->cellList);
	dp->cellList = (CellList **) NULL;
}

static void
free_struct(demonstruct * dp)
{
	if (dp->cellList != NULL) {
		free_list(dp);
	}
	if (dp->ncells != NULL) {
		free((void *) dp->ncells);
		dp->ncells = (int *) NULL;
	}
	if (dp->oldcell != NULL) {
		free((void *) dp->oldcell);
		dp->oldcell = (unsigned char *) NULL;
	}
	if (dp->newcell != NULL) {
		free((void *) dp->newcell);
		dp->newcell = (unsigned char *) NULL;
	}
}

static void
free_demon(demonstruct *dp)
{

	dp->init_bits = 0;
	free_struct(dp);
}

static Bool
draw_state(int state)
{
	demonstruct *dp = &demons;
	CellList   *current;
	Image 		*colr;
	Point p;
	int i;

	if (!state) {
		//XSetForeground(MI_DISPLAY(mi), MI_GC(mi), MI_BLACK_PIXEL(mi));
		//gc = MI_GC(mi);
		colr = display->black;
	} else {
		//XSetForeground(MI_DISPLAY(mi), MI_GC(mi),
		//    MI_PIXEL(mi, (((int) state - 1) * MI_NPIXELS(mi) /
		//    (dp->states - 1)) % MI_NPIXELS(mi)));
		//gc = MI_GC(mi);
		colr = colors[(((int) state - 1) * 256 / (dp->states - 1)) % 256];
	}

	if (dp->neighbors == 6) {	/* Draw right away, slow */
		current = dp->cellList[state];
		while (current) {
			Point p;
			int         col, row, ccol, crow, i;

			col = current->pt.x;
			row = current->pt.y;
			ccol = 2 * col + !(row & 1), crow = 2 * row;
			p.x = dp->xb + ccol * dp->xs;
			p.y = dp->yb + crow * dp->ys;
			p = addpt(p, screen->r.min);

			for (i = 0; i < 6; i++) {
				dp->shape.hexagon[i].x = p.x + (dp->xs - 1) * hexagonUnit[i].x;
				dp->shape.hexagon[i].y = p.y + ((dp->ys - 1) * hexagonUnit[i].y / 2) * 4 / 3;
			}

			//XFillPolygon(MI_DISPLAY(mi), MI_WINDOW(mi), gc,
			//    dp->shape.hexagon, 6, Convex, CoordModePrevious);
			fillpoly(screen, dp->shape.hexagon, 6, 1, 
				colr, screen->r.min);


			current = current->next;
		}
	} else if (dp->neighbors == 4 || dp->neighbors == 8) {
		/* Take advantage of XDrawRectangles */
		int         ncells = 0;

		/* Create Rectangle list from part of the cellList */
		//		if ((rects = (XRectangle *) malloc(dp->ncells[state] *
		//		    sizeof (XRectangle))) == NULL) {
		//			return False;
		//		}
		current = dp->cellList[state];
		while (current) {
			//rects[ncells].x = dp->xb + current->pt.x * dp->xs;
			//rects[ncells].y = dp->yb + current->pt.y * dp->ys;
			//rects[ncells].width = dp->xs - (dp->xs > 3);
			//rects[ncells].height = dp->ys - (dp->ys > 3);
			draw(screen, rectaddpt(
				Rect(dp->xb + current->pt.x * dp->xs,
				dp->yb + current->pt.y * dp->ys,
				dp->xb + current->pt.x * dp->xs + dp->xs - (dp->xs > 3),
				dp->yb + current->pt.y * dp->ys + dp->ys - (dp->ys > 3)), 
				screen->r.min), colr, nil, ZP);

			current = current->next;
			ncells++;
		}
		/* Finally get to draw */
		//XFillRectangles(MI_DISPLAY(mi), MI_WINDOW(mi), gc, rects, ncells);
		/* Free up rects list and the appropriate part of the cellList */
		//(void) free((void *) rects);
	} else {		/* TRI */
		current = dp->cellList[state];
		while (current) {
			int         col, row, orient;

			col = current->pt.x;
			row = current->pt.y;
			orient = (col + row) % 2;	/* O left 1 right */

			p.x = dp->xb + col * dp->xs + screen->r.min.x;
			p.y = dp->yb + row * dp->ys + screen->r.min.y;

			for (i = 0; i < 6; i++) {
				dp->shape.triangle[orient][i].x = p.x + (dp->xs - 2) * triangleUnit[orient][i].x;
				dp->shape.triangle[orient][i].y = p.y + (dp->ys - 2) * triangleUnit[orient][i].y;
			}


			if (dp->xs <= 3 || dp->ys <= 3)
				//XDrawPoint(MI_DISPLAY(mi), MI_WINDOW(mi), gc,
				//    ((orient) ? -1 : 1) + dp->shape.triangle[orient][0].x,
				//    dp->shape.triangle[orient][0].y);
				draw(screen, rectaddpt(
					Rect(((orient) ? -1 : 1) + dp->shape.triangle[orient][0].x, 
					dp->shape.triangle[orient][0].y, 
					((orient) ? -1 : 1) + dp->shape.triangle[orient][0].x+1, 
					dp->shape.triangle[orient][0].y+1), 
					screen->r.min), colr, nil, ZP);
			else {
				//			if (orient)
				//				dp->shape.triangle[orient][0].x += (dp->xs / 2 - 1);
				//			else
				//				dp->shape.triangle[orient][0].x -= (dp->xs / 2 - 1);
				//XFillPolygon(MI_DISPLAY(mi), MI_WINDOW(mi), gc,
				//    dp->shape.triangle[orient], 3, Convex, CoordModePrevious);
				fillpoly(screen, dp->shape.triangle[orient], 3, 1, 
					colr, screen->r.min);
			}
			current = current->next;
		}
	}
	free_state(dp, state);
	//XFlush(MI_DISPLAY(mi));
	flushimage(display, 1);
	return True;
}

static void
RandomSoup(void)
{
	demonstruct *dp = &demons;
	int         row, col, mrow = 0;

	for (row = 0; row < dp->nrows; ++row) {
		for (col = 0; col < dp->ncols; ++col) {
			dp->oldcell[col + mrow] =
				(unsigned char) LRAND() % (dp->states);
			if (!addtolist(col, row, dp->oldcell[col + mrow]))
				return; /* sparse soup */
		}
		mrow += dp->ncols;
	}
}

void
init_demon(void)
{
	int         size = -7, nk;
	demonstruct *dp;

	dp = &demons;

	dp->generation = 0;
	dp->redrawing = 0;
	free_struct(dp);

	switch(nrand(3)) {
	case 0: 
		neighbors = 4; 
		break;
	case 1: 
		neighbors = 6; 
		break;
	case 2: 
		neighbors = 8; 
		break;
	}

	for (nk = 0; nk < NEIGHBORKINDS; nk++) {
		if (neighbors == plots[0][nk]) {
			dp->neighbors = plots[0][nk];
			break;
		}
		if (nk == NEIGHBORKINDS - 1) {
			nk = NRAND(NEIGHBORKINDS);
			dp->neighbors = plots[0][nk];
			break;
		}
	}


	dp->states = 0;
	if (dp->states < -MINSTATES)
		dp->states = NRAND(-dp->states - MINSTATES + 1) + MINSTATES;
	else if (dp->states < MINSTATES)
		dp->states = plots[1][nk];
	if ((dp->cellList = (CellList **) calloc(dp->states,
		sizeof (CellList *))) == NULL) {
		free_demon(dp);
		return;
	}
	if ((dp->ncells = (int *) calloc(dp->states, sizeof (int))) == NULL) {
		free_demon(dp);
		return;
	}

	dp->state = 0;

	dp->width = Dx(screen->r);
	dp->height = Dy(screen->r);

	if (dp->neighbors == 6) {
		int         nccols, ncrows;

		if (dp->width < 8)
			dp->width = 8;
		if (dp->height < 8)
			dp->height = 8;
		if (size < -MINSIZE)
			dp->ys = NRAND(MIN(-size, MAX(MINSIZE, MIN(dp->width, dp->height) /
				MINGRIDSIZE)) - MINSIZE + 1) + MINSIZE;
		else if (size < MINSIZE) {
			if (!size)
				dp->ys = MAX(MINSIZE, MIN(dp->width, dp->height) / MINGRIDSIZE);
			else
				dp->ys = MINSIZE;
		} else
			dp->ys = MIN(size, MAX(MINSIZE, MIN(dp->width, dp->height) /
				MINGRIDSIZE));
		dp->xs = dp->ys;
		nccols = MAX(dp->width / dp->xs - 2, 2);
		ncrows = MAX(dp->height / dp->ys - 1, 4);
		dp->ncols = nccols / 2;
		dp->nrows = 2 * (ncrows / 4);
		dp->xb = (dp->width - dp->xs * nccols) / 2 + dp->xs / 2;
		dp->yb = (dp->height - dp->ys * (ncrows / 2) * 2) / 2 + dp->ys - 2;
	} else if (dp->neighbors == 4 || dp->neighbors == 8) {
		if (size < -MINSIZE)
			dp->ys = NRAND(MIN(-size, MAX(MINSIZE, MIN(dp->width, dp->height) /
				MINGRIDSIZE)) - MINSIZE + 1) + MINSIZE;
		else if (size < MINSIZE) {
			if (!size)
				dp->ys = MAX(MINSIZE, MIN(dp->width, dp->height) / MINGRIDSIZE);
			else
				dp->ys = MINSIZE;
		} else
			dp->ys = MIN(size, MAX(MINSIZE, MIN(dp->width, dp->height) /
				MINGRIDSIZE));
		dp->xs = dp->ys;
		dp->ncols = MAX(dp->width / dp->xs, 2);
		dp->nrows = MAX(dp->height / dp->ys, 2);
		dp->xb = (dp->width - dp->xs * dp->ncols) / 2;
		dp->yb = (dp->height - dp->ys * dp->nrows) / 2;
	} else {		/* TRI */

		if (dp->width < 2)
			dp->width = 2;
		if (dp->height < 2)
			dp->height = 2;
		if (size < -MINSIZE)
			dp->ys = NRAND(MIN(-size, MAX(MINSIZE, MIN(dp->width, dp->height) /
				MINGRIDSIZE)) - MINSIZE + 1) + MINSIZE;
		else if (size < MINSIZE) {
			if (!size)
				dp->ys = MAX(MINSIZE, MIN(dp->width, dp->height) / MINGRIDSIZE);
			else
				dp->ys = MINSIZE;
		} else
			dp->ys = MIN(size, MAX(MINSIZE, MIN(dp->width, dp->height) /
				MINGRIDSIZE));
		dp->xs = (int) (1.52 * dp->ys);
		dp->ncols = (MAX(dp->width / dp->xs - 1, 2) / 2) * 2;
		dp->nrows = (MAX(dp->height / dp->ys - 1, 2) / 2) * 2;
		dp->xb = (dp->width - dp->xs * dp->ncols) / 2 + dp->xs / 2;
		dp->yb = (dp->height - dp->ys * dp->nrows) / 2 + dp->ys / 2;
	}

	//MI_CLEARWINDOW(mi);
	draw(screen, screen->r, display->black, nil, ZP);

	if ((dp->oldcell = (unsigned char *)
		malloc(dp->ncols * dp->nrows * sizeof (unsigned char))) == NULL) {
		free_demon(dp);
		return;
	}

	if ((dp->newcell = (unsigned char *)
		malloc(dp->ncols * dp->nrows * sizeof (unsigned char))) == NULL) {
		free_demon(dp);
		return;
	}

	RandomSoup();
}

void
draw_demon(void)
{
	int         i, j, k, l, mj = 0, ml;
	demonstruct *dp;

	dp = &demons;
	if (dp->cellList == NULL)
		return;

	if (dp->state >= dp->states) {
		memcpy((char *) dp->newcell, (char *) dp->oldcell,
			dp->ncols * dp->nrows * sizeof (unsigned char));

		if (dp->neighbors == 6) {
			for (j = 0; j < dp->nrows; j++) {
				for (i = 0; i < dp->ncols; i++) {
					/* NE */
					if (!(j & 1))
						k = (i + 1 == dp->ncols) ? 0 : i + 1;
					else
						k = i;
					l = (!j) ? dp->nrows - 1 : j - 1;
					ml = l * dp->ncols;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
					/* E */
					k = (i + 1 == dp->ncols) ? 0 : i + 1;
					ml = mj;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
					/* SE */
					if (!(j & 1))
						k = (i + 1 == dp->ncols) ? 0 : i + 1;
					else
						k = i;
					l = (j + 1 == dp->nrows) ? 0 : j + 1;
					ml = l * dp->ncols;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
					/* SW */
					if (j & 1)
						k = (!i) ? dp->ncols - 1 : i - 1;
					else
						k = i;
					l = (j + 1 == dp->nrows) ? 0 : j + 1;
					ml = l * dp->ncols;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
					/* W */
					k = (!i) ? dp->ncols - 1 : i - 1;
					ml = mj;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
					/* NW */
					if (j & 1)
						k = (!i) ? dp->ncols - 1 : i - 1;
					else
						k = i;
					l = (!j) ? dp->nrows - 1 : j - 1;
					ml = l * dp->ncols;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
				}
				mj += dp->ncols;
			}
		} else if (dp->neighbors == 4 || dp->neighbors == 8) {
			for (j = 0; j < dp->nrows; j++) {
				for (i = 0; i < dp->ncols; i++) {
					/* N */
					k = i;
					l = (!j) ? dp->nrows - 1 : j - 1;
					ml = l * dp->ncols;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
					/* E */
					k = (i + 1 == dp->ncols) ? 0 : i + 1;
					ml = mj;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
					/* S */
					k = i;
					l = (j + 1 == dp->nrows) ? 0 : j + 1;
					ml = l * dp->ncols;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
					/* W */
					k = (!i) ? dp->ncols - 1 : i - 1;
					//					l = j;
					ml = mj;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
				}
				mj += dp->ncols;
			}
			if (dp->neighbors == 8) {
				mj = 0;
				for (j = 0; j < dp->nrows; j++) {
					for (i = 0; i < dp->ncols; i++) {
						/* NE */
						k = (i + 1 == dp->ncols) ? 0 : i + 1;
						l = (!j) ? dp->nrows - 1 : j - 1;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
						/* SE */
						k = (i + 1 == dp->ncols) ? 0 : i + 1;
						l = (j + 1 == dp->nrows) ? 0 : j + 1;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
						/* SW */
						k = (!i) ? dp->ncols - 1 : i - 1;
						l = (j + 1 == dp->nrows) ? 0 : j + 1;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
						/* NW */
						k = (!i) ? dp->ncols - 1 : i - 1;
						l = (!j) ? dp->nrows - 1 : j - 1;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
					}
					mj += dp->ncols;
				}
			}
		} else if (dp->neighbors == 3 || dp->neighbors == 9 ||
			dp->neighbors == 12) {
			for (j = 0; j < dp->nrows; j++) {
				for (i = 0; i < dp->ncols; i++) {
					if ((i + j) % 2) {	/* right */
						/* W */
						k = (!i) ? dp->ncols - 1 : i - 1;
						ml = mj;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
					} else {	/* left */
						/* E */
						k = (i + 1 == dp->ncols) ? 0 : i + 1;
						ml = mj;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
					}
					/* N */
					k = i;
					l = (!j) ? dp->nrows - 1 : j - 1;
					ml = l * dp->ncols;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
					/* S */
					k = i;
					l = (j + 1 == dp->nrows) ? 0 : j + 1;
					ml = l * dp->ncols;
					if (dp->oldcell[k + ml] ==
						(int) (dp->oldcell[i + mj] + 1) % dp->states)
						dp->newcell[i + mj] = dp->oldcell[k + ml];
				}
				mj += dp->ncols;
			}
			if (dp->neighbors == 9 || dp->neighbors == 12) {
				mj = 0;
				for (j = 0; j < dp->nrows; j++) {
					for (i = 0; i < dp->ncols; i++) {
						/* NN */
						k = i;
						if (!j)
							l = dp->nrows - 2;
						else if (!(j - 1))
							l = dp->nrows - 1;
						else
							l = j - 2;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
						/* SS */
						k = i;
						if (j + 1 == dp->nrows)
							l = 1;
						else if (j + 2 == dp->nrows)
							l = 0;
						else
							l = j + 2;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
						/* NW */
						k = (!i) ? dp->ncols - 1 : i - 1;
						l = (!j) ? dp->nrows - 1 : j - 1;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
						/* NE */
						k = (i + 1 == dp->ncols) ? 0 : i + 1;
						l = (!j) ? dp->nrows - 1 : j - 1;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
						/* SW */
						k = (!i) ? dp->ncols - 1 : i - 1;
						l = (j + 1 == dp->nrows) ? 0 : j + 1;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
						/* SE */
						k = (i + 1 == dp->ncols) ? 0 : i + 1;
						l = (j + 1 == dp->nrows) ? 0 : j + 1;
						ml = l * dp->ncols;
						if (dp->oldcell[k + ml] ==
							(int) (dp->oldcell[i + mj] + 1) % dp->states)
							dp->newcell[i + mj] = dp->oldcell[k + ml];
					}
					mj += dp->ncols;
				}
				if (dp->neighbors == 12) {
					mj = 0;
					for (j = 0; j < dp->nrows; j++) {
						for (i = 0; i < dp->ncols; i++) {
							if ((i + j) % 2) {	/* right */
								/* NNW */
								k = (!i) ? dp->ncols - 1 : i - 1;
								if (!j)
/* code folded from here */
	l = dp->nrows - 2;
/* unfolding */
								else if (!(j - 1))
/* code folded from here */
	l = dp->nrows - 1;
/* unfolding */
								else
/* code folded from here */
	l = j - 2;
/* unfolding */
								ml = l * dp->ncols;
								if (dp->oldcell[k + ml] ==
									(int) (dp->oldcell[i + mj] + 1) % dp->states)
/* code folded from here */
	dp->newcell[i + mj] = dp->oldcell[k + ml];
/* unfolding */
								/* SSW */
								k = (!i) ? dp->ncols - 1 : i - 1;
								if (j + 1 == dp->nrows)
/* code folded from here */
	l = 1;
/* unfolding */
								else if (j + 2 == dp->nrows)
/* code folded from here */
	l = 0;
/* unfolding */
								else
/* code folded from here */
	l = j + 2;
/* unfolding */
								ml = l * dp->ncols;
								if (dp->oldcell[k + ml] ==
									(int) (dp->oldcell[i + mj] + 1) % dp->states)
/* code folded from here */
	dp->newcell[i + mj] = dp->oldcell[k + ml];
/* unfolding */
								/* EE */
								k = (i + 1 == dp->ncols) ? 0 : i + 1;
								//								l = j;
								ml = mj;
								if (dp->oldcell[k + ml] ==
									(int) (dp->oldcell[i + mj] + 1) % dp->states)
/* code folded from here */
	dp->newcell[i + mj] = dp->oldcell[k + ml];
/* unfolding */
							} else {	/* left */
								/* NNE */
								k = (i + 1 == dp->ncols) ? 0 : i + 1;
								if (!j)
/* code folded from here */
	l = dp->nrows - 2;
/* unfolding */
								else if (!(j - 1))
/* code folded from here */
	l = dp->nrows - 1;
/* unfolding */
								else
/* code folded from here */
	l = j - 2;
/* unfolding */
								ml = l * dp->ncols;
								if (dp->oldcell[k + ml] ==
									(int) (dp->oldcell[i + mj] + 1) % dp->states)
/* code folded from here */
	dp->newcell[i + mj] = dp->oldcell[k + ml];
/* unfolding */
								/* SSE */
								k = (i + 1 == dp->ncols) ? 0 : i + 1;
								if (j + 1 == dp->nrows)
/* code folded from here */
	l = 1;
/* unfolding */
								else if (j + 2 == dp->nrows)
/* code folded from here */
	l = 0;
/* unfolding */
								else
/* code folded from here */
	l = j + 2;
/* unfolding */
								ml = l * dp->ncols;
								if (dp->oldcell[k + ml] ==
									(int) (dp->oldcell[i + mj] + 1) % dp->states)
/* code folded from here */
	dp->newcell[i + mj] = dp->oldcell[k + ml];
/* unfolding */
								/* WW */
								k = (!i) ? dp->ncols - 1 : i - 1;
								//								l = j;
								ml = mj;
								if (dp->oldcell[k + ml] ==
									(int) (dp->oldcell[i + mj] + 1) % dp->states)
/* code folded from here */
	dp->newcell[i + mj] = dp->oldcell[k + ml];
/* unfolding */
							}
						}
						mj += dp->ncols;
					}
				}
			}
		}
		mj = 0;
		for (j = 0; j < dp->nrows; j++) {
			for (i = 0; i < dp->ncols; i++)
				if (dp->oldcell[i + mj] != dp->newcell[i + mj]) {
					dp->oldcell[i + mj] = dp->newcell[i + mj];
					if (!addtolist(i, j, dp->oldcell[i + mj])) {
						free_demon(dp);
						return;
					}
				}
			mj += dp->ncols;
		}
		if (++dp->generation > 500)
			init_demon();
		dp->state = 0;
	} else {
		if (dp->ncells[dp->state])
			if (!draw_state(dp->state)) {
				free_demon(dp);
				return;
			}
		dp->state++;
	}
	if (dp->redrawing) {
		for (i = 0; i < REDRAWSTEP; i++) {
			if (dp->oldcell[dp->redrawpos]) {
				drawcell(dp->redrawpos % dp->ncols, dp->redrawpos / dp->ncols,
					dp->oldcell[dp->redrawpos]);
			}
			if (++(dp->redrawpos) >= dp->ncols * dp->nrows) {
				dp->redrawing = 0;
				break;
			}
		}
	}
}
void
release_demon(void)
{
	free_demon(&demons);
}

void
refresh_demon(void)
{
	demonstruct *dp;

	dp = &demons;

	dp->redrawing = 1;
	dp->redrawpos = 0;
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	release_demon();
	init_demon();
}


void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "demon") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			cmap2rgba(i<<2));
	}

	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_demon();
		frames++;
	}
}
