/* discrete --- chaotic mappings */

/*-
 * Copyright (c) 1996 by Tim Auckland <Tim.Auckland@Procket.com>
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * "discrete" shows a number of fractals based on the "discrete map"
 * type of dynamical systems.  They include a different way of looking
 * at the HOPALONG system, an inverse julia-set iteration, the "Standard
 * Map" and the "Bird in a Thornbush" fractal.
 *
 * Revision History:
 * 01-Nov-2000: Allocation checks
 * 31-Jul-1997: Ported to xlockmore-4
 * 08-Aug-1996: Adapted from hop.c Copyright (c) 1991 by Patrick J. Naughton.
 */

/*
  * ported to Plan 9 by andrey@lanl.gov, 07/02
  */

/* plan9-related stuff */
#define COUNT 5000
#define CYCLES 200
#include "plan9.h"
#include "smooth.h"


enum ftypes {
	SQRT, BIRDIE, STANDARD, TRIG, CUBIC, HENON, AILUJ, HSHOE, DELOG
};


#define BIASES 18
static enum ftypes bias[BIASES] =
{
	STANDARD, STANDARD, STANDARD, STANDARD,
	SQRT, SQRT, SQRT, SQRT,
	BIRDIE, BIRDIE, BIRDIE,
	AILUJ, AILUJ, AILUJ,
	TRIG, TRIG,
	CUBIC,
	HENON,
};


typedef struct {
	int         maxx;
	int         maxy;	/* max of the screen */
	double      a;
	double      b;
	double      c;
	double      d;
	double      e;
	double      i;
	double      j;		/* discrete parameters */
	double      ic;
	double      jc;
	double      is;
	double      js;
	int         inc;
	int         pix;
	enum ftypes op;
	int         count;
	XPoint     *pointBuffer;	/* pointer for XDrawPoints */
} discretestruct;

static discretestruct *discretes = (discretestruct *) NULL;

void
init_discrete(void)
{
	double      range;
	discretestruct *hp;

	if (discretes == NULL) {
		if ((discretes =
			(discretestruct *) calloc(1,
			sizeof (discretestruct))) == NULL)
			return;
	}
	hp = &discretes[0];

	hp->maxx = Dx(screen->r);
	hp->maxy = Dy(screen->r);

	hp->op = bias[LRAND() % BIASES];

	switch (hp->op) {
	case HSHOE:
		hp->ic = 0;
		hp->jc = 0;
		hp->is = hp->maxx / 4.0;
		hp->js = hp->maxy / 4.0;
		hp->a = 0.5;
		hp->b = 0.5;
		hp->c = 0.2;
		hp->d = -1.25;
		hp->e = 1;
		hp->i = hp->j = 0.0;
		break;
	case DELOG:
		hp->ic = 0.5;
		hp->jc = 0.3;
		hp->is = hp->maxx / 1.5;
		hp->js = hp->maxy / 1.5;
		hp->a = 2.176399;
		hp->i = hp->j = 0.01;
		break;
	case HENON:
		hp->jc = ((frand()) * 2.0 - 1.0) * 0.4;
		hp->ic = 1.3 * (1 - (hp->jc * hp->jc) / (0.4 * 0.4));
		hp->is = hp->maxx;
		hp->js = hp->maxy * 1.5;
		hp->a = 1;
		hp->b = 1.4;
		hp->c = 0.3;
		hp->i = hp->j = 0;
		break;
	case SQRT:
		hp->ic = 0;
		hp->jc = 0;
		hp->is = 1;
		hp->js = 1;
		range = sqrt((double) hp->maxx * 2 * hp->maxx * 2 +
			(double) hp->maxy * 2 * hp->maxy * 2) /
			(10.0 + LRAND() % 10);

		hp->a = (frand()) * range - range / 2.0;
		hp->b = (frand()) * range - range / 2.0;
		hp->c = (frand()) * range - range / 2.0;
		if (!(LRAND() % 2))
			hp->c = 0.0;
		hp->i = hp->j = 0.0;
		break;
	case STANDARD:
		hp->ic = M_PI;
		hp->jc = M_PI;
		hp->is = hp->maxx / (M_PI * 2);
		hp->js = hp->maxy / (M_PI * 2);
		hp->a = 0;	/* decay */
		hp->b = (frand()) * 2.0;
		hp->c = 0;
		hp->i = M_PI;
		hp->j = M_PI;
		break;
	case BIRDIE:
		hp->ic = 0;
		hp->jc = 0;
		hp->is = hp->maxx / 2;
		hp->js = hp->maxy / 2;
		hp->a = 1.99 + ((frand()) * 2.0 - 1.0) * 0.2;
		hp->b = 0;
		hp->c = 0.8 + ((frand()) * 2.0 - 1.0) * 0.1;
		hp->i = hp->j = 0;
		break;
	case TRIG:
		hp->a = 5;
		hp->b = 0.5 + ((frand()) * 2.0 - 1.0) * 0.3;
		hp->ic = hp->a;
		hp->jc = 0;
		hp->is = hp->maxx / (hp->b * 20);
		hp->js = hp->maxy / (hp->b * 20);
		hp->i = hp->j = 0;
		break;
	case CUBIC:
		hp->a = 2.77;
		hp->b = 0.1 + ((frand()) * 2.0 - 1.0) * 0.1;
		hp->ic = 0;
		hp->jc = 0;
		hp->is = hp->maxx / 4;
		hp->js = hp->maxy / 4;
		hp->i = hp->j = 0.1;
		break;
	case AILUJ:
		{
			int         i;
			double      x, y, xn, yn;

			hp->ic = 0;
			hp->jc = 0;
			hp->is = hp->maxx / 4;
			hp->js = hp->maxx / 4;
			do 
			{
				hp->a = ((frand()) * 2.0 - 1.0) * 1.5 - 0.5;
				hp->b = ((frand()) * 2.0 - 1.0) * 1.5;
				x = y = 0;
#define MAXITER 10
				for (i = 0; i < MAXITER && x * x + y * y < 13; i++) {	/* 'Brot calc */
					xn = x * x - y * y + hp->a;
					yn = 2 * x * y + hp->b;
					x = xn;
					y = yn;
				}
			}			while (i < MAXITER);	/* wait for a connected set */
			hp->i = hp->j = 0.1;
			break;
		}
	}
	hp->pix = 0;
	hp->inc = 0;

	if (hp->pointBuffer == NULL) {
		hp->pointBuffer = (XPoint *) malloc(sizeof (XPoint) * COUNT);
		/* if fails will check later */
	}

	draw(screen, screen->r, display->black, nil, ZP);
	hp->count = 0;
}


void
draw_discrete(void)
{
	double      oldj, oldi;
	int         count = COUNT;
	int         cycles = CYCLES;
	int         k, i2;
	XPoint     *xp;
	discretestruct *hp;

	if (discretes == NULL)
		return;
	hp = &discretes[0];
	if (hp->pointBuffer == NULL)
		return;

	k = count;
	xp = hp->pointBuffer;

	hp->inc++;

	hp->pix++;
	hp->pix %= 256;

	while (k--) {
		oldj = hp->j;
		oldi = hp->i;
		switch (hp->op) {
		case HSHOE:
			{
				int         i;


				if (k < count / 4) {
					hp->i = ((double) k / count) * 8 - 1;
					hp->j = 1;
				}
				else if (k < count / 2) {
					hp->i = 1;
					hp->j = 3 - ((double) k / count) * 8;
				}
				else if (k < 3 * count / 4) {
					hp->i = 5 - ((double) k / count) * 8;
					hp->j = -1;
				}
				else {
					hp->i = -1;
					hp->j = ((double) k / count) * 8 - 7;
				}
				for (i = 1; i < (hp->inc % 15); i++) {
					oldj = hp->j;
					oldi = hp->i;
					hp->i = (hp->a * oldi + hp->b) * oldj;
					hp->j = (hp->e - hp->d + hp->c * oldi) * oldj * oldj - hp->c * oldi + hp->d;
				}
				break;
			}
		case DELOG:
			hp->j = oldi;
			hp->i = hp->a * oldi * (1 - oldj);
			break;
		case HENON:
			hp->i = oldj + hp->a - hp->b * oldi * oldi;
			hp->j = hp->c * oldi;
			break;
		case SQRT:
			if (k) {
				hp->j = hp->a + hp->i;
				hp->i = -oldj + (hp->i < 0
					? sqrt(fabs(hp->b * (hp->i - hp->c)))
					: -sqrt(fabs(hp->b * (hp->i - hp->c))));
			}
			else {
				static int  s = 1;

				hp->i = s * hp->inc * hp->maxx / cycles / 2;
				hp->j = hp->a + hp->i;
				s = -s;
			}
			break;
		case STANDARD:
			if (k) {
				hp->j = (1 - hp->a) * oldj + hp->b * sin(oldi) + hp->a * hp->c;
				hp->j = fmod(hp->j + 2 * M_PI, 2 * M_PI);
				hp->i = oldi + hp->j;
				hp->i = fmod(hp->i + 2 * M_PI, 2 * M_PI);
			}
			else {
				static int  s = 1;

				hp->j = M_PI + fmod(s * hp->inc * 2 * M_PI / (cycles - 0.5), M_PI);
				hp->i = M_PI;
				s = -s;
			}
			break;
		case BIRDIE:
			hp->j = oldi;
			hp->i = (1 - hp->c) * cos(M_PI * hp->a * oldj) + hp->c * hp->b;
			hp->b = oldj;
			break;
		case TRIG:
			{
				double      r2 = oldi * oldi + oldj * oldj;

				hp->i = hp->a + hp->b * (oldi * cos(r2) - oldj * sin(r2));
				hp->j = hp->b * (oldj * cos(r2) + oldi * sin(r2));
			}
			break;
		case CUBIC:
			hp->i = oldj;
			hp->j = hp->a * oldj - oldj * oldj * oldj - hp->b * oldi;
			break;
		case AILUJ:
			hp->i = ((LRAND() < MAXRAND / 2) ? -1 : 1) *
				sqrt(((oldi - hp->a) +
				sqrt((oldi - hp->a) * (oldi - hp->a) + (oldj - hp->b) * (oldj - hp->b))) / 2);
			if (hp->i < 0.00000001 && hp->i > -0.00000001)
				hp->i = (hp->i > 0.0) ? 0.00000001 : -0.00000001;
			hp->j = (oldj - hp->b) / (2 * hp->i);
			break;
		}
		xp->x = hp->maxx / 2 + (int) ((hp->i - hp->ic) * hp->is);
		xp->y = hp->maxy / 2 - (int) ((hp->j - hp->jc) * hp->js);
		xp++;
	}
	//	XDrawPoints(dsp, win, gc, hp->pointBuffer, count, CoordModeOrigin);

	for(i2 = 0; i2 < count; i2++) {
		Rectangle r;
		Point p = Pt(screen->r.min.x, screen->r.min.y);

		p = addpt(p, hp->pointBuffer[i2]);
		r = Rpt(p, addpt(p, Pt(1,1)));
		draw(screen, r, colors[hp->pix], nil, ZP);

	}

	if (++hp->count > cycles) {
		draw(screen, screen->r, display->black, nil, ZP);
		init_discrete();
	}
	flushimage(display, 1);
}

void
release_discrete(void)
{
	if (discretes != NULL) {

		discretestruct *hp = &discretes[0];

		if (hp->pointBuffer != NULL) {
			free((void *) hp->pointBuffer);
			/* hp->pointBuffer = NULL; */
		}
		free((void *) discretes);
		discretes = (discretestruct *) NULL;
	}
}


void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	if(discretes)
		release_discrete();
	init_discrete();
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "discrete") < 0)
		sysfatal("initdraw failed: %r");


	smooth(colors);

	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_discrete();
		frames++;
	}
}
