/* xscreensaver, Copyright (c) 1993, 1995, 1996, 1998
 *  Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* This file was ported from xlock for use in xscreensaver (and standalone)
 * by jwz on 18-Oct-93.  (And again, 11-May-97.)  Original copyright reads:
 *
 *   static char sccsid[] = "@(#)flame.c 1.4 91/09/27 XLOCK";
 *
 * flame.c - recursive fractal cosmic flames.
 *
 * Copyright (c) 1991 by Patrick J. Naughton.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Comments and additions should be sent to the author:
 *
 *		       naughton@eng.sun.com
 *
 *		       Patrick J. Naughton
 *		       MS 21-14
 *		       Sun Laboritories, Inc.
 *		       2550 Garcia Ave
 *		       Mountain View, CA  94043
 *
 * Revision History:
 * 01-Jun-95: This should look more like the original with some updates by
 *            Scott Draves.
 * 27-Jun-91: vary number of functions used.
 * 24-Jun-91: fixed portability problem with integer mod (%).
 * 06-Jun-91: Written. (received from Scott Draves, spot@cs.cmu.edu).
 */

/* 
  * ported to Plan 9 by andrey@lanl.gov, 07/02

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"

#define POINT_BUFFER_SIZE 10
#define MAXLEV 4
#define MAXKINDS  10

static double f[2][3][MAXLEV];	/* three non-homogeneous transforms */
static int max_total;
static int max_levels;
static int max_points;
static int cur_level;
static int variation;
static int snum;
static int anum;
static int num_points;
static int total_points;
static int pixcol;
static int ncolors;
static XPoint points [POINT_BUFFER_SIZE];

static int delay, delay2;
static int width, height;

static short
halfrandom (int mv)
{
	static short lasthalf = 0;
	unsigned long r;

	if (lasthalf)
	{
		r = lasthalf;
		lasthalf = 0;
	}
	else
	{
		r = random ();
		lasthalf = r >> 16;
	}
	return (r % mv);
}

static void
init_flame (void)
{


	width = Dx(screen->r);
	height = Dy(screen->r);

	max_points = 25;

	max_levels = max_points;

	max_total = 10000;
	delay = 10;
	delay2 = 1000;

	variation = random() % MAXKINDS;


	ncolors = 256;

	pixcol = halfrandom (ncolors);

}

static int
recurse (double x, double y, int l)
{
	int i, i2;
	double nx, ny;

	if (l == max_levels)
	{
		total_points++;
		if (total_points > max_total) /* how long each fractal runs */
			return 0;

		if (x > -1.0 && x < 1.0 && y > -1.0 && y < 1.0)
		{
			points[num_points].x = (int) ((width / 2) * (x + 1.0));
			points[num_points].y = (int) ((height / 2) * (y + 1.0));
			num_points++;
			if (num_points >= POINT_BUFFER_SIZE)
			{
				//	      XDrawPoints (dpy, win, gc, points, num_points, CoordModeOrigin);

				for(i2 = 0; i2 < num_points; i2++) {
					Rectangle r;
					Point p = Pt(screen->r.min.x, screen->r.min.y);

					p = addpt(p, points[i2]);
					r = Rpt(p, addpt(p, Pt(1,1)));
					draw(screen, r, colors[pixcol], nil, ZP);

				}

				num_points = 0;
				/* if (delay) usleep (delay); */
				/* XSync (dpy, False); */
				flushimage(display, 1);
			}
		}
	}
	else
	{
		for (i = 0; i < snum; i++)
		{

			/* Scale back when values get very large. Spot sez:
	     "I think this happens on HPUX.  I think it's non-IEEE
	     to generate an exception instead of a silent NaN."
	   */
			if ((abs(x) > 1.0E5) || (abs(y) > 1.0E5))
				x = x / y;

			nx = f[0][0][i] * x + f[0][1][i] * y + f[0][2][i];
			ny = f[1][0][i] * x + f[1][1][i] * y + f[1][2][i];
			if (i < anum)
			{
				switch (variation)
				{
				case 0:	/* sinusoidal */
					nx = sin(nx);
					ny = sin(ny);
					break;
				case 1:	/* complex */
					{
						double r2 = nx * nx + ny * ny + 1e-6;
						nx = nx / r2;
						ny = ny / r2;
					}
					break;
				case 2:	/* bent */
					if (nx < 0.0)
						nx = nx * 2.0;
					if (ny < 0.0)
						ny = ny / 2.0;
					break;
				case 3:	/* swirl */
					{
						double r = (nx * nx + ny * ny);	/* times k here is fun */
						double c1 = sin(r);
						double c2 = cos(r);
						double t = nx;

						if (nx > 1e4 || nx < -1e4 || ny > 1e4 || ny < -1e4)
							ny = 1e4;
						else
							ny = c2 * t + c1 * ny;
						nx = c1 * nx - c2 * ny;
					}
					break;
				case 4:	/* horseshoe */
					{
						double r, c1, c2, t;

						/* Avoid atan2: DOMAIN error message */
						if (nx == 0.0 && ny == 0.0)
							r = 0.0;
						else
							r = atan2(nx, ny);      /* times k here is fun */
						c1 = sin(r);
						c2 = cos(r);
						t = nx;

						nx = c1 * nx - c2 * ny;
						ny = c2 * t + c1 * ny;
					}
					break;
				case 5:	/* drape */
					{
						double t;

						/* Avoid atan2: DOMAIN error message */
						if (nx == 0.0 && ny == 0.0)
							t = 0.0;
						else
							t = atan2(nx, ny) / M_PI;

						if (nx > 1e4 || nx < -1e4 || ny > 1e4 || ny < -1e4)
							ny = 1e4;
						else
							ny = sqrt(nx * nx + ny * ny) - 1.0;
						nx = t;
					}
					break;
				case 6:	/* broken */
					if (nx > 1.0)
						nx = nx - 1.0;
					if (nx < -1.0)
						nx = nx + 1.0;
					if (ny > 1.0)
						ny = ny - 1.0;
					if (ny < -1.0)
						ny = ny + 1.0;
					break;
				case 7:	/* spherical */
					{
						double r = 0.5 + sqrt(nx * nx + ny * ny + 1e-6);

						nx = nx / r;
						ny = ny / r;
					}
					break;
				case 8:	/*  */
					nx = atan(nx) / M_PI_2;
					ny = atan(ny) / M_PI_2;
					break;

				default:
					nx = sin(nx);
					ny = sin(ny);
				}
			}
			if (!recurse (nx, ny, l + 1))
				return 0;
		}
	}
	return 1;
}


static void
flame (void)
{
	int i, i2, j, k;
	static int alt = 0;

	if (!(cur_level++ % max_levels))
	{
		if (delay2) sleep (delay2);
		draw(screen, screen->r, display->black, nil, ZP);
		alt = !alt;

		variation = random() % MAXKINDS;
	}
	else
	{
		if (ncolors > 2)
		{
			if (--pixcol <= 0)
				pixcol = ncolors - 1;
		}
	}

	/* number of functions */
	snum = 2 + (cur_level % (MAXLEV - 1));

	/* how many of them are of alternate form */
	if (alt)
		anum = 0;
	else
		anum = halfrandom (snum) + 2;

	/* 6 coefs per function */
	for (k = 0; k < snum; k++)
	{
		for (i = 0; i < 2; i++)
			for (j = 0; j < 3; j++)
				f[i][j][k] = ((double) (random() & 1023) / 512.0 - 1.0);
	}
	num_points = 0;
	total_points = 0;
	recurse (0.0, 0.0, 0);
	//  XDrawPoints (dpy, window, gc, points, num_points, CoordModeOrigin);

	for(i2 = 0; i2 < num_points; i2++) {
		Rectangle r;
		Point p = Pt(screen->r.min.x, screen->r.min.y);

		p = addpt(p, points[i2]);
		r = Rpt(p, addpt(p, Pt(1,1)));
		draw(screen, r, colors[pixcol], nil, ZP);

	}


	if (delay) sleep (delay);
	flushimage(display, 1);
}



void
screenhack (void)
{
	init_flame ();
	while (1)
	{
		flame ();
		frames++;
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	init_flame();
}



void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "flame") < 0)
		sysfatal("initdraw failed: %r");


	smooth(colors);
	einit(Emouse);

	screenhack();
}
