/* fluidballs, Copyright (c) 2000 by Peter Birtles <peter@bqdesign.com.au>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 *
 * Ported to X11 and xscreensaver by jwz, 27-Feb-2002.
 *
 * http://astronomy.swin.edu.au/~pbourke/modelling/fluid/
 *
 * Some physics improvements by Steven Barker <steve@blckknght.org>
 */

/* Future ideas:
 * Specifying a distribution in the ball sizes (with a gamma curve, possibly).
 * Brownian motion, for that extra touch of realism.
 *
 * It would be nice to detect when there are more balls than fit in
 * the window, and scale the number of balls back.  Useful for the
 * xscreensaver-demo preview, which is often too tight by default.
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 08/03
 */

/* plan9-related stuff */
#include "plan9.h"

typedef struct b_state b_state;
struct b_state {
	Image * fg;
	Image * fg2;

	int count;		/* number of balls */
	float xmin, ymin;	/* rectangle of window, relative to root */
	float xmax, ymax;

	int mouse_ball;	/* index of ball being dragged, or 0 if none. */

	float tc;		/* time constant (time-warp multiplier) */
	float accx;		/* horizontal acceleration (wind) */
	float accy;		/* vertical acceleration (gravity) */

	float *vx,  *vy;	/* current ball velocities */
	float *px,  *py;	/* current ball positions */
	float *opx, *opy;	/* previous ball positions */
	float *r;		/* ball radiuses */

	float *m;		/* ball mass, precalculated */
	float e;		/* coeficient of elasticity */
	float max_radius;	/* largest radius of any ball */

	Bool random_sizes_p;  /* Whether balls should be various sizes up to max. */
	Bool shake_p;		/* Whether to mess with gravity when things settle. */
	Bool dbuf;            /* Whether we're using double buffering. */
	Bool dbeclear_p;      /* ? */
	float shake_threshold;
	int time_since_shake;

	int frame_count;
	int collision_count;

};


/* Re-pick the colors of the balls, and the mouse-ball.
 */
static void
recolor (b_state *state)
{
	int red, green, blue;

	if (state->fg)
		freeimage(state->fg);
	if (state->fg2)
		freeimage(state->fg2);

	red    = 0x88 + (random() % 0x8888);
	green  = 0x88 + (random() % 0x8888);
	blue   = 0x88 + (random() % 0x8888);
	state->fg = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
		red << 24 | green << 16 | blue << 8 | 0xFF);

	red    = 0x88 + (random() % 0x8888);
	green  = 0x88 + (random() % 0x8888);
	blue   = 0x88 + (random() % 0x8888);
	state->fg2 = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
		red << 24 | green << 16 | blue << 8 | 0xFF);
}

/* Initialize the state structure and various X data.
 */
static b_state *
init_balls (void)
{
	int i;
	float extx, exty;
	b_state *state = (b_state *) calloc (1, sizeof(*state));

	recolor (state);

	extx = Dx(screen->r);
	exty = Dy(screen->r);
	state->xmin = 0;
	state->ymin = 0;
	state->xmax = Dx(screen->r);
	state->ymax = Dy(screen->r);

	state->count = nrand(500) + 100;
	state->collision_count = 0;

	state->max_radius = 10.0;

	state->random_sizes_p = 1;

	state->accx = frand()-0.5;

	state->accy = 0;//frand()- 0.5;

	state->frame_count = 0;

	state->shake_p = 1;
	state->shake_threshold = 100;
	state->e = frand()*0.3 + 0.7;
	state->tc = 1.0;

	state->m   = (float *) malloc (sizeof (*state->m)   * (state->count + 1));
	state->r   = (float *) malloc (sizeof (*state->r)   * (state->count + 1));
	state->vx  = (float *) malloc (sizeof (*state->vx)  * (state->count + 1));
	state->vy  = (float *) malloc (sizeof (*state->vy)  * (state->count + 1));
	state->px  = (float *) malloc (sizeof (*state->px)  * (state->count + 1));
	state->py  = (float *) malloc (sizeof (*state->py)  * (state->count + 1));
	state->opx = (float *) malloc (sizeof (*state->opx) * (state->count + 1));
	state->opy = (float *) malloc (sizeof (*state->opy) * (state->count + 1));

	for (i=1; i<=state->count; i++)
	{
		state->px[i] = frand()*extx + state->xmin;
		state->py[i] = frand()*exty + state->ymin;
		state->opx[i] = state->px[i];
		state->opy[i] = state->py[i];
		state->vx[i] = frand()*0.2 - 0.1;
		state->vy[i] = frand()*0.2 - 0.1;

		state->r[i] = (state->random_sizes_p
			? ((0.2 + frand()*0.8) * state->max_radius)
			: state->max_radius);
		state->m[i] = pow(state->r[i],3) * M_PI * 1.3333;
	}

	return state;
}


/* Messes with gravity: permute "down" to be in a random direction.
 */
static void
shake (b_state *state)
{
	float a = state->accx;
	float b = state->accy;
	int i = random() % 4;

	switch (i)
	{
	case 0:
		state->accx = a;
		state->accy = b;
		break;
	case 1:
		state->accx = -a;
		state->accy = -b;
		break;
	case 2:
		state->accx = b;
		state->accy = a;
		break;
	case 3:
		state->accx = -b;
		state->accy = -a;
		break;
	default:
		abort();
		break;
	}

	state->time_since_shake = 0;
	state->frame_count = 0;
	recolor (state);
}


/* Erases the balls at their previous positions, and draws the new ones.
 */
static void
repaint_balls (b_state *state)
{
	int a;
	int x1b, y1b;
	float max_d = 0;

	state->frame_count++;

	for (a=1; a <= state->count; a++)
	{
		x1b = state->px[a] - state->xmin;
		y1b = state->py[a] - state->ymin;

		fillellipse(screen, addpt(screen->r.min, Pt(x1b, y1b)), state->r[a], state->r[a], state->fg, ZP);

		if (state->shake_p)
		{
			/* distance this ball moved this frame */
			float d = ((state->px[a] - state->opx[a]) *
				(state->px[a] - state->opx[a]) +
				(state->py[a] - state->opy[a]) *
				(state->py[a] - state->opy[a]));
			if (d > max_d) max_d = d;
		}
		if (state->shake_p && state->frame_count > 500)
		{
			max_d /= state->max_radius;
			if (max_d < state->shake_threshold ||  /* when its stable */
			state->frame_count > 1000)      /* or when 30 secs has passed */
			{
				shake (state);
			}
		}


		state->opx[a] = state->px[a];
		state->opy[a] = state->py[a];
	}

}


/* Implements the laws of physics: move balls to their new positions.
 */
static void
update_balls (b_state *state)
{
	int a, b;
	float d, vxa, vya, vxb, vyb, dd, cdx, cdy;
	float ma, mb, vca, vcb, dva, dvb;
	float dee2;

	/* For each ball, compute the influence of every other ball. */
	for (a=1; a <= state->count -  1; a++)
		for (b=a + 1; b <= state->count; b++)
		{
			d = ((state->px[a] - state->px[b]) *
				(state->px[a] - state->px[b]) +
				(state->py[a] - state->py[b]) *
				(state->py[a] - state->py[b]));
			dee2 = (state->r[a] + state->r[b]) *
				(state->r[a] + state->r[b]);
			if (d < dee2)
			{
				state->collision_count++;
				d = sqrt(d);
				dd = state->r[a] + state->r[b] - d;

				cdx = (state->px[b] - state->px[a]) / d;
				cdy = (state->py[b] - state->py[a]) / d;

				/* Move each ball apart from the other by half the
             	 * 'collision' distance.
             	 */
				state->px[a] -= 0.5 * dd * cdx;
				state->py[a] -= 0.5 * dd * cdy;
				state->px[b] += 0.5 * dd * cdx;
				state->py[b] += 0.5 * dd * cdy;

				ma = state->m[a];
				mb = state->m[b];

				vxa = state->vx[a];
				vya = state->vy[a];
				vxb = state->vx[b];
				vyb = state->vy[b];

				vca = vxa * cdx + vya * cdy; /* the component of each velocity */
				vcb = vxb * cdx + vyb * cdy; /* along the axis of the collision */

				/* elastic collison */
				dva = (vca * (ma - mb) + vcb * 2 * mb) / (ma + mb) - vca;
				dvb = (vcb * (mb - ma) + vca * 2 * ma) / (ma + mb) - vcb;

				dva *= state->e; /* some energy lost to inelasticity */
				dvb *= state->e;


				vxa += dva * cdx;
				vya += dva * cdy;
				vxb += dvb * cdx;
				vyb += dvb * cdy;

				state->vx[a] = vxa;
				state->vy[a] = vya;
				state->vx[b] = vxb;
				state->vy[b] = vyb;
			}
		}

	/* Force all balls to be on screen.
    	*/
	for (a=1; a <= state->count; a++)
	{
		if (state->px[a] <= (state->xmin + state->r[a]))
		{
			state->px[a] = state->xmin + state->r[a];
			state->vx[a] = -state->vx[a] * state->e;
		}
		if (state->px[a] >= (state->xmax - state->r[a]))
		{
			state->px[a] = state->xmax - state->r[a];
			state->vx[a] = -state->vx[a] * state->e;
		}
		if (state->py[a] <= (state->ymin + state->r[a]))
		{
			state->py[a] = state->ymin + state->r[a];
			state->vy[a] = -state->vy[a] * state->e;
		}
		if (state->py[a] >= (state->ymax - state->r[a]))
		{
			state->py[a] = state->ymax - state->r[a];
			state->vy[a] = -state->vy[a] * state->e;
		}
	}

	/* Apply gravity to all balls.
   	*/
	for (a=1; a <= state->count; a++)
		if (a != state->mouse_ball)
		{
			state->vx[a] += state->accx * state->tc;
			state->vy[a] += state->accy * state->tc;
			state->px[a] += state->vx[a] * state->tc;
			state->py[a] += state->vy[a] * state->tc;
		}
}



void
screenhack (void)
{
	b_state *state = init_balls();

	while (1)
	{
		draw(screen, screen->r, display->black, nil, ZP);
		repaint_balls(state);
		update_balls(state);
		frames++;
		flushimage(display, 1);
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
	}
}


void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	fprint(2, "resize not supported...\n");
	threadexitsall("resize not supported");

}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "fluidballs") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);

	screenhack();
}
