/* halftone, Copyright (c) 2002 by Peter Jaric <peter@jaric.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 *
 * Description:
 * Draws the gravitational force in each point on the screen seen
 * through a halftone dot pattern. The force is calculated from a set
 * of moving mass points. View it from a distance for best effect.
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/03
 */

/* plan9-related stuff */
#include "plan9.h"

#include "smooth.h"

#define DEFAULT_DELAY          10
#define DEFAULT_SPACING        14
#define DEFAULT_SIZE_FACTOR    10
#define DEFAULT_COUNT          10
#define DEFAULT_MIN_MASS       0.001
#define DEFAULT_MAX_MASS       0.02
#define DEFAULT_MIN_SPEED      0.001
#define DEFAULT_MAX_SPEED      0.02

typedef struct
{
	/* halftone dots */
	double * dots;
	int dots_width;
	int dots_height;
	int spacing;
	int max_dot_size;

	/* Moving gravity points */
	int gravity_point_count;

	double* gravity_point_x;
	double* gravity_point_y;
	double* gravity_point_mass;
	double* gravity_point_x_inc;
	double* gravity_point_y_inc;

	int ncolors;
	int color0, color1;
	int color_tick, cycle_speed;

} halftone_screen;

static void update_dot_attributes(halftone_screen *halftone)
{
	double dots_width = Dx(screen->r) / halftone->spacing + 1;
	double dots_height = Dy(screen->r) / halftone->spacing + 1;

	if (halftone->dots == NULL ||
		(dots_width != halftone->dots_width ||
		dots_height != halftone->dots_height))
	{
		if (halftone->dots != NULL)
			free(halftone->dots);

		halftone->dots_width = dots_width;
		halftone->dots_height = dots_height;
		halftone->dots = (double *) malloc(halftone->dots_width * halftone->dots_height * sizeof(double));
	}
}

static halftone_screen * init_halftone(void)
{
	int x, y, i;
	double min_mass;
	double max_mass;
	double min_speed;
	double max_speed;
	halftone_screen *halftone;

	halftone = (halftone_screen *) calloc (1, sizeof(halftone_screen));

	halftone->dots = NULL;

	halftone->gravity_point_count = DEFAULT_COUNT;
	halftone->spacing = DEFAULT_SPACING;
	halftone->max_dot_size = DEFAULT_SIZE_FACTOR;
	min_mass = DEFAULT_MIN_MASS;
	max_mass = DEFAULT_MAX_MASS;
	min_speed = DEFAULT_MIN_SPEED;
	max_speed = DEFAULT_MAX_SPEED;


	/* Set up the moving gravity points. */
	halftone->gravity_point_x = (double *) malloc(halftone->gravity_point_count * sizeof(double));
	halftone->gravity_point_y = (double *) malloc(halftone->gravity_point_count * sizeof(double));
	halftone->gravity_point_mass = (double *) malloc(halftone->gravity_point_count * sizeof(double));
	halftone->gravity_point_x_inc = (double *) malloc(halftone->gravity_point_count * sizeof(double));
	halftone->gravity_point_y_inc = (double *) malloc(halftone->gravity_point_count * sizeof(double));

	for (i = 0; i < halftone->gravity_point_count; i++)
	{
		halftone->gravity_point_x[i] = frand();
		halftone->gravity_point_y[i] = frand();
		halftone->gravity_point_mass[i] = min_mass + (max_mass - min_mass) * frand();
		halftone->gravity_point_x_inc[i] = min_speed + (max_speed - min_speed) * frand();
		halftone->gravity_point_y_inc[i] = min_speed + (max_speed - min_speed) * frand();
	}



	halftone->ncolors = 256;


	halftone->color0 = 0;
	halftone->color1 = halftone->ncolors / 2;
	halftone->cycle_speed = nrand(20) + 10;
	halftone->color_tick = 0;

	update_dot_attributes(halftone);

	for (x = 0; x < halftone->dots_width; x++)
		for (y = 0; y < halftone->dots_height; y++)
		{
			halftone->dots[x + y * halftone->dots_width] = 0;
		}

	return halftone;
}



static void fill_circle(int x, int y, int size, int color)
{
	int start_x = x - (size / 2);
	int start_y = y - (size / 2);

	//XFillArc (display, window, gc,
	//    start_x, start_y, width, height,
	//    angle1, angle2);

	fillellipse(img, Pt(start_x, start_y), size+1, size+1, colors[color], ZP);
}

static void repaint_halftone(halftone_screen *halftone)
{
	int x, y;
	int x_offset = 0;
	int y_offset = 0;


	/* Fill buffer with background color */
	//XSetForeground (halftone->display, halftone->buffer_gc,
	//   halftone->colors[halftone->color0].pixel);
	//XFillRectangle(halftone->display, halftone->buffer, halftone->buffer_gc, 0, 0, halftone->buffer_width, halftone->buffer_height);

	draw(img, img->r, colors[halftone->color0], nil, ZP);

	/* Draw dots on buffer */
	//XSetForeground (halftone->display, halftone->buffer_gc,
	//    halftone->colors[halftone->color1].pixel);

	if (halftone->color_tick++ >= halftone->cycle_speed)
	{
		halftone->color_tick = 0;
		halftone->color0 = (halftone->color0 + 1) % halftone->ncolors;
		halftone->color1 = (halftone->color1 + 1) % halftone->ncolors;
	}

	for (x = 0; x < halftone->dots_width; x++)
		for (y = 0; y < halftone->dots_height; y++)
			fill_circle(x_offset + x * halftone->spacing, y_offset + y * halftone->spacing, 
				(int)((double)halftone->max_dot_size * halftone->dots[x + y * halftone->dots_width]), 
				halftone->color1);

	/* Copy buffer to window */
	//XCopyArea(halftone->display, halftone->buffer, halftone->window, halftone->gc, 0, 0, halftone->buffer_width, halftone->buffer_height, 0, 0);

	draw(screen, screen->r, img, nil, ZP);
	flushimage(display, 1);
}

static double calculate_gravity(halftone_screen *halftone, int x, int y)
{
	int i;
	double gx = 0;
	double gy = 0;

	for (i = 0; i < halftone->gravity_point_count; i++)
	{
		double dx = ((double) x) - halftone->gravity_point_x[i] * halftone->dots_width;
		double dy = ((double) y) - halftone->gravity_point_y[i] * halftone->dots_height;
		double distance = sqrt(dx * dx + dy * dy);

		if (distance != 0)
		{
			double gravity = halftone->gravity_point_mass[i] / (distance * distance  / (halftone->dots_width * halftone->dots_height));

			gx += (dx / distance) * gravity;
			gy += (dy / distance) * gravity;
		}
	}

	return sqrt(gx * gx + gy * gy);
}

static void update_halftone(halftone_screen *halftone)
{
	int x, y, i;

	/* Make sure all dot attributes (spacing, width, height, etc) are correct */
	update_dot_attributes(halftone);

	/* Move gravity points */
	for (i = 0; i < halftone->gravity_point_count; i++)
	{
		halftone->gravity_point_x_inc[i] = 
			(halftone->gravity_point_x[i] >= 1 || halftone->gravity_point_x[i] <= 0 ?
			-halftone->gravity_point_x_inc[i] : 
			halftone->gravity_point_x_inc[i]);
		halftone->gravity_point_y_inc[i] = 
			(halftone->gravity_point_y[i] >= 1 || halftone->gravity_point_y[i] <= 0 ?
			-halftone->gravity_point_y_inc[i] : 
			halftone->gravity_point_y_inc[i]);

		halftone->gravity_point_x[i] += halftone->gravity_point_x_inc[i];
		halftone->gravity_point_y[i] += halftone->gravity_point_y_inc[i];
	}

	/* Update gravity in each dot .*/
	for (x = 0; x < halftone->dots_width; x++)
		for (y = 0; y < halftone->dots_height; y++)
		{
			double gravity = calculate_gravity(halftone, x, y);

			halftone->dots[x + y * halftone->dots_width] = (gravity > 1 ? 1 : (gravity < 0 ? 0 : gravity));
		}
}


void screenhack (void)
{
	halftone_screen *halftone = init_halftone();
	int delay =  DEFAULT_DELAY;

	while (1)
	{
		repaint_halftone(halftone);
		update_halftone(halftone);
		frames++;
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}

		if (delay != 0)
			sleep (delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}


	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");
	smooth(colors);

}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "halftone") < 0)
		sysfatal("initdraw failed: %r");



	einit(Emouse);
	eresized(0);
	screenhack();
}
