/* hop --- real plane fractals */

/*-
 * Copyright (c) 1991 by Patrick J. Naughton.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * Changes in xlockmore distribution
 * 01-Nov-2000: Allocation checks
 * 24-Jun-1997: EJK and RR functions stolen from xmartin2.2
 *              Ed Kubaitis <ejk@ux2.cso.uiuc.edu> ejk functions and xmartin
 *              Renaldo Recuerdo rr function, generalized exponent version
 *              of the Barry Martin's square root function
 * 10-May-1997: Compatible with xscreensaver
 * 27-Jul-1995: added Peter de Jong's hop from Scientific American
 *              July 87 p. 111.  Sometimes they are amazing but there are a
 *              few duds (I did not see a pattern in the parameters).
 * 29-Mar-1995: changed name from hopalong to hop
 * 09-Dec-1994: added Barry Martin's sine hop
 * Changes in original xlock
 * 29-Oct-1990: fix bad (int) cast.
 * 29-Jul-1990: support for multiple screens.
 * 08-Jul-1990: new timing and colors and new algorithm for fractals.
 * 15-Dec-1989: Fix for proper skipping of {White,Black}Pixel() in colors.
 * 08-Oct-1989: Fixed long standing typo bug in RandomInitHop();
 *	            Fixed bug in memory allocation in init_hop();
 *	            Moved seconds() to an extern.
 *	            Got rid of the % mod since .mod is slow on a sparc.
 * 20-Sep-1989: Lint.
 * 31-Aug-1988: Forked from xlock.c for modularity.
 * 23-Mar-1988: Coded HOPALONG routines from Scientific American Sept. 86 p. 14.
 *              Hopalong was attributed to Barry Martin of Aston University
 *              (Birmingham, England)
 */


/*
  * ported to Plan 9 by andrey@lanl.gov, 07/02
  */

/* plan9-related stuff */
#define COUNT 1000
#define CYCLES 500
#define NCOLORS 256

#include "plan9.h"



#define DEF_MARTIN "False"
#define DEF_POPCORN "False"
#define DEF_EJK1 "False"
#define DEF_EJK2 "False"
#define DEF_EJK3 "False"
#define DEF_EJK4 "False"
#define DEF_EJK5 "False"
#define DEF_EJK6 "False"
#define DEF_RR "False"
#define DEF_JONG "False"
#define DEF_SINE "False"

static Bool martin;
static Bool popcorn;
static Bool ejk1;
static Bool ejk2;
static Bool ejk3;
static Bool ejk4;
static Bool ejk5;
static Bool ejk6;
static Bool rr;
static Bool jong;
static Bool sine;

#define MARTIN 0
#define POPCORN 7
#define SINE 8
#define EJK1 1
#define EJK2 2
#define EJK3 9
#define EJK4 3
#define EJK5 4
#define EJK6 10
#define RR 5
#define JONG 6
#ifdef OFFENDING
#define OPS 8			/* 8, 9, 10 might be too close to a swastika for some... */
#else
#define OPS 11
#endif

typedef struct {
	int         centerx, centery;	/* center of the screen */
	double      a, b, c, d;
	double      i, j;	/* hopalong parameters */
	int         inc;
	int         pix;
	int         op;
	int         count;
	int         bufsize;
	XPoint     *pointBuffer;	/* pointer for XDrawPoints */
} hopstruct;

static hopstruct *hops = (hopstruct *) NULL;

void
init_hop(void)
{
	double      range;
	hopstruct  *hp;

	if (hops == NULL) {
		if ((hops = (hopstruct *) calloc(1,
			sizeof (hopstruct))) == NULL)
			return;
	}
	hp = &hops[0];

	hp->centerx = Dx(screen->r) / 2;
	hp->centery = Dy(screen->r) / 2;
	/* Make the other operations less common since they are less interesting */
	hp->op = NRAND(OPS);


	range = sqrt((double) hp->centerx * hp->centerx +
		(double) hp->centery * hp->centery) / (1.0 + LRAND() / MAXRAND);
	hp->i = hp->j = 0.0;
	hp->inc = (int) ((frand()) * 200) - 100;
#undef XMARTIN
	switch (hp->op) {
	case MARTIN:
#ifdef XMARTIN
		hp->a = (frand()) * 1500.0 + 40.0;
		hp->b = (frand()) * 17.0 + 3.0;
		hp->c = (frand()) * 3000.0 + 100.0;
#else
		hp->a = ((frand()) * 2.0 - 1.0) * range / 20.0;
		hp->b = ((frand()) * 2.0 - 1.0) * range / 20.0;
		if (LRAND() & 1)
			hp->c = ((frand()) * 2.0 - 1.0) * range / 20.0;
		else
			hp->c = 0.0;
#endif
		break;
	case EJK1:
#ifdef XMARTIN
		hp->a = (frand()) * 500.0;
		hp->c = (frand()) * 100.0 + 10.0;
#else
		hp->a = ((frand()) * 2.0 - 1.0) * range / 30.0;
		hp->c = ((frand()) * 2.0 - 1.0) * range / 40.0;
#endif
		hp->b = (frand()) * 0.4;
		break;
	case EJK2:
#ifdef XMARTIN
		hp->a = (frand()) * 500.0;
#else
		hp->a = ((frand()) * 2.0 - 1.0) * range / 30.0;
#endif
		hp->b = pow(10.0, 6.0 + (frand()) * 24.0);
		if (LRAND() & 1)
			hp->b = -hp->b;
		hp->c = pow(10.0, (frand()) * 9.0);
		if (LRAND() & 1)
			hp->c = -hp->c;
		break;
	case EJK3:
#ifdef XMARTIN
		hp->a = (frand()) * 500.0;
		hp->c = (frand()) * 80.0 + 30.0;
#else
		hp->a = ((frand()) * 2.0 - 1.0) * range / 30.0;
		hp->c = ((frand()) * 2.0 - 1.0) * range / 70.0;
#endif
		hp->b = (frand()) * 0.35 + 0.5;
		break;
	case EJK4:
#ifdef XMARTIN
		hp->a = (frand()) * 1000.0;
		hp->c = (frand()) * 40.0 + 30.0;
#else
		hp->a = ((frand()) * 2.0 - 1.0) * range / 2.0;
		hp->c = ((frand()) * 2.0 - 1.0) * range / 200.0;
#endif
		hp->b = (frand()) * 9.0 + 1.0;
		break;
	case EJK5:
#ifdef XMARTIN
		hp->a = (frand()) * 600.0;
		hp->c = (frand()) * 90.0 + 20.0;
#else
		hp->a = ((frand()) * 2.0 - 1.0) * range / 2.0;
		hp->c = ((frand()) * 2.0 - 1.0) * range / 200.0;
#endif
		hp->b = (frand()) * 0.3 + 0.1;
		break;
	case EJK6:
#ifdef XMARTIN
		hp->a = (frand()) * 100.0 + 550.0;
#else
		hp->a = ((frand()) * 2.0 - 1.0) * range / 30.0;
#endif
		hp->b = (frand()) + 0.5;
		break;
	case RR:
#ifdef XMARTIN
		hp->a = (frand()) * 100.0;
		hp->b = (frand()) * 20.0;
		hp->c = (frand()) * 200.0;
#else
		hp->a = ((frand()) * 2.0 - 1.0) * range / 40.0;
		hp->b = ((frand()) * 2.0 - 1.0) * range / 200.0;
		hp->c = ((frand()) * 2.0 - 1.0) * range / 20.0;
#endif
		hp->d = (frand()) * 0.9;
		break;
	case POPCORN:
		hp->a = 0.0;
		hp->b = 0.0;
		hp->c = ((frand()) * 2.0 - 1.0) * 0.24 + 0.25;
		hp->inc = 100;
		break;
	case JONG:
		hp->a = ((frand()) * 2.0 - 1.0) * M_PI;
		hp->b = ((frand()) * 2.0 - 1.0) * M_PI;
		hp->c = ((frand()) * 2.0 - 1.0) * M_PI;
		hp->d = ((frand()) * 2.0 - 1.0) * M_PI;
		break;
	case SINE:	/* MARTIN2 */
#ifdef XMARTIN
		hp->a = M_PI + ((frand()) * 2.0 - 1.0) * 0.07;
#else
		hp->a = M_PI + ((frand()) * 2.0 - 1.0) * 0.7;
#endif
		break;
	}
	hp->pix = NRAND(256);
	hp->bufsize = COUNT;

	if (hp->pointBuffer == NULL) {
		if ((hp->pointBuffer = (XPoint *) malloc(hp->bufsize *
			sizeof (XPoint))) == NULL)
			return;
	}

	draw(screen, screen->r, display->black, nil, ZP);
	hp->count = 0;
}


void
draw_hop(void)
{
	double      oldj, oldi;
	XPoint     *xp;
	int         k, i2;
	hopstruct  *hp;

	if (hops == NULL)
		return;
	hp = &hops[0];
	if (hp->pointBuffer == NULL)
		return;
	xp = hp->pointBuffer;
	k = hp->bufsize;

	hp->inc++;
	if (++hp->pix >= NCOLORS)
		hp->pix = 1;

	while (k--) {
		oldj = hp->j;
		switch (hp->op) {
		case MARTIN:	/* SQRT, MARTIN1 */
			oldi = hp->i + hp->inc;
			hp->j = hp->a - hp->i;
			hp->i = oldj + ((hp->i < 0)
				? sqrt(fabs(hp->b * oldi - hp->c))
				: -sqrt(fabs(hp->b * oldi - hp->c)));
			xp->x = hp->centerx + (int) (hp->i + hp->j);
			xp->y = hp->centery - (int) (hp->i - hp->j);
			break;
		case EJK1:
			oldi = hp->i + hp->inc;
			hp->j = hp->a - hp->i;
			hp->i = oldj - ((hp->i > 0) ? (hp->b * oldi - hp->c) :
				-(hp->b * oldi - hp->c));
			xp->x = hp->centerx + (int) (hp->i + hp->j);
			xp->y = hp->centery - (int) (hp->i - hp->j);
			break;
		case EJK2:
			oldi = hp->i + hp->inc;
			hp->j = hp->a - hp->i;
			hp->i = oldj - ((hp->i < 0) ? log(fabs(hp->b * oldi - hp->c)) :
				-log(fabs(hp->b * oldi - hp->c)));
			xp->x = hp->centerx + (int) (hp->i + hp->j);
			xp->y = hp->centery - (int) (hp->i - hp->j);
			break;
		case EJK3:
			oldi = hp->i + hp->inc;
			hp->j = hp->a - hp->i;
			hp->i = oldj - ((hp->i > 0) ? sin(hp->b * oldi) - hp->c :
				-sin(hp->b * oldi) - hp->c);
			xp->x = hp->centerx + (int) (hp->i + hp->j);
			xp->y = hp->centery - (int) (hp->i - hp->j);
			break;
		case EJK4:
			oldi = hp->i + hp->inc;
			hp->j = hp->a - hp->i;
			hp->i = oldj - ((hp->i > 0) ? sin(hp->b * oldi) - hp->c :
				-sqrt(fabs(hp->b * oldi - hp->c)));
			xp->x = hp->centerx + (int) (hp->i + hp->j);
			xp->y = hp->centery - (int) (hp->i - hp->j);
			break;
		case EJK5:
			oldi = hp->i + hp->inc;
			hp->j = hp->a - hp->i;
			hp->i = oldj - ((hp->i > 0) ? sin(hp->b * oldi) - hp->c :
				-(hp->b * oldi - hp->c));
			xp->x = hp->centerx + (int) (hp->i + hp->j);
			xp->y = hp->centery - (int) (hp->i - hp->j);
			break;
		case EJK6:
			oldi = hp->i + hp->inc;
			hp->j = hp->a - hp->i;
			hp->i = oldj - asin((hp->b * oldi) - (long) (hp->b * oldi));
			xp->x = hp->centerx + (int) (hp->i + hp->j);
			xp->y = hp->centery - (int) (hp->i - hp->j);
			break;
		case RR:	/* RR1 */
			oldi = hp->i + hp->inc;
			hp->j = hp->a - hp->i;
			hp->i = oldj - ((hp->i < 0) ? -pow(fabs(hp->b * oldi - hp->c), hp->d) :
				pow(fabs(hp->b * oldi - hp->c), hp->d));
			xp->x = hp->centerx + (int) (hp->i + hp->j);
			xp->y = hp->centery - (int) (hp->i - hp->j);
			break;
		case POPCORN:
#define HVAL 0.05
#define INCVAL 50
			{
				double      tempi, tempj;

				if (hp->inc >= 100)
					hp->inc = 0;
				if (hp->inc == 0) {
					if (hp->a++ >= INCVAL) {
						hp->a = 0;
						if (hp->b++ >= INCVAL)
							hp->b = 0;
					}
					hp->i = (-hp->c * INCVAL / 2 + hp->c * hp->a) * M_PI / 180.0;
					hp->j = (-hp->c * INCVAL / 2 + hp->c * hp->b) * M_PI / 180.0;
				}
				tempi = hp->i - HVAL * sin(hp->j + tan(3.0 * hp->j));
				tempj = hp->j - HVAL * sin(hp->i + tan(3.0 * hp->i));
				xp->x = hp->centerx + (int) (Dx(screen->r) / 40 * tempi);
				xp->y = hp->centery + (int) (Dy(screen->r) / 40 * tempj);
				hp->i = tempi;
				hp->j = tempj;
			}
			break;
		case JONG:
			if (hp->centerx > 0)
				oldi = hp->i + 4 * hp->inc / hp->centerx;
			else
				oldi = hp->i;
			hp->j = sin(hp->c * hp->i) - cos(hp->d * hp->j);
			hp->i = sin(hp->a * oldj) - cos(hp->b * oldi);
			xp->x = hp->centerx + (int) (hp->centerx * (hp->i + hp->j) / 4.0);
			xp->y = hp->centery - (int) (hp->centery * (hp->i - hp->j) / 4.0);
			break;
		case SINE:	/* MARTIN2 */
			oldi = hp->i + hp->inc;
			hp->j = hp->a - hp->i;
			hp->i = oldj - sin(oldi);
			xp->x = hp->centerx + (int) (hp->i + hp->j);
			xp->y = hp->centery - (int) (hp->i - hp->j);
			break;
		}
		xp++;
	}
	//	XDrawPoints(MI_DISPLAY(mi), MI_WINDOW(mi), MI_GC(mi),
	//		    hp->pointBuffer, hp->bufsize, CoordModeOrigin);

	for(i2 = 0; i2 < hp->bufsize; i2++) {
		Rectangle r;
		Point p = Pt(screen->r.min.x, screen->r.min.y);

		p = addpt(p, hp->pointBuffer[i2]);
		r = Rpt(p, addpt(p, Pt(1,1)));
		draw(screen, r, colors[hp->pix], nil, ZP);

	}
	if (++hp->count > CYCLES) {
		draw(screen, screen->r, display->black, nil, ZP);
		init_hop();
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	init_hop();
}

void 
threadmain(int argc, char **argv)
{
	Mouse m;
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "hopalong") < 0)
		sysfatal("initdraw failed: %r");


	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(i<<2));
	}
	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_hop();
		frames++;
	}
}
