/*
 *  InterAggregate (dagraz@gmail.com)
 *  Based on code from complexification.net Intersection Aggregate
 *  http://www.complexification.net/gallery/machines/interAggregate/index.php
 *
 *  Intersection Aggregate code:
 *  j.tarbell   May, 2004
 *  Albuquerque, New Mexico
 *  complexification.net
 *
 *  Also based on substrate, a port of j.tarbell's Substrate Art done
 *  by dragorn@kismetwireless.net
 *
 *  
 *  Interesting command line options, all of which serve to
 *  concentrate the painting in some way:
 *  
 *  -percent-orbits 100 -base-orbits 50 -base-on-center -growth-delay 0
 *
 *  Paint should be concentrated in the center of the canvas, orbiting
 *  about it.  -percent-orbits 100 implies -base-on-center, so that's
 *  not really needed.
 *
 *
 *  -percent-orbits 99 -base-orbits 50 -growth-delay 0
 *
 *  Like the above example, but the 'center' will rove about the screen.
 *
 *  -percent-orbits 98 -base-orbits 50 -growth-delay 0
 *
 *  Like the above example, but there will be two roving centers.
 *
 *
 *  TODO:
 *  -fix alpha blending / byte ordering
 *
 *  CHANGES
 *
 *
 * Directly based the hacks of: 
 * 
 * xscreensaver, Copyright (c) 1997, 1998, 2002 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/06
 */

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"


static const char *interaggregate_defaults[] = 
{
	".background: white",
	".foreground: black",
	"*maxCycles: 100000",
	"*growthDelay: 18000", 
	"*numCircles: 100",
	"*percentOrbits: 0",
	"*baseOrbits: 75",
	"*baseOnCenter: False",
	"*drawCenters: False",
	0
};

/* black white brown olive grey camel */

typedef enum { 
	LINEAR, ORBIT } PathType;

typedef struct 
{

	unsigned long color;
	double gain;
	double p;

} SandPainter;

typedef struct _circle
{
	double radius;

	double x;
	double y;

	PathType path_type;

	/* for a linear path */
	double dx;
	double dy;

	/* for orbital path */
	double theta;
	double r;
	double dtheta;

	struct _circle* center;

	int num_painters;
	SandPainter* painters;

} Circle;


struct field 
{
	int height;
	int width;

	int num_circles;
	Circle* circles;

	int percent_orbits;
	int base_orbits;
	Bool base_on_center;

	/* used for orbits circling the center of the screen */
	Circle center_of_universe;

	/* Raw map of pixels we need to keep for alpha blending */
	unsigned long int *off_img;

	/* color parms */
	int numcolors;
	unsigned long *parsedcolors;
	unsigned long fgcolor;
	unsigned long bgcolor;
	int visdepth;

	unsigned int cycles;

	double max_gain;

	/* for debugging */
	Bool draw_centers;

	/* for profiling whatnot */
	int possible_intersections;
	int intersection_count;
};

struct state {
	unsigned int max_cycles;
	int growth_delay;
	struct field *f;
};

struct state st;

static struct field *
init_field(void)
{
	struct field *f = (struct field*) malloc(sizeof(struct field));
	if ( f == NULL )
		sysfatal("malloc\n");

	f->height = 0;
	f->width = 0;
	f->num_circles = 0;
	f->circles = NULL;
	f->percent_orbits = 0;
	f->base_orbits = 0;
	f->base_on_center = False;
	f->off_img = NULL;
	f->numcolors = 256;
	f->parsedcolors = NULL;
	f->fgcolor = 0;
	f->bgcolor = 0;
	f->visdepth = 0;

	f->cycles = 0;

	f->max_gain = 0.22;

	f->draw_centers = False;

	f->possible_intersections = 0;
	f->intersection_count = 0;

	return f;
}

/* Quick references to pixels in the offscreen map and in the crack grid */
#define ref_pixel(f, x, y)   ((f)->off_img[(y) * (f)->width + (x)])

#define in_bounds(f, x, y) ((x >= 0) && (x < f->width) && (y >= 0) && (y < f->height))

/* Consider rewriting with XQueryColor, or ImageByteOrder */

static void point2rgb(int depth, unsigned long c, int *r, int *g, int *b)
{
	switch(depth)
	{
	case 32:
	case 24:
		*b = c & 0xff;
		*g = (c & 0xff00) >> 8;
		*r = (c & 0xff0000) >> 16;
		break;
	case 16:
		*b = (c & 0x1f) << 3;
		*g = ((c >> 5) & 0x3f) << 2;
		*r = ((c >> 11) & 0x1f) << 3;
		break;
	case 15:
		*b = (c & 0x1f) << 3;
		*g = ((c >> 5) & 0x1f) << 3;
		*r = ((c >> 10) & 0x1f) << 3;
		break;
	}
}

static unsigned long rgb2point(int depth, int r, int g, int b)
{
	unsigned long ret = 0;

	switch(depth)
	{
	case 32:
		ret = 0xff000000;
	case 24:
		ret |= (r << 16) | (g << 8) | b;
		break;
	case 16:
		ret = ((r>>3) << 11) | ((g>>2)<<5) | (b>>3);
		break;
	case 15:
		ret = ((r>>3) << 10) | ((g>>3)<<5) | (b>>3);
		break;
	}

	return ret;
}

/* alpha blended point drawing -- this is Not Right and will likely fail on 
 * non-intel platforms as it is now, needs fixing */
static unsigned long 
trans_point(int x1, int y1, unsigned long myc, double a, struct field *f)
{
	if (a >= 1.0)
	{
		ref_pixel(f, x1, y1) = myc;
		return myc;
	}
	else 
	{
		int or=0, og=0, ob=0;
		int r=0, g=0, b=0;
		int nr, ng, nb;
		unsigned long c;

		c = ref_pixel(f, x1, y1);

		point2rgb(f->visdepth, c, &or, &og, &ob);
		point2rgb(f->visdepth, myc, &r, &g, &b);

		nr = or + (r - or) * a;
		ng = og + (g - og) * a;
		nb = ob + (b - ob) * a;

		c = rgb2point(f->visdepth, nr, ng, nb);

		ref_pixel(f, x1, y1) = c;

		return c;
	}
}

static void drawPoint(int x, int y, unsigned long color, double intensity, struct field *f)

{

	USED(intensity);

	while ( x >= f->width ) x -= f->width;
	while ( x < 0 ) x += f->width;

	while ( y >= f->height ) y -= f->height;
	while ( y < 0 ) y += f->height;

	/* if ( in_bounds(f, x, y) ) ... */

	//XSetForeground(dpy, fgc, c);
	//XDrawPoint(dpy, window, fgc, x, y);
	draw(screen, rectaddpt(Rect(x, y, x+1, y+1), screen->r.min), colors[color%255], nil, ZP);
}

static void paint(SandPainter* painter, double ax, double ay, double bx, double by,
struct field *f)
{
	/* the sand painter */

	double inc, sandp;
	int i;

	/* XXX try adding tpoint here, like orig */

	/* jitter the painter's values */
	painter->gain += frand()*0.05 - 0.025;

	if ( painter->gain > f->max_gain )
		painter->gain = -f->max_gain;
	else if ( painter->gain < -f->max_gain )
		painter->gain = f->max_gain;

	painter->p += frand()*0.1 - 0.05;

	if ( 0 < painter->p )
		painter->p = 0;
	else if ( painter->p > 1.0 )
		painter->p = 1.0;

	/* replace 0.1 with 1 / f->grains */
	inc = painter->gain * 0.1;
	sandp = 0;

	for(i = 0; i <= 10; ++i)
	{
		int drawx, drawy;
		double sp, sm;
		double intensity = 0.1 - 0.009 * i;

		sp = sin(painter->p + sandp);
		drawx = ax + (bx - ax) * sp;
		drawy = ay + (by - ay) * sp;

		drawPoint(drawx, drawy, painter->color,
			intensity, f);

		sm = sin(painter->p - sandp);
		drawx = ax + (bx - ax) * sm;
		drawy = ay + (by - ay) * sm;

		drawPoint(drawx, drawy, painter->color,
			intensity, f);

		sandp += inc;
	}
	flushimage(display, 1);
}


/* used when the window is resized */
static void build_img(struct field *f)
{
	if (f->off_img) {
		free(f->off_img);
		f->off_img = NULL;
	}

	f->off_img = (unsigned long *) calloc(f->width * f->height,
		sizeof(unsigned long));


	if ( f->off_img == NULL )
	{
		sysfatal("failed to allocate off_img\n");
	}

	memset(f->off_img, f->bgcolor, 
		sizeof(unsigned long) * f->width * f->height);
}

static void free_circles(struct field *f)
{
	int i;

	if ( f->circles != NULL )
	{
		for(i = 0; i < f->num_circles; ++i)
		{
			free (f->circles[i].painters);
		}

		free (f->circles);
		f->circles = NULL;
	}
}

static void build_field(struct field *f)
{
	int i;
	int num_orbits;
	int base_orbits;
	int orbit_start;
	build_img(f);

	f->center_of_universe.x = f->width / 2.0;
	f->center_of_universe.y = f->height / 2.0;
	f->center_of_universe.r = MAX(f->width, f->height) / 2.0;

	num_orbits = (f->percent_orbits * f->num_circles) / 100;
	orbit_start = f->num_circles - num_orbits;
	base_orbits = orbit_start + (num_orbits * f->base_orbits) / 100;

	free_circles(f);

	f->circles = (Circle*) calloc(f->num_circles, sizeof(Circle));
	if ( f->circles == NULL )
	{
		sysfatal("failed to allocate off_img\n");
	}

	for(i = 0; i < f->num_circles; ++i)
	{
		int j;
		Circle *circle = f->circles + i;

		/* make this a pref */

		if ( i >= orbit_start )
			circle->path_type = ORBIT;
		else
			circle->path_type = LINEAR;


		if ( circle->path_type == LINEAR )
		{
			circle->x = frand()*f->width;
			circle->y = frand()*f->height;

			circle->dx = frand()*0.5 - 0.25;
			circle->dy = frand()*0.5 - 0.25;
			/* circle->dy = 0; */
			/* circle->r  = f->height * (0.05 + frand()*0.1); */
			circle->radius = 5 + frand()*55;

			/* in case we want orbits based on lines */
			circle->r = MIN(f->width, f->height) / 2.0;
			circle->center = NULL;
		}
		else /* == ORBIT */
		{
			if (i < base_orbits )
			{
				if ( f->base_on_center )
					circle->center = &f->center_of_universe;
				else
				{
					circle->center = f->circles + 
						(int)(frand()*(orbit_start - 0.1));
				}

				circle->r = 1 + frand()*(MIN(f->width, f->height) / 2.0);

				/* circle->radius = 5 + frand(55); */
			}
			else
			{
				/* give a preference for the earlier circles */

				double p = frand()*0.9;

				circle->center = f->circles + (int) (p*i);

				circle->r = 1 + 0.5 * circle->center->r + 0.5 * frand()*(circle->center->r);
				/* circle->r = 1 + frand()*(circle->center->r / 2); */


				/* circle->radius = MAX(5, frand()*(circle->r)); */
				/* circle->radius = 5 + frand(55); */
			}

			circle->radius = 5 + frand()*(MIN(55, circle->r));
			circle->dtheta = (frand()*(0.5) - 0.25) / circle->r;
			circle->theta = frand()*(2 * M_PI);

			circle->x = circle->r * cos(circle->theta) + circle->center->x;
			circle->y = circle->r * sin(circle->theta) + circle->center->y;

		}

		/* make this a command line option */
		circle->num_painters = 3;
		circle->painters = (SandPainter*) calloc(circle->num_painters, 
			sizeof(SandPainter));
		if ( circle->painters == NULL )
		{
			sysfatal("failed to allocate painters");
		}

		for(j = 0; j < circle->num_painters; ++j)
		{
			SandPainter *painter = circle->painters + j;

			painter->gain = frand()*(0.09) + 0.01;
			painter->p = frand()*(1.0);
			painter->color = frand()*f->numcolors;
		}
	}
}

static void moveCircles(struct field *f)
{
	int i;

	for(i = 0; i < f->num_circles; ++i)
	{
		Circle *circle = f->circles + i;

		if ( circle->path_type == LINEAR )
		{
			circle->x += circle->dx;
			circle->y += circle->dy;

			if ( circle->x < 0 ) circle->x += f->width;
			else if ( circle->x >= f->width ) circle->x -= f->width;

			if ( circle->y < 0 ) circle->y += f->height;
			else if ( circle->y >= f->height ) circle->y -= f->height;
		}
		else /* (circle->path_type == ORBIT) */
		{
			circle->theta += circle->dtheta;

			if ( circle->theta < 0 ) circle->theta += 2 * M_PI;
			else if ( circle->theta > 2 * M_PI ) circle->theta -= 2 * M_PI;

			circle->x = circle->r * cos(circle->theta) + circle->center->x;
			circle->y = circle->r * sin(circle->theta) + circle->center->y;

			if ( circle->x < 0 ) circle->x += f->width;
			else if ( circle->x >= f->width ) circle->x -= f->width;

			if ( circle->y < 0 ) circle->y += f->height;
			else if ( circle->y >= f->height ) circle->y -= f->height;
		}
	}
}

static void drawIntersections(struct field *f)
{
	int i,j;

	/* One might be tempted to think 'hey, this is a crude algorithm
	 * that is going to check each of the n (n-1) / 2 possible
	 * intersections!  Why not try bsp trees, quad trees, etc, etc,
	 * etc'
	 *
	 * In practice the time spent drawing the intersection of two
	 * circles dwarfs the time takes to check for intersection.
	 * Profiling on a 640x480 screen with 100 circles shows possible
	 * speed gains to be only a couple of percent.
	 * 
	 * But hey, if you're bored, go have fun.  Let me know how it
	 * turns out.
	 */


	for(i = 0; i < f->num_circles; ++i)
	{
		Circle *c1 = f->circles + i;

		if ( !f->draw_centers )
		{
			/* the default branch */

			for(j = i + 1; j < f->num_circles; ++j)
			{
				double d, dsqr, dx, dy;
				Circle *c2 = f->circles + j;

				dx = c2->x - c1->x;
				dy = c2->y - c1->y;

				dsqr = dx * dx + dy * dy;
				d = sqrt(dsqr);

				if ( (fabs(dx) < (c1->radius + c2->radius)) &&
					(fabs(dy) < (c1->radius + c2->radius)) &&
					( d < (c1->radius + c2->radius) ) &&
					( d > fabs(c1->radius - c2->radius) ) )
				{
					double d1, d2, r1sqr;
					double bx, by;
					double midpx, midpy;
					double int1x, int1y;
					double int2x, int2y;
					int s;

					/* woo-hoo.  the circles are neither outside nor
				 * inside each other.  they intersect.  
				 *
				 * Now, compute the coordinates of the points of
				 * intersection 
				 */

					/* unit vector in direction of c1 to c2 */
					bx = dx / d;
					by = dy / d;

					r1sqr = c1->radius * c1->radius;

					/* distance from c1's center midpoint of intersection
				 * points */

					d1 = 0.5 * (r1sqr - c2->radius * c2->radius + dsqr) / d;

					midpx = c1->x + d1 * bx;
					midpy = c1->y + d1 * by;

					/* distance from midpoint to points of intersection */

					d2 = sqrt(r1sqr - d1 * d1);

					int1x = midpx + d2 * by;
					int1y = midpy - d2 * bx;

					int2x = midpx - d2 * by;
					int2y = midpy + d2 * bx;

					for(s = 0; s < c1->num_painters; ++s)
					{
						paint(c1->painters + s, int1x, int1y, int2x, int2y, f);
					}
				}
			}
		}
		else /* f->draw_centers */
		{
			//XDrawPoint(dpy, window, fgc, c1->x, c1->y);
			draw(screen, rectaddpt(Rect(c1->x, c1->y, c1->x+1, c1->y+1), screen->r.min), display->black, nil, ZP);
		}
	}
}

static unsigned long
interaggregate_draw (void *closure)
{
	struct state *st = (struct state *) closure;

	moveCircles(st->f);
	drawIntersections(st->f);

	st->f->cycles++;


	if (st->f->cycles >= st->max_cycles && st->max_cycles != 0)
	{
		build_field(st->f);
		//XSetForeground(st->dpy, st->fgc, st->gcv.background);
		//XFillRectangle(st->dpy, st->window, st->fgc, 0, 0, st->xgwa.width, st->xgwa.height);
		//XSetForeground(st->dpy, st->fgc, st->gcv.foreground);
		draw(screen, screen->r, display->white, nil, ZP);
	}
	flushimage(display, 1);

	return st->growth_delay;
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	smooth(colors);

	if(st.f)
		free(st.f);

	st.f = init_field();
	st.growth_delay = 18000;
	st.max_cycles = 100000;
	st.f->width = Dx(screen->r);
	st.f->height = Dy(screen->r);
	st.f->num_circles = 100;
	st.f->percent_orbits = 0;
	st.f->base_orbits = 75;
	st.f->base_on_center = 0;
	st.f->draw_centers = 0;

	if ( st.f->percent_orbits == 100 )
		st.f->base_on_center = True;

	st.f->height = Dy(screen->r);
	st.f->width = Dx(screen->r);

	/* Initialize stuff */
	build_field(st.f);
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));

	if(initdraw(nil, nil, "interaggregate") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);

	init_field();

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(nil);
		}
		interaggregate_draw(&st);
		frames++;
	}

}
