/* interference.c --- colored fields via decaying sinusoidal waves.
 * An entry for the RHAD Labs Screensaver Contest.
 *
 * Author: Hannu Mallat <hmallat@cs.hut.fi>
 *
 * Copyright (C) 1998 Hannu Mallat.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 * decaying sinusoidal waves, which extend spherically from their
 * respective origins, move around the plane. a sort of interference
 * between them is calculated and the resulting twodimensional wave
 * height map is plotted in a grid, using softly changing colours.
 *
 * not physically (or in any sense) accurate, but fun to look at for 
 * a while. you may tune the speed/resolution/interestingness tradeoff 
 * with X resources, see below.
 *
 * Created      : Wed Apr 22 09:30:30 1998, hmallat
 * Last modified: Wed Apr 22 09:30:30 1998, hmallat
 * Last modified: Sun Aug 31 23:40:14 2003, 
 *              david slimp <rock808@DavidSlimp.com>
 * 		added -hue option to specify base color hue
 *
 * TODO:
 *
 *    This really needs to be sped up.
 *
 *    I've tried making it use XPutPixel/XPutImage instead of XFillRectangle,
 *    but that doesn't seem to help (it's the same speed at gridsize=1, and
 *    it actually makes it slower at larger sizes.)
 *
 *    I played around with shared memory, but clearly I still don't know how
 *    to use the XSHM extension properly, because it doesn't work yet.
 *
 *    Hannu had put in code to use the double-buffering extension, but that
 *    code didn't work for me on Irix.  I don't think it would help anyway,
 *    since it's not the swapping of frames that is the bottleneck (or a source
 *    of visible flicker.)
 *
 *     -- jwz, 4-Jun-98
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/03
 */

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"


struct inter_source {
	int x;
	int y;
	double x_theta;
	double y_theta;
};

struct inter_context {

	int count;
	int grid_size;
	int colors;
	float hue;
	int speed;
	int delay;
	int shift;
	int radius;

	int w;
	int h;
	int pal;

	int* wave_height;

	struct inter_source* source;
};

void inter_init(struct inter_context* c)
{
	int i;

	memset (c, 0, sizeof(*c));

	c->count = nrand(3)+2;
	c->grid_size = nrand(10)+1;
	c->colors = 256;
	c->hue = (double)nrand(360);
	c->speed = nrand(10) + 25;
	c->shift = 60;
	c->radius = 800;


	c->wave_height = calloc(c->radius, sizeof(int));
	for(i = 0; i < c->radius; i++) {
		float max = 
		((float)c->colors) * 
		((float)c->radius - (float)i) /
		((float)c->radius);
		c->wave_height[i] = 
			(int)
			((max + max*cos((double)i/50.0)) / 2.0);
	}

	c->source = calloc(c->count, sizeof(struct inter_source));
	for(i = 0; i < c->count; i++) {
		c->source[i].x_theta = frand()*2*PI;
		c->source[i].y_theta = frand()*2*PI;
	}

}

#define source_x(c, i) \
  (Dx(screen->r)/2 + ((int)(cos(c->source[i].x_theta)*((float)Dx(screen->r)/2.0))))
#define source_y(c, i) \
  (Dy(screen->r)/2 + ((int)(cos(c->source[i].y_theta)*((float)Dy(screen->r)/2.0))))

/*
 * this is rather suboptimal. the sqrt() doesn't seem to be a big
 * performance hit, but all those separate XFillRectangle()'s are.
 * hell, it's just a quick hack anyway -- if someone wishes to tune
 * it, go ahead! 
 */

void do_inter(struct inter_context* c)
{
	int i, j, k;
	uint result;
	int dist;
	int g;

	int dx, dy;

	for(i = 0; i < c->count; i++) {
		c->source[i].x_theta += (c->speed/1000.0);
		if(c->source[i].x_theta > 2.0*3.14159)
			c->source[i].x_theta -= 2.0*3.14159;
		c->source[i].y_theta += (c->speed/1000.0);
		if(c->source[i].y_theta > 2.0*3.14159)
			c->source[i].y_theta -= 2.0*3.14159;
		c->source[i].x = source_x(c, i);
		c->source[i].y = source_y(c, i);
	}

	g = c->grid_size;

	for(j = 0; j < Dy(screen->r)/g; j++) {
		for(i = 0; i < Dx(screen->r)/g; i++) {
			result = 0;
			for(k = 0; k < c->count; k++) {
				dx = i*g + g/2 - c->source[k].x;
				dy = j*g + g/2 - c->source[k].y;
				dist = sqrt(dx*dx + dy*dy); /* what's the performance penalty here? */
				result += (dist > c->radius ? 0 : c->wave_height[dist]);
			}
			result %= c->colors;

			//XFillRectangle(c->dpy, TARGET(c), c->gcs[result], g*i, g*j, g, g);
			draw(img, Rect(g*i, g*j, g*i + g, g*j + g), 
				colors[result], nil, ZP);
		}
	}

	//XSync(c->dpy, False);
	draw(screen, screen->r, img, nil, ZP);
	flushimage(display, 1);
}

void screenhack(void)
{
	struct inter_context c;
	int delay;

	delay = 10;

	inter_init(&c);
	while(1) {
		do_inter(&c);
		frames++;
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		if(delay) sleep(delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}


	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "interference") < 0)
		sysfatal("initdraw failed: %r");


	smooth(colors);

	einit(Emouse);
	eresized(0);

	screenhack();
}
