/*
 * julia --- continuously varying Julia set.
 */

/* Copyright (c) 1995 Sean McCullough <bankshot@mailhost.nmt.edu>.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * 28-May-97: jwz@jwz.org: added interactive frobbing with the mouse.
 * 10-May-97: jwz@jwz.org: turned into a standalone program.
 * 02-Dec-95: snagged boilerplate from hop.c
 *           used ifs {w0 = sqrt(x-c), w1 = -sqrt(x-c)} with random iteration 
 *           to plot the julia set, and sinusoidially varied parameter for set 
 *	     and plotted parameter with a circle.
 */

/*-
 * One thing to note is that batchcount is the *depth* of the search tree,
 * so the number of points computed is 2^batchcount - 1.  I use 8 or 9
 * on a dx266 and it looks okay.  The sinusoidal variation of the parameter
 * might not be as interesting as it could, but it still gives an idea of
 * the effect of the parameter.
 */

/*
  * ported to Plan 9 by andrey@lanl.gov, 06/02
  */

/* plan9-related stuff */
#include "plan9.h"

#include "smooth.h"


#define COUNT 10
#define CYCLES 20


#define numpoints ((0x2<<jp->depth)-1)

typedef struct {
	int         centerx;
	int         centery;	/* center of the screen */
	double      cr;
	double      ci;		/* julia params */
	int         depth;
	int         inc;
	int         circsize;
	int         erase;
	int         pix;
	long        itree;
	int         buffer;
	int         nbuffers;
	int         redrawing, redrawpos;
	XPoint    **pointBuffer;	/* pointer for XDrawPoints */

} juliastruct;

static juliastruct *julias = NULL;

/* How many segments to draw per cycle when redrawing */
#define REDRAWSTEP 3

static void
apply(juliastruct * jp, register double xr, register double xi, int d)
{
	double      theta, r;

	jp->pointBuffer[jp->buffer][jp->itree].x =
		(int) (0.5 * xr * jp->centerx + jp->centerx);
	jp->pointBuffer[jp->buffer][jp->itree].y =
		(int) (0.5 * xi * jp->centery + jp->centery);
	jp->itree++;

	if (d > 0) {
		xi -= jp->ci;
		xr -= jp->cr;

		/* Avoid atan2: DOMAIN error message */
		if (xi == 0.0 && xr == 0.0)
			theta = 0.0;
		else
			theta = atan2(xi, xr) / 2.0;

		/*r = pow(xi * xi + xr * xr, 0.25); */
		r = sqrt(sqrt(xi * xi + xr * xr));	/* 3 times faster */

		xr = r * cos(theta);
		xi = r * sin(theta);

		d--;
		apply(jp, xr, xi, d);
		apply(jp, -xr, -xi, d);
	}
}

static void
incr(juliastruct * jp)
{




	jp->cr = 1.5 * (sin(M_PI * (jp->inc / 300.0)) *
		sin(jp->inc * M_PI / 200.0));
	jp->ci = 1.5 * (cos(M_PI * (jp->inc / 300.0)) *
		cos(jp->inc * M_PI / 200.0));

	jp->cr += 0.5 * cos(M_PI * jp->inc / 400.0);
	jp->ci += 0.5 * sin(M_PI * jp->inc / 400.0);
}

void
init_julia(void)
{
	juliastruct *jp;
	int         i;

	if (julias == NULL) {
		if ((julias = (juliastruct *) calloc(1,
			sizeof (juliastruct))) == NULL)
			return;
	}
	jp = &julias[0];

	jp->centerx = Dx(screen->r) / 2;
	jp->centery = Dy(screen->r) / 2;

	jp->depth = COUNT;
	if (jp->depth > 10)
		jp->depth = 10;


	jp->pix = NRAND(256);
	jp->inc = ((LRAND() & 1) * 2 - 1) * NRAND(2000);
	jp->nbuffers = (CYCLES + 1);
	if (!jp->pointBuffer)
		jp->pointBuffer = (XPoint **) calloc(jp->nbuffers, sizeof (XPoint *));
	for (i = 0; i < jp->nbuffers; ++i)
		if (jp->pointBuffer[i])
			memset((char *) jp->pointBuffer[i], 0,
				numpoints * sizeof (XPoint));
		else
			jp->pointBuffer[i] = (XPoint *) calloc(numpoints, sizeof (XPoint));
	jp->buffer = 0;
	jp->redrawing = 0;
	jp->erase = 0;
}


/* hack: moved here by jwz. */

XPoint old_circle, new_circle;

void
draw_julia(void)
{
	juliastruct *jp = &julias[0];
	double      r, theta;
	register double xr = 0.0, xi = 0.0;
	int         k = 64, rnd = 0, i, i2, j;
	XPoint     *xp = jp->pointBuffer[jp->buffer];

	old_circle.x = (int) (jp->centerx * jp->cr / 2) + jp->centerx - 2;
	old_circle.y = (int) (jp->centery * jp->ci / 2) + jp->centery - 2;
	incr(jp);
	new_circle.x = (int) (jp->centerx * jp->cr / 2) + jp->centerx - 2;
	new_circle.y = (int) (jp->centery * jp->ci / 2) + jp->centery - 2;



	fillellipse(img, Pt(old_circle.x, old_circle.y), 
		1, 1, display->black, ZP);
	fillellipse(img, Pt(new_circle.x, new_circle.y), 
		1, 1, display->white, ZP);


	if (jp->erase == 1) {
		//		XDrawPoints(display, window, gc,
		//		    jp->pointBuffer[jp->buffer], numpoints, CoordModeOrigin);

		for(i2 = 0; i2 < numpoints; i2++) {
			Rectangle r = Rpt((jp->pointBuffer[jp->buffer])[i2], addpt((jp->pointBuffer[jp->buffer])[i2], Pt(1,1)));
			draw(img, r, display->black, nil, ZP);

		}

	}
	jp->inc++;
	if (++jp->pix >= 256)
		jp->pix = 1;

	while (k--) {

		/* save calls to LRAND by using bit shifts over and over on the same
		   int for 32 iterations, then get a new random int */
		if (!(k % 32))
			rnd = LRAND();

		/* complex sqrt: x^0.5 = radius^0.5*(cos(theta/2) + i*sin(theta/2)) */

		xi -= jp->ci;
		xr -= jp->cr;

		/* Avoid atan2: DOMAIN error message */
		if (xi == 0.0 && xr == 0.0)
			theta = 0.0;
		else
			theta = atan2(xi, xr) / 2.0;

		/*r = pow(xi * xi + xr * xr, 0.25); */
		r = sqrt(sqrt(xi * xi + xr * xr));	/* 3 times faster */

		xr = r * cos(theta);
		xi = r * sin(theta);

		if ((rnd >> (k % 32)) & 0x1) {
			xi = -xi;
			xr = -xr;
		}
		xp->x = jp->centerx + (int) ((jp->centerx >> 1) * xr);
		xp->y = jp->centery + (int) ((jp->centery >> 1) * xi);
		xp++;
	}

	jp->itree = 0;
	apply(jp, xr, xi, jp->depth);

	//	XDrawPoints(display, window, gc,
	//		    jp->pointBuffer[jp->buffer], numpoints, CoordModeOrigin);

	for(i2 = 0; i2 < numpoints; i2++) {
		Rectangle r = Rpt((jp->pointBuffer[jp->buffer])[i2], addpt((jp->pointBuffer[jp->buffer])[i2], Pt(1,1)));
		draw(img, r,  colors[jp->pix], nil, ZP);

	}

	jp->buffer++;
	if (jp->buffer > jp->nbuffers - 1) {
		jp->buffer -= jp->nbuffers;
		jp->erase = 1;
	}
	if (jp->redrawing) {
		for (i = 0; i < REDRAWSTEP; i++) {
			j = (jp->buffer - jp->redrawpos + jp->nbuffers) % jp->nbuffers;
			//			XDrawPoints(display, window, gc,
			//			     jp->pointBuffer[j], numpoints, CoordModeOrigin);

			for(i2 = 0; i2 < numpoints; i2++) {
				Rectangle r = Rpt((jp->pointBuffer[j])[i2], addpt((jp->pointBuffer[j])[i2], Pt(1,1)));
				draw(img, r, colors[jp->pix], nil, ZP);

			}

			if (++(jp->redrawpos) >= jp->nbuffers) {
				jp->redrawing = 0;
				break;
			}
		}
	}
	draw(screen, screen->r, img, nil, ZP);
	flushimage(display, 1);
}

void
release_julia(void)
{
	if (julias != NULL) {

		juliastruct *jp = &julias[0];
		int         buffer;

		if (jp->pointBuffer) {
			for (buffer = 0; buffer < jp->nbuffers; buffer++)
				if (jp->pointBuffer[buffer])
					free((void *) jp->pointBuffer[buffer]);
			free((void *) jp->pointBuffer);
		}
	}
	free((void *) julias);
	julias = NULL;
}


void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

	draw(img, img->r, display->black, nil, ZP);

	if(julias != nil)
		release_julia();
	init_julia();
}

void 
threadmain(int argc, char **argv)
{
	Mouse m;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "julia") < 0)
		sysfatal("initdraw failed: %r");


	smooth(colors);
	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
			if(m.buttons&1) {

			}

		}
		draw_julia();
		frames++;
	}
}
