/* Lyap - calculate and display Lyapunov exponents */

/* Written by Ron Record (rr@sco) 03 Sep 1991 */

/* The idea here is to calculate the Lyapunov exponent for a periodically
 * forced logistic map (later i added several other nonlinear maps of the unit
 * interval). In order to turn the 1-dimensional parameter space of the
 * logistic map into a 2-dimensional parameter space, select two parameter
 * values ('a' and 'b') then alternate the iterations of the logistic map using
 * first 'a' then 'b' as the parameter. This program accepts an argument to
 * specify a forcing function, so instead of just alternating 'a' and 'b', you
 * can use 'a' as the parameter for say 6 iterations, then 'b' for 6 iterations
 * and so on. An interesting forcing function to look at is abbabaab (the
 * Morse-Thue sequence, an aperiodic self-similar, self-generating sequence).
 * Anyway, step through all the values of 'a' and 'b' in the ranges you want,
 * calculating the Lyapunov exponent for each pair of values. The exponent
 * is calculated by iterating out a ways (specified by the variable "settle")
 * then on subsequent iterations calculating an average of the logarithm of
 * the absolute value of the derivative at that point. Points in parameter
 * space with a negative Lyapunov exponent are colored one way (using the
 * value of the exponent to index into a color map) while points with a
 * non-negative exponent are colored differently.
 *
 * The algorithm was taken from the September 1991 Scientific American article
 * by A. K. Dewdney who gives credit to Mario Markus of the Max Planck Institute
 * for its creation. Additional information and ideas were gleaned from the
 * discussion on alt.fractals involving Stephen Hall, Ed Kubaitis, Dave Platt
 * and Baback Moghaddam. Assistance with colormaps and spinning color wheels
 * and X was gleaned from Hiram Clawson. Rubber banding code was adapted from
 * an existing Mandelbrot program written by Stacey Campbell.
 */

/*
 * Ported to Plan 9 by mirtchovski@gmail.com, 06/04
 */

#include "plan9.h"
#include "smooth.h"

int Etimer;
int resized;
int done;
#define M_LOG2E       1.4426950408889634074

#undef ABS
#define ABS(a)  (((a)<0) ? (0-(a)) : (a) )
#define Min(x,y) ((x < y)?x:y)
#define Max(x,y) ((x > y)?x:y)

#define MAXPOINTS  256
#define MAXFRAMES 8
#define MAXCOLOR 256
static int maxcolor=256, startcolor=17, color_offset=96, mincolindex=33;
static int dwell=100, settle=50;
static int width=256, height=256;
#endif

#define TRUE 1
#define FALSE 0


static ulong foreground,  background;


typedef struct {
	int x, y;
} xy_t;

typedef struct {
	int start_x, start_y;
	int last_x, last_y;
} rubber_band_data_t;

typedef struct {
	double p_min, p_max, q_min, q_max;
	rubber_band_data_t rubber_band;
} image_data_t;

typedef struct points_t {
	XPoint data[MAXCOLOR][MAXPOINTS];
	int npoints[MAXCOLOR];
} points_t;

static points_t Points;
static image_data_t rubber_data;


#define MAXINDEX 64
#define FUNCMAXINDEX 16
#define MAXWHEELS 7
#define NUMMAPS 5

typedef double (*PFD)(double,double);

static double logistic(double,double), circle(double,double), leftlog(double,double), rightlog(double,double), doublelog(double,double);
static double dlogistic(double,double), dcircle(double,double), dleftlog(double,double), drightlog(double,double), ddoublelog(double,double);
static PFD map, deriv;
static PFD Maps[NUMMAPS] = {
	logistic, circle, leftlog, rightlog, doublelog };
static PFD Derivs[NUMMAPS] = {
	dlogistic, dcircle, dleftlog, drightlog, ddoublelog };

static int aflag=0, bflag=0, wflag=0, hflag=0, Rflag=0;
static double pmins[NUMMAPS] = {
	2.0, 0.0, 0.0, 0.0, 0.0 };
static double pmaxs[NUMMAPS] = {
	4.0, 1.0, 6.75, 6.75, 16.0 };
static double amins[NUMMAPS] = {
	2.0, 0.0, 0.0, 0.0, 0.0 };
static double aranges[NUMMAPS] = {
	2.0, 1.0, 6.75, 6.75, 16.0 };
static double bmins[NUMMAPS] = {
	2.0, 0.0, 0.0, 0.0, 0.0 };
static double branges[NUMMAPS] = {
	2.0, 1.0, 6.75, 6.75, 16.0 };

static int   forcing[MAXINDEX] = {
	0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,
	0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,
	0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1 };
static int   Forcing[FUNCMAXINDEX] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };

static int   maxindex = MAXINDEX;
static int   funcmaxindex = FUNCMAXINDEX;
static double	min_a=2.0, min_b=2.0, a_range=2.0, b_range=2.0, minlyap=1.0;
static double  max_a=4.0, max_b=4.0;
static double  start_x=0.65, lyapunov, a_inc, b_inc, a, b;
static int	numcolors=16, numfreecols, displayplanes, lowrange;
static xy_t	point;
static XColor	Colors[MAXCOLOR];
static double  *exponents[MAXFRAMES];
static double  a_minimums[MAXFRAMES], b_minimums[MAXFRAMES];
static double  a_maximums[MAXFRAMES], b_maximums[MAXFRAMES];
static double  minexp, maxexp, prob=0.5;
static int     expind[MAXFRAMES]={
	0};
static int 	resiz[MAXFRAMES]={
	0};
static int	numwheels=MAXWHEELS, force=0, Force=0, negative=1;
static int     rgb_max=65000, nostart=1, stripe_interval=7;
static int	save=1, show=0, useprod=1, spinlength=256, savefile=0;
static int	maxframe=0, frame=0, dorecalc=0, mapindex=0, run=1;
static char	*outname="lyap.out";


static void resize(void);
static void redisplay(void);
static void Spin(void);
static void show_defaults(void);
static void InitBuffer(void);
static void BufferPoint(int color, int x, int y);
static void FlushBuffer(void);
static void init_canvas(void);
static void init_data(void);
static void parseargs(void);
static void Clear1(void);
static void setupmem(void);
static void main_event(void);
static int complyap(void);
static void Getkey(void);
static int sendpoint(double expo);
static void setforcing(void);
static void check_params(int mapnum, int parnum);
static void Destroy_frame(void);
static void freemem(void);
static void Redraw(void);
static void redraw(double *exparray, int index, int cont);
static void recalc(void);

void
initfun(void)
{
	int builtin;
	builtin = nrand(22);
	{
		char *ff = 0;
		switch (builtin) {
		case 0:
			min_a = 3.75;
			aflag++;
			min_b = 3.299999;
			bflag++;
			a_range = 0.05;
			wflag++;
			b_range = 0.05;
			hflag++;
			dwell = 200;
			settle = 100;
			ff = "abaabbaaabbb";
			break;

		case 1:
			min_a = 3.8;
			aflag++;
			min_b = 3.2;
			bflag++;
			b_range = .05;
			hflag++;
			a_range = .05;
			wflag++;
			ff = "bbbbbaaaaa";
			break;

		case 2:
			min_a =  3.4;
			aflag++;
			min_b =  3.04;
			bflag++;
			a_range =  .5;
			wflag++;
			b_range =  .5;
			hflag++;
			ff = "abbbbbbbbb";
			settle = 500;
			dwell = 1000;
			break;

		case 3:
			min_a = 3.5;
			aflag++;
			min_b = 3.0;
			bflag++;
			a_range = 0.2;
			wflag++;
			b_range = 0.2;
			hflag++;
			dwell = 600;
			settle = 300;
			ff = "aaabbbab";
			break;

		case 4:
			min_a = 3.55667;
			aflag++;
			min_b = 3.2;
			bflag++;
			b_range = .05;
			hflag++;
			a_range = .05;
			wflag++;
			ff = "bbbbbaaaaa";
			break;

		case 5:
			min_a = 3.79;
			aflag++;
			min_b = 3.22;
			bflag++;
			b_range = .02999;
			hflag++;
			a_range = .02999;
			wflag++;
			ff = "bbbbbaaaaa";
			break;

		case 6:
			min_a = 3.7999;
			aflag++;
			min_b = 3.299999;
			bflag++;
			a_range = 0.2;
			wflag++;
			b_range = 0.2;
			hflag++;
			dwell = 300;
			settle = 150;
			ff = "abaabbaaabbb";
			break;

		case 7:
			min_a = 3.89;
			aflag++;
			min_b = 3.22;
			bflag++;
			b_range = .028;
			hflag++;
			a_range = .02999;
			wflag++;
			ff = "bbbbbaaaaa";
			settle = 600;
			dwell = 1000;
			break;

		case 8:
			min_a = 3.2;
			aflag++;
			min_b = 3.7;
			bflag++;
			a_range = 0.05;
			wflag++;
			b_range = .005;
			hflag++;
			ff = "abbbbaa";
			break;

		case 9:
			ff = "aaaaaabbbbbb";
			mapindex = 1;
			dwell =  400;
			settle =  200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 10:
			ff = "aaaaaabbbbbb";
			mapindex = 1;
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 11:
			mapindex = 1;
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 12:
			ff = "abbb";
			mapindex = 1;
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 13:
			ff = "abbabaab";
			mapindex = 1;
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 14:
			ff = "abbabaab";
			dwell =  800;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			/* ####  -x 0.05 */
			min_a = 3.91;
			aflag++;
			a_range =  0.0899999999;
			wflag++;
			min_b =  3.28;
			bflag++;
			b_range =  0.35;
			hflag++;
			break;

		case 15:
			ff = "aaaaaabbbbbb";
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 16:
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 17:
			ff = "abbb";
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 18:
			ff = "abbabaab";
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 19:
			mapindex = 2;
			ff = "aaaaaabbbbbb";
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 20:
			mapindex = 2;
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 21:
			mapindex = 2;
			ff = "abbb";
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;

		case 22:
			mapindex = 2;
			ff = "abbabaab";
			dwell =  400;
			settle = 200;
			minlyap = maxexp = ABS(-0.85);
			minexp = -1.0 * minlyap;
			break;
		}

		if (ff) {
			char *ch;
			int bindex = 0;
			maxindex = strlen(ff);
			ch = ff;
			force++;
			while (bindex < maxindex) {
				if (*ch == 'a')
					forcing[bindex++] = 0;
				else if (*ch == 'b')
					forcing[bindex++] = 1;
				ch++;
			}
		}
	}

}

void
screenhack (void)
{


	width = Dx(screen->r);
	height = Dy(screen->r);

	parseargs();

	initfun();
	init_data();
	init_canvas();
	for(;;)
		main_event();
}

static void
main_event(void)
{


	if(resized) {
		resized = 0;
		resize();
	}

	if (complyap() == TRUE)
		run=0;
	if(run == 0) {
		sleep(10000);
		resize();
	}
	frames++;
	if(ecanmouse()) {
		m = emouse();
		if(m.buttons&4)
			if(emenuhit(3, &m, &menu) == 0)
				threadexitsall(nil);
	}
}

/* complyap() is the guts of the program. This is where the Lyapunov exponent
 * is calculated. For each iteration (past some large number of iterations)
 * calculate the logarithm of the absolute value of the derivative at that
 * point. Then average them over some large number of iterations. Some small
 * speed up is achieved by utilizing the fact that log(a*b) = log(a) + log(b).
 */
static int
complyap(void)
{
	int i, bindex;
	double total, prod, x, dx, r;

	if (!run)
		return TRUE;
	a += a_inc;
	if (a >= max_a) {
		if (sendpoint(lyapunov) == TRUE)
			return FALSE;
		else {
			FlushBuffer();
			return TRUE;
		}
	}
	if (b >= max_b) {
		FlushBuffer();
		return TRUE;
	}
	prod = 1.0;
	total = 0.0;
	bindex = 0;
	x = start_x;
	r = (forcing[bindex]) ? b : a;
	for (i=0;i<settle;i++) {     /* Here's where we let the thing */
		x = (*map)(x, r);	 /* "settle down". There is usually */
		if (++bindex >= maxindex) { /* some initial "noise" in the */
			bindex = 0;    /* iterations. How can we optimize */
			if (Rflag)      /* the value of settle ??? */
				setforcing();
		}
		r = (forcing[bindex]) ? b : a;
	}
	for (i=0;i<dwell;i++) {
		x = (*map)(x, r);
		dx = (*deriv)(x, r); /* ABS is a macro, so don't be fancy */
		dx = ABS(dx);
		if (dx == 0.0) /* log(0) is nasty so break out. */
		{
			i++;
			break;
		}
		prod *= dx;
		/* we need to prevent overflow and underflow */
		if ((prod > 1.0e12) || (prod < 1.0e-12)) {
			total += log(prod);
			prod = 1.0;
		}
		if (++bindex >= maxindex) {
			bindex = 0;
			if (Rflag)
				setforcing();
		}
		r = (forcing[bindex]) ? b : a;
	}
	total += log(prod);
	lyapunov = (total * M_LOG2E) / (double)i;

	if (sendpoint(lyapunov) == TRUE)
		return FALSE;
	else {
		FlushBuffer();
		return TRUE;
	}
}

static double
logistic(double x, double r)	    /* the familiar logistic map */
{
	return(r * x * (1.0 - x));
}

static double
dlogistic(double x, double r)	    /* the derivative of logistic map */
{
	return(r - (2.0 * r * x));
}

static double
circle(double x, double r)	  /* sin() hump or sorta like the circle map */
{
	return(r * sin(M_PI * x));
}

static double
dcircle(double x, double r)	  /* derivative of the "sin() hump" */
{
	return(r * M_PI * cos(M_PI * x));
}

static double
leftlog(double x, double r)	  /* left skewed logistic */
{
	double d;

	d = 1.0 - x;
	return(r * x * d * d);
}

static double
dleftlog(double x, double r)    /* derivative of the left skewed logistic */
{
	return(r * (1.0 - (4.0 * x) + (3.0 * x * x)));
}

static double
rightlog(double x, double r)    /* right skewed logistic */
{
	return(r * x * x * (1.0 - x));
}

static double
drightlog(double x, double r)    /* derivative of the right skewed logistic */
{
	return(r * ((2.0 * x) - (3.0 * x * x)));
}

static double
doublelog(double x, double r)    /* double logistic */
{
	double d;

	d = 1.0 - x;
	return(r * x * x * d * d);
}

static double
ddoublelog(double x, double r)   /* derivative of the double logistic */
{
	double d;

	d = x * x;
	return(r * ((2.0 * x) - (6.0 * d) + (4.0 * x * d)));
}

static void
init_data(void)
{
	numcolors = 256;
	displayplanes = 2;
	if (numcolors > maxcolor)
		numcolors = maxcolor;
	numfreecols = numcolors - mincolindex;
	lowrange = mincolindex - startcolor;
	a_inc = a_range / (double)width;
	b_inc = b_range / (double)height;
	point.x = -1;
	point.y = 0;
	a = rubber_data.p_min = min_a;
	b = rubber_data.q_min = min_b;
	rubber_data.p_max = max_a;
	rubber_data.q_max = max_b;
	InitBuffer();
}

static void
init_canvas(void)
{
	static int i;
}

static void
parseargs()
{
	static int i;
	int bindex=0;
	char *s, *ch;

	map = Maps[0];
	deriv = Derivs[0];

	mincolindex = 0;
	dwell = 50;

	minlyap=1.0;
	maxexp=minlyap;
	minexp= -1.0 * minlyap;


	color_offset = 0;

	maxcolor=256;
	if ((maxcolor - startcolor) <= 0)
		startcolor = 0;
	if ((maxcolor - mincolindex) <= 0) {
		mincolindex = 1;
		color_offset = 0;
	}

	prob = 0.5;

	settle = 50;

	min_a = 2.0;

	min_b = 2.0;

	numwheels = 7;

	s = "abbabaab";
	if (s && *s) {
		maxindex = strlen(s);
		ch = s;
		force++;
		while (bindex < maxindex) {
			if (*ch == 'a')
				forcing[bindex++] = 0;
			else
				forcing[bindex++] = 1;
			ch++;
		}
	}
	b_range = 2.0;
	start_x = 0.65;

	negative--;

	rgb_max = 65535;
	spinlength = 256;
	show = 0;

	a_range = 2.0;

	max_a = min_a + a_range;
	max_b = min_b + b_range;

	a_minimums[0] = min_a;
	b_minimums[0] = min_b;
	a_maximums[0] = max_a;
	b_maximums[0] = max_b;

}

/* Here's where we index into a color map. After the Lyapunov exponent is
 * calculated, it is used to determine what color to use for that point.
 * I suppose there are a lot of ways to do this. I used the following :
 * if it's non-negative then there's a reserved area at the lower range
 * of the color map that i index into. The ratio of some "minimum exponent
 * value" and the calculated value is used as a ratio of how high to index
 * into this reserved range. Usually these colors are dark red (see init_color).
 * If the exponent is negative, the same ratio (expo/minlyap) is used to index
 * into the remaining portion of the colormap (which is usually some light
 * shades of color or a rainbow wheel). The coloring scheme can actually make
 * a great deal of difference in the quality of the picture. Different colormaps
 * bring out different details of the dynamics while different indexing
 * algorithms also greatly effect what details are seen. Play around with this.
 */
static int
sendpoint(double expo)
{
	static int index;
	static double tmpexpo;

	point.x++;
	tmpexpo = (negative) ? expo : -1.0 * expo;
	if (tmpexpo > 0) {

		index = (int)(tmpexpo*lowrange/maxexp);
		index = (index % lowrange) + startcolor;
	}
	else {
		index = (int)(tmpexpo*numfreecols/minexp);
		index = (index % numfreecols) + mincolindex;
	}
	BufferPoint(index, point.x, point.y);
	if (point.x >= width) {
		point.y++;
		point.x = 0;
		if (save) {
			b += b_inc;
			a = min_a;
		}
		if (point.y >= height)
			return FALSE;
		else
			return TRUE;
	}
	return TRUE;
}


static void
resize(void)
{
	int n;

	width = Dx(screen->r);
	height = Dy(screen->r);
	a_inc = a_range / (double)width;
	b_inc = b_range / (double)height;
	point.x = -1;
	point.y = 0;
	run = 1;
	a = rubber_data.p_min = min_a;
	b = rubber_data.q_min = min_b;
	rubber_data.p_max = max_a;
	rubber_data.q_max = max_b;

	initfun();
	smooth(colors);

	for (n=0;n<MAXFRAMES;n++)
		if ((n <= maxframe) && (n != frame))
			resiz[n] = 1;
	InitBuffer();
	Clear1();
	Redraw();
}

static void
redraw(double *exparray, int index, int cont)
{
	static int i;
	static int x_sav, y_sav;

	x_sav = point.x;
	y_sav = point.y;

	point.x = -1;
	point.y = 0;

	save=0;
	for (i=0;i<index;i++)
		sendpoint(exparray[i]);
	save=1;

	if (cont) {
		point.x = x_sav;
		point.y = y_sav;
	}
	else {
		a = point.x * a_inc + min_a;
		b = point.y * b_inc + min_b;
	}
	FlushBuffer();
}

static void
Redraw(void)
{
	FlushBuffer();
	point.x = -1;
	point.y = 0;
	run = 1;
	a = min_a;
	b = min_b;
	expind[frame] = 0;
	resiz[frame] = 0;
}


static void
Clear1(void)
{
	draw(screen, screen->r, display->black, nil, ZP);
	InitBuffer();
}


static void
set_new_params(image_data_t *data)
{
	frame = (maxframe + 1) % MAXFRAMES;
	if (frame > maxframe)
		maxframe = frame;
	a_range = data->p_max - data->p_min;
	b_range = data->q_max - data->q_min;
	a_minimums[frame] = min_a = data->p_min;
	b_minimums[frame] = min_b = data->q_min;
	a_inc = a_range / (double)width;
	b_inc = b_range / (double)height;
	point.x = -1;
	point.y = 0;
	run = 1;
	a = min_a;
	b = min_b;
	a_maximums[frame] = max_a = data->p_max;
	b_maximums[frame] = max_b = data->q_max;
	expind[frame] = 0;
	Clear1();
}

static void
InitBuffer(void)
{
	int i;

	for (i = 0 ; i < maxcolor; ++i)
		Points.npoints[i] = 0;
}

static void
BufferPoint(int color, int x, int y)
{
	int i;
	Rectangle r;
	Point p;

	/* Guard against bogus color values. Shouldn't be necessary but paranoia  is good. */
	if (color < 0)
		color = 0;
	else if (color >= maxcolor)
		color = maxcolor - 1;

	if (Points.npoints[color] == MAXPOINTS)
	{
		//XDrawPoints(display, window, Data_GC[color],
		//    Points.data[color], Points.npoints[color], CoordModeOrigin);
		//XDrawPoints(display, pixmap, Data_GC[color],
		//    Points.data[color], Points.npoints[color], CoordModeOrigin);

		for(i = 0; i < Points.npoints[color]; i++) {

			p = (Points.data[color])[i];
			p = addpt(p, screen->r.min);
			r = Rpt(p, addpt(p, Pt(1,1)));
			draw(screen, r, colors[color] , nil, ZP);
		}
		flushimage(display, 1);
		Points.npoints[color] = 0;
	}
	Points.data[color][Points.npoints[color]].x = x;
	Points.data[color][Points.npoints[color]].y = y;
	++Points.npoints[color];
}

static void
FlushBuffer(void)
{
	int color;
	int i;
	Rectangle r;
	Point p;

	for (color = 0; color < maxcolor; ++color)
		if (Points.npoints[color])
		{
			//XDrawPoints(dpy, canvas, Data_GC[color],
			//    Points.data[color], Points.npoints[color],
			//    CoordModeOrigin);
			//XDrawPoints(dpy, pixmap, Data_GC[color],
			//    Points.data[color], Points.npoints[color],
			//    CoordModeOrigin);
			for(i = 0; i < Points.npoints[color]; i++) {

				p = (Points.data[color])[i];
				p = addpt(p, screen->r.min);
				r = Rpt(p, addpt(p, Pt(1,1)));
				draw(screen, r, colors[color] , nil, ZP);
			}

			flushimage(display, 1);
			Points.npoints[color] = 0;
		}
}


static void
setforcing(void)
{
	static int i;
	for (i=0;i<MAXINDEX;i++)
		forcing[i] = (random() > prob) ? 0 : 1;
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	resized = 1;


	draw(screen, screen->r,display->black, nil, ZP);
	flushimage(display, 1);

	smooth(colors);

}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "lyap") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);

	eresized(0);
	resized = 0;
	screenhack();

}
