/* xscreensaver, Copyright (c) 2002, 2004 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 * Memscroller -- scrolls a dump of its own RAM across the screen.
 */

/*
 * Ported to Plan 9 by mirtchovski@gmail.com, 08/04
 */

#include "plan9.h"
#include "smooth.h"

typedef struct {
	int which;
	XRectangle rect;
	XImage *image;
	int rez;
	int speed;
	int scroll_tick;
	unsigned char *data;
	int count_zero;
} scroller;

typedef struct {
	int border;

	enum {
		SEED_RAM, SEED_RANDOM 		} seed_mode;
	enum {
		DRAW_COLOR, DRAW_MONO 		} draw_mode;

	int nscrollers;
	scroller *scrollers;
} state;


state *
init_memscroller (void)
{
	int i;
	state *st = (state *) calloc (1, sizeof (*st));
	char *s;

	st->seed_mode = SEED_RAM;

	st->nscrollers = 3;
	st->scrollers = (scroller *) calloc (st->nscrollers, sizeof(scroller));

	for (i = 0; i < st->nscrollers; i++)
	{
		scroller *sc = &st->scrollers[i];
		int max_height = 4096;

		sc->which = i;
		sc->speed = i+1;

		sc->image = 0;

		if (!sc->image) {
			sc->image = allocimage(display, Rect(0, 0, 1, max_height), screen->chan, 0, DBlack);
			if(sc->image == nil)
				sysfatal("cannot allocate buffer image");
		}
	}

	return st;
}


static void
reshape_memscroller (state *st)
{
	int i, width, height;

	for (i = 0; i < st->nscrollers; i++)
	{
		scroller *sc = &st->scrollers[i];

		if (i == 0)
		{
			sc->rez = 6;  /* #### */

			width  = (((int) (Dx(screen->r)* 0.8)
				/ sc->rez) * sc->rez);
			height = (((int) (Dy(sfreen->r) * 0.3)
				/ sc->rez) * sc->rez);

			sc->rect.min.x = (Dx(screen->r)  - width)  / 2;
			sc->rect.min.y = (Dx(screen->r) - height) / 2;
			sc->rect.max = addpt(sc->rect.min, Pt(width, height));

		}
		else
		{
			scroller *sc0 = &st->scrollers[i-1];
			sc->rez = sc0->rez * 1.8;

			sc->rect.min.x      = sc0->rect.min.x;
			sc->rect.min.y      = (sc0->rect.max.y + st->border + (st->border + 2) * 7);
			sc->rect.max.x  = sc->rect.max.x;
			sc->rect.max.y = sc->rect.min.y + (((int) (Dy(screen->r) * 0.1)
				/ sc->rez) * sc->rez);
		}

		//XDrawRectangle (st->dpy, st->window, st->draw_gc,
		//   sc->rect.x - st->border*2,
		//    sc->rect.y - st->border*2,
		//    sc->rect.width  + st->border*4,
		//    sc->rect.height + st->border*4);
		draw(img, sc->rect, colors[Border], nil, ZP);
		draw(img, insetrect(sc->rect, sc->border), display->black, nil, ZP);
	}
}



static unsigned int
more_bits (state *st, scroller *sc)
{
	static unsigned char *lomem = 0;
	static unsigned char *himem = 0;
	unsigned char r, g, b;
	unsigned int v;

# undef PACK
# define PACK(r,g,b) ((r) | ((g) << 8) | ((b) << 16) | ((r) << 24))

	switch (st->seed_mode)
	{
	case SEED_RAM:
		if (himem == 0)
		{
			lomem = (unsigned char *) progname; /* not first malloc, but early */
			himem = (unsigned char *)           /* not last malloc, but late */
				st->scrollers[st->nscrollers-1].image->data;
		}

		if (sc->data < lomem)
			sc->data = lomem;

		if (lomem >= himem) abort();

RETRY:
		if (sc->data >= himem)
			sc->data = lomem;

		switch (st->draw_mode)
		{
		case DRAW_COLOR:
			r = *sc->data++;
			g = *sc->data++;
			b = *sc->data++;
			break;
		case DRAW_MONO:
			r = 0;
			g = *sc->data++;
			b = 0;
			break;
		default:
			abort();
		}

		v = PACK(r,g,b);

		/* avoid having many seconds of blackness: truncate zeros at 24K. */
		if (v == 0)
			sc->count_zero++;
		else
			sc->count_zero = 0;
		if (sc->count_zero > 1024 * (st->draw_mode == DRAW_COLOR ? 24 : 8))
			goto RETRY;

		break;

	case SEED_RANDOM:
		v = random();
		switch (st->draw_mode)
		{
		case DRAW_COLOR:
			r = (v >> 16) & 0xFF;
			g = (v >>  8) & 0xFF;
			b = (v      ) & 0xFF;
			break;
		case DRAW_MONO:
			r = 0;
			g = v & 0xFF;
			b = 0;
			break;
		default:
			abort();
		}
		v = PACK(r,g,b);
		break;

	default:
		abort();
	}

# undef PACK
	return v;
}


static void
draw_string (state *st, unsigned int n)
{
	char buf[40];
	int direction, ascent, descent;
	XCharStruct overall;
	int x, y, w, h;
	int bot = st->scrollers[0].rect.y;
	const char *fmt = "%08X";

	sprintf (buf, fmt, 0);
	XTextExtents (st->font, buf, strlen(buf), 
		&direction, &ascent, &descent, &overall);
	sprintf (buf, "%08X", n);

	w = overall.width;
	h = st->font->ascent + st->font->descent + 1;
	x = (st->xgwa.width - w) / 2;
	y = (bot - h) / 2;

	if (y + h + 10 > bot) return;

	XFillRectangle (st->dpy, st->window, st->erase_gc,
		x-w, y, w*3, h);
	XDrawString (st->dpy, st->window, st->text_gc,
		x, y + st->font->ascent, buf, strlen(buf));
}


static void
draw_memscroller (state *st)
{
	int i;

	draw_string (st, *((unsigned int *) st->scrollers[0].image->data));

	for (i = 0; i < st->nscrollers; i++)
	{
		scroller *sc = &st->scrollers[i];
		int j;

		XCopyArea (st->dpy, st->window, st->window, st->draw_gc,
			sc->rect.x + sc->speed, sc->rect.y,
			sc->rect.width - sc->speed, sc->rect.height,
			sc->rect.x, sc->rect.y);

		if (sc->scroll_tick == 0)
		{
			int top = ((sc->image->bytes_per_line * sc->rect.height) /
				(4 * sc->rez));
			unsigned int *out = (unsigned int *) sc->image->data;
			for (j = 0; j < top; j++)
			{
				unsigned int v = more_bits(st, sc);
				int k;
				for (k = 0; k < sc->rez; k++)
					*out++ = v;
			}
		}

		sc->scroll_tick++;
		if (sc->scroll_tick * sc->speed >= sc->rez)
			sc->scroll_tick = 0;

		for (j = 0; j < sc->speed; j++)
		{
			XPutImage (st->dpy, st->window, st->draw_gc, sc->image,
				0, 0,
				sc->rect.x + sc->rect.width - sc->image->width - j,
				sc->rect.y,
				sc->rect.width, sc->rect.height);
		}
	}
}


char *defaults [] = {
	".background:		   black",
	"*drawMode:		   color",
	"*seedMode:		   ram",
	".textColor:		   #00FF00",
	".foreground:		   #00FF00",
	"*borderSize:		   2",
	".font:		   -*-courier-medium-r-*-*-*-1400-*-*-m-*-*-*",
	"*delay:		   10000",
	"*offset:		   0",
	0
};

XrmOptionDescRec options [] = {
	{ "-delay",		".delay",		XrmoptionSepArg, 0 },
	{ "-font",		".font",		XrmoptionSepArg, 0 },
	{ "-color",		".drawMode",		XrmoptionNoArg, "color"  },
	{ "-mono",		".drawMode",		XrmoptionNoArg, "mono"   },
	{ "-ram",		".seedMode",		XrmoptionNoArg, "ram"    },
	{ "-random",		".seedMode",		XrmoptionNoArg, "random" },
	{ 0, 0, 0, 0 }
};




void
screenhack (void)
{
	int delay = 0;
	while(1)
	{
		draw_memscroller (st);
		flushimage(display, 1);
		frames++;
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4) {
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(0);
			}
		}
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}
	state *st = init_memscroller();
	reshape_memscroller(st);

	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

	smooth(colors);
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "memscroller") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);

	eresized(0);
	screenhack();

}
