/*
 * pedal
 *
 * Based on a program for some old PDP-11 Graphics Display Processors
 * at CMU.
 *
 * X version by
 *
 *  Dale Moore  <Dale.Moore@cs.cmu.edu>
 *  24-Jun-1994
 *
 *  Copyright (c) 1994, by Carnegie Mellon University.  Permission to use,
 *  copy, modify, distribute, and sell this software and its documentation
 *  for any purpose is hereby granted without fee, provided fnord that the
 *  above copyright notice appear in all copies and that both that copyright
 *  notice and this permission notice appear in supporting documentation.
 *  No representations are made about the  suitability of fnord this software
 *  for any purpose.  It is provided "as is" without express or implied
 *  warranty.
 */

/* 
  * ported to Plan 9 by andrey@lanl.gov, 07/02

/* plan9-related stuff */
#include "plan9.h"
int col;


/* end of plan9-related stuff */

/* If MAXLINES is too big, we might not be able to get it
 * to the X server in the 2byte length field. Must be less
 * than 16k
 */
#define MAXLINES (16 * 1024)
#define MAXPOINTS MAXLINES
XPoint *points;

/* 
 * If the pedal has only this many lines, it must be ugly and we dont
 * want to see it.
 */
#define MINLINES 7

static int sizex, sizey;
static int delay;
static int fadedelay;
static int maxlines;

static Bool fade_p;


/*
 * Routine (Macro actually)
 *   mysin
 * Description:
 *   Assume that degrees is .. oh 360... meaning that
 *   there are 360 degress in a circle.  Then this function
 *   would return the sin of the angle in degrees.  But lets
 *   say that they are really big degrees, with 4 big degrees
 *   the same as one regular degree.  Then this routine
 *   would be called mysin(t, 90) and would return sin(t degrees * 4)
 */
#define mysin(t, degrees) sin(t * 2 * M_PI / (degrees))
#define mycos(t, degrees) cos(t * 2 * M_PI / (degrees))

/*
 * Macro:
 *   rand_range
 * Description:
 *   Return a random number between a inclusive  and b exclusive.
 *    rand (3, 6) returns 3 or 4 or 5, but not 6.
 */
#define rand_range(a, b) (a + random() % (b - a))


static int
gcd(int m, int n) /* Greatest Common Divisor (also Greates common factor). */
{
	int r;

	for (;;) {
		r = m % n;
		if (r == 0) return (n);
		m = n;
		n = r;
	}

}

static int numlines (int a, int b, int d)
/*
 * Description:
 *
 *      Given parameters a and b, how many lines will we have to draw?
 *
 * Algorithm:
 *
 *      This algorithm assumes that r = sin (theta * a), where we
 *      evaluate theta on multiples of b.
 *
 *      LCM (i, j) = i * j / GCD (i, j);
 *
 *      So, at LCM (b, 360) we start over again.  But since we
 *      got to LCM (b, 360) by steps of b, the number of lines is
 *      LCM (b, 360) / b.
 *
 *      If a is odd, then at 180 we cross over and start the
 *      negative.  Someone should write up an elegant way of proving
 *      this.  Why?  Because I'm not convinced of it myself. 
 *
 */
{
#define odd(x) (x & 1)
#define even(x) (!odd(x))
	if ( odd(a) && odd(b) && even(d)) d /= 2;
	return  (d / gcd (d, b));
#undef odd
}

static int
compute_pedal(XPoint *points, int maxpoints)
/*
 * Description:
 *
 *    Basically, it's combination spirograph and string art.
 *    Instead of doing lines, we just use a complex polygon,
 *    and use an even/odd rule for filling in between.
 *
 *    The spirograph, in mathematical terms is a polar
 *    plot of the form r = sin (theta * c);
 *    The string art of this is that we evaluate that
 *    function only on certain multiples of theta.  That is
 *    we let theta advance in some random increment.  And then
 *    we draw a straight line between those two adjacent points.
 *
 *    Eventually, the lines will start repeating themselves
 *    if we've evaluated theta on some rational portion of the
 *    whole.
 *
 *    The number of lines generated is limited to the
 *    ratio of the increment we put on theta to the whole.
 *    If we say that there are 360 degrees in a circle, then we
 *    will never have more than 360 lines.   
 *
 * Return:
 *
 *    The number of points.
 *
 */
{
	int a, b, d;  /* These describe a unique pedal */

	double r;
	int theta = 0;
	XPoint *pp = points;
	int count;
	int numpoints;


	/* Just to make sure that this division is not done inside the loop */
	int h_width = sizex / 2, h_height = sizey / 2 ;

	USED(maxpoints);

	for (;;) {
		d = rand_range (MINLINES, maxlines);

		a = rand_range (1, d);
		b = rand_range (1, d);
		numpoints = numlines(a, b, d);
		if (numpoints > MINLINES) break;
	}

	/* it might be nice to try to move as much sin and cos computing
     * (or at least the argument computing) out of the loop.
     */
	for (count = numpoints; count-- ; )
	{
		r = mysin (theta * a, d);

		/* Convert from polar to cartesian coordinates */
		/* We could round the results, but coercing seems just fine */
		pp->x = screen->r.min.x + mysin (theta, d) * r * h_width  + h_width;
		pp->y = screen->r.min.y + mycos (theta, d) * r * h_height + h_height;

		/* Advance index into array */
		pp++;

		/* Advance theta */
		theta += b;
		theta %= d;
	}

	return(numpoints);
}

static void
init_pedal (void)
{


	delay = 2000;
	if (delay < 0) delay = 0;

	fadedelay = 2;
	if (fadedelay < 0) fadedelay = 0;

	maxlines = 1000;
	if (maxlines < MINLINES) maxlines = MINLINES;
	else if (maxlines > MAXLINES) maxlines = MAXLINES;

	points = (XPoint *)malloc(sizeof(XPoint) * maxlines);

	sizex = Dx(screen->r);
	sizey = Dy(screen->r);

}


static void
pedal (void)
/*
 *    Since the XFillPolygon doesn't require that the last
 *    point == first point, the number of points is the same
 *    as the number of lines.  We just let XFillPolygon supply
 *    the line from the last point to the first point.
 *
 */
{
	int numpoints;

	numpoints = compute_pedal(points, maxlines);
	col = nrand(255) + 1;


	draw(screen, screen->r, display->black, nil, ZP);


	fillpoly(screen, points, numpoints, 1, colors[col], ZP);



}


void
screenhack (void)
{
	init_pedal ();
	for (;;) {
		pedal ();

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		flushimage(display, 1);
		frames++;
		if (delay) 
			sleep (delay);

	}
}

void
eresized(int new)
{

	USED(new);

	fprint(2, "sorry, resize not supported\n");
	threadexitsall("resize");
}



void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "pedal") < 0)
		sysfatal("initdraw failed: %r");


	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(nrand(65535)));
	}
	einit(Emouse);

	screenhack();
}
