/* petri, simulate mold in a petri dish. v2.7
 * by Dan Bornstein, danfuzz@milk.com
 * with help from Jamie Zawinski, jwz@jwz.org
 * Copyright (c) 1992-1999 Dan Bornstein.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 *
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/03
 */

/* plan9-related stuff */
#include "plan9.h"

#define FLOAT float
#define RAND_FLOAT (((FLOAT) (random() & 0xffff)) / ((FLOAT) 0x10000))

typedef struct cell_s 
{
	unsigned char col;              /*  0      */
	unsigned char isnext;           /*  1      */
	unsigned char nextcol;          /*  2      */
	/*  3      */
	struct cell_s *next;            /*  4      */
	struct cell_s *prev;            /*  8    - */
	FLOAT speed;                    /* 12      */
	FLOAT growth;                   /* 16 20 - */
	FLOAT nextspeed;                /* 20 28   */
	/* 24 36 - */
} cell;

static int arr_width;
static int arr_height;
static int count;

static cell *arr;
static cell *head;
static cell *tail;
static int blastcount;

static int windowWidth;
static int windowHeight;
static int xOffset;
static int yOffset;
static int xSize;
static int ySize;

static FLOAT orthlim = 1.0;
static FLOAT diaglim;
static FLOAT anychan;
static FLOAT minorchan;
static FLOAT instantdeathchan;
static int minlifespan;
static int maxlifespan;
static FLOAT minlifespeed;
static FLOAT maxlifespeed;
static FLOAT mindeathspeed;
static FLOAT maxdeathspeed;
static Bool originalcolors;

#define cell_x(c) (((c) - arr) % arr_width)
#define cell_y(c) (((c) - arr) / arr_width)


static int random_life_value (void)
{
	return (int) ((RAND_FLOAT * (maxlifespan - minlifespan)) + minlifespan);
}


static void setup_display (void)
{

	int cell_size = nrand(2) + 1;

	count = 8;
	diaglim  = 1.414;
	diaglim *= orthlim;

	anychan  = 0.0015;
	minorchan = 0.5;
	instantdeathchan = 0.2;
	minlifespan = 500;
	maxlifespan = 1500;
	if (maxlifespan < minlifespan)
	{
		maxlifespan = minlifespan;
	}
	minlifespeed = 0.04;
	maxlifespeed = 0.13;
	if (maxlifespeed < minlifespeed)
	{
		maxlifespeed = minlifespeed;
	}
	mindeathspeed = 0.42;
	maxdeathspeed = 0.46;
	if (maxdeathspeed < mindeathspeed)
	{
		maxdeathspeed = mindeathspeed;
	}

	minlifespeed *= diaglim;
	maxlifespeed *= diaglim;
	mindeathspeed *= diaglim;
	maxdeathspeed *= diaglim;

	windowWidth = Dx(screen->r);
	windowHeight = Dy(screen->r);

	arr_width = windowWidth / cell_size;
	arr_height = windowHeight / cell_size;



	xSize = windowWidth / arr_width;
	ySize = windowHeight / arr_height;
	if (xSize > ySize)
	{
		xSize = ySize;
	}
	else
	{
		ySize = xSize;
	}

	xOffset = (windowWidth - (arr_width * xSize)) / 2;
	yOffset = (windowHeight - (arr_height * ySize)) / 2;
}

static void drawblock (int x, int y, unsigned char c)
{
	//	if (xSize == 1 && ySize == 1)
	//		XDrawPoint (display, window, coloredGCs[c], x + xOffset, y + yOffset);
	//	else
	//		XFillRectangle (display, window, coloredGCs[c],
	//		    x * xSize + xOffset, y * ySize + yOffset,
	//		    xSize, ySize);
	draw(screen, rectaddpt(Rect(x * xSize + xOffset, y * ySize + yOffset,
		x * xSize + xOffset + xSize, y * ySize + yOffset + ySize),
		screen->r.min),
		colors[c], nil, ZP);
}

static void setup_arr (void)
{
	int x, y;

	if (arr != NULL)
	{
		free (arr);
	}

	//XFillRectangle (display, window, coloredGCs[0], 0, 0, 
	//    windowWidth, windowHeight);
	draw(screen, screen->r, display->black, nil, ZP);

	arr = (cell *) calloc (sizeof(cell), arr_width * arr_height);
	if (!arr)
	{
		fprint (2, "%s: out of memory allocating %dx%d grid\n",
			"petri", arr_width, arr_height);
		exits ("out of memory");
	}

	for (y = 0; y < arr_height; y++)
	{
		int row = y * arr_width;
		for (x = 0; x < arr_width; x++)
		{
			arr[row+x].speed = 0.0;
			arr[row+x].growth = 0.0;
			arr[row+x].col = 0;
			arr[row+x].isnext = 0;
			arr[row+x].next = 0;
			arr[row+x].prev = 0;
		}
	}

	if (head == NULL)
	{
		head = (cell *) malloc (sizeof (cell));
	}

	if (tail == NULL)
	{
		tail = (cell *) malloc (sizeof (cell));
	}

	head->next = tail;
	head->prev = head;
	tail->next = tail;
	tail->prev = head;

	blastcount = random_life_value ();
}

static void newcell (cell *c, unsigned char col, FLOAT sp)
{
	if (! c) return;

	if (c->col == col) return;

	c->nextcol = col;
	c->nextspeed = sp;
	c->isnext = 1;

	if (c->prev == 0) {
		c->next = head->next;
		c->prev = head;
		head->next = c;
		c->next->prev = c;
	}
}

static void killcell (cell *c)
{
	c->prev->next = c->next;
	c->next->prev = c->prev;
	c->prev = 0;
	c->speed = 0.0;
	drawblock (cell_x(c), cell_y(c), c->col);
}


static void randblip (int doit)
{
	int n;
	int b = 0;
	if (!doit 
		&& (blastcount-- >= 0)
		&& (RAND_FLOAT > anychan))
	{
		return;
	}

	if (blastcount < 0)
	{
		b = 1;
		n = 2;
		blastcount = random_life_value ();
		if (RAND_FLOAT < instantdeathchan)
		{
			/* clear everything every so often to keep from getting into a
	     * rut */
			setup_arr ();
			b = 0;
		}
	}
	else if (RAND_FLOAT <= minorchan)
	{
		n = 2;
	}
	else 
	{
		n = random () % 3 + 3;
	}
	while (n--)
	{
		int x = random () % arr_width;
		int y = random () % arr_height;
		int c;
		FLOAT s;
		if (b)
		{
			c = 0;
			s = RAND_FLOAT * (maxdeathspeed - mindeathspeed) + mindeathspeed;
		}
		else
		{
			c = (random () % (count-1)) + 1;
			s = RAND_FLOAT * (maxlifespeed - minlifespeed) + minlifespeed;
		}
		newcell (&arr[y * arr_width + x], c, s);
	}
}

static void update (void)
{
	cell *a;

	for (a = head->next; a != tail; a = a->next)
	{
		static XPoint all_coords[] = {
			{-1, -1},
			{-1, 1},
			{1, -1},
			{1, 1},
			{-1,  0},
			{ 1, 0},
			{0, -1},
			{0, 1},
			{99, 99}
		};		



		XPoint *coords;

		if (a->speed == 0) continue;
		a->growth += a->speed;

		if (a->growth >= diaglim)
		{
			coords = all_coords;
		}
		else if (a->growth >= orthlim)
		{
			coords = &all_coords[4];
		}
		else
		{
			continue;
		}

		while (coords->x != 99)
		{
			int x = cell_x(a) + coords->x;
			int y = cell_y(a) + coords->y;
			coords++;

			if (x < 0) x = arr_width - 1;
			else if (x >= arr_width) x = 0;

			if (y < 0) y = arr_height - 1;
			else if (y >= arr_height) y = 0;

			newcell (&arr[y * arr_width + x], a->col, a->speed);
		}

		if (a->growth >= diaglim)
			killcell (a);
	}

	randblip ((head->next) == tail);

	for (a = head->next; a != tail; a = a->next)
	{
		if (a->isnext)
		{
			a->isnext = 0;
			a->speed = a->nextspeed;
			a->growth = 0.0;
			a->col = a->nextcol;
			drawblock (cell_x(a), cell_y(a), a->col + count);
		}
	}
}


void screenhack (void)
{
	int delay = 10;
	setup_display ();

	setup_arr ();

	randblip (1);

	for (;;)
	{
		update ();
		flushimage(display, 1);
		frames++;

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		sleep (delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	setup_display ();

	setup_arr ();

	randblip (1);

	draw(screen, screen->r, display->black, nil, ZP);

}

void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "petri") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 16; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			cmap2rgba(nrand(65535)+1));
	}
	freeimage(colors[0]);
	colors[0] = display->black;
	freeimage(colors[8]);
	colors[8] = display->white;

	einit(Emouse);

	screenhack();
}
