/* xscreensaver, Copyright (c) 1992, 1994, 1996, 1998, 2001
 *  Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* Draw some fireworks.  Inspired from TI Explorer Lisp code by 
   John S. Pezaris <pz@hx.lcs.mit.edu>
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 09/03.
 */

/* plan9-related stuff */
#include "plan9.h"

struct projectile {
	int x, y;	/* position */
	int dx, dy;	/* velocity */
	int decay;
	int size;
	int fuse;
	Bool primary;
	Bool dead;
	int color;
	struct projectile *next_free;
};


static struct projectile *projectiles, *free_projectiles;
static struct projectile **sorted_projectiles;


/* Slightly whacked, for better explosions
 */
#define PI_2000 6284

static int sin_cache[PI_2000];
static int cos_cache[PI_2000];

static void
cache(void)
{               /*needs to be run once. Could easily be */
	int i;        /*reimplemented to run and cache at compile-time,*/
	double dA;    /*saving on init_pyro time */
	for (i=0; i<PI_2000; i++)
	{
		dA=sin(((double) (random() % (PI_2000/2)))/1000.0);
		/*Emulation of spherical distribution*/
		dA+=asin(frand())/M_PI_2*0.1;
		/*Approximating the integration of the binominal, for
        well-distributed randomness*/
		cos_cache[i]=(int) (cos(((double)i)/1000.0)*dA*2500.0);
		sin_cache[i]=(int) (sin(((double)i)/1000.0)*dA*2500.0);
	}
}


static struct projectile *
get_projectile (void)
{
	struct projectile *p;
	if (free_projectiles)
	{
		p = free_projectiles;
		free_projectiles = p->next_free;
		p->next_free = 0;
		p->dead = False;
		return p;
	}
	else
		return 0;
}

static void
free_projectile (struct projectile *p)
{
	p->next_free = free_projectiles;
	free_projectiles = p;
	p->dead = True;
}

static void
launch (int xlim, int ylim, int g)
{
	struct projectile *p = get_projectile ();
	int x, dx, xxx;
	if (! p) return;

	do {
		x = (random () % xlim);
		dx = 30000 - (random () % 60000);
		xxx = x + (dx * 200);
	} while (xxx <= 0 || xxx >= xlim);

	p->x = x;
	p->y = ylim;
	p->dx = dx;
	p->size = 8000;
	p->decay = 0;
	p->dy = (random () % 4000) - 13000;
	p->fuse = ((((random () % 500) + 500) * abs (p->dy / g)) / 1000);
	p->primary = True;

	/* cope with small windows -- those constants assume big windows. */
	{
		int div = 1000000 / ylim;
		if (div > 1)
			p->fuse /= div;
	}

	p->color = nrand(256);
}

static struct projectile *
shrapnel (struct projectile *parent)
{
	struct projectile *p = get_projectile ();
	int v;
	if (! p) return 0;
	p->x = parent->x;
	p->y = parent->y;
	v=random () % PI_2000;
	p->dx =(sin_cache[v]) + parent->dx;
	p->dy =(cos_cache[v]) + parent->dx;
	p->decay = (random () % 50) - 60;
	p->size = (parent->size * 2) / 3;
	p->fuse = 0;
	p->primary = False;

	p->color = parent->color;

	return p;
}

static unsigned int default_fg_pixel;

static int how_many, frequency, scatter;

static void
init_pyro (void)
{
	int i;

	how_many = 600;
	frequency = 50;
	scatter = 100;

	projectiles = 0;
	free_projectiles = 0;
	projectiles = (struct projectile *)
		calloc (how_many, sizeof (*projectiles));
	sorted_projectiles = (struct projectile **)
		calloc (how_many, sizeof (*sorted_projectiles));
	for (i = 0; i < how_many; i++)
		free_projectile (&projectiles [i]);
	for (i = 0; i < how_many; i++)
		sorted_projectiles[i] = &projectiles[i];


	cache();
}


static int
projectile_pixel_sorter (const void *a, const void *b)
{
	struct projectile *pa = *(struct projectile **) a;
	struct projectile *pb = *(struct projectile **) b;
	if (pa->color == pb->color) return 0;
	else if (pa->color < pb->color) return -1;
	else return 1;
}

static void
sort_by_pixel (int length)
{
	qsort ((void *) sorted_projectiles,
		length,
		sizeof(*sorted_projectiles),
		projectile_pixel_sorter);
}


static void
pyro (void)
{
	static int xlim, ylim, real_xlim, real_ylim;
	int g = 100;
	int resort = 0;
	int i;
	Rectangle r;

	for (i = 0; i < how_many; i++)
	{
		struct projectile *p = sorted_projectiles [i];
		int old_x, old_y, old_size;
		int size, x, y;
		if (p->dead) continue;
		old_x = p->x >> 10;
		old_y = p->y >> 10;
		old_size = p->size >> 10;
		size = (p->size += p->decay) >> 10;
		x = (p->x += p->dx) >> 10;
		y = (p->y += p->dy) >> 10;
		p->dy += (p->size >> 6);
		if (p->primary) p->fuse--;

		/* erase old one */
		if (old_size > 0) {
			if(old_size < 4) {
				r = rectaddpt(Rect(old_x, old_y, old_x+old_size, old_y+old_size), screen->r.min);
				draw(screen, r, display->black, nil, ZP);
			} else {
				fillellipse(screen, addpt(screen->r.min, 
					Pt(old_x, old_y)), 
					old_size, old_size, display->black, ZP);
			}
		}

		if ((p->primary ? (p->fuse > 0) : (p->size > 0)) &&
			x < real_xlim &&
			y < real_ylim &&
			x > 0 &&
			y > 0)
		{
			if (size > 0)
			{
				static unsigned long last_pixel = ~0;
				unsigned long pixel;


				pixel = p->color;

				if (size < 4) {
					//XFillRectangle (dpy, window, draw_gc, x, y, size, size);
					r = rectaddpt(Rect(x, y, x+size, y+size), screen->r.min);
					draw(screen, r, colors[pixel], nil, ZP);
				} else {
					//XFillArc (dpy, window, draw_gc, x, y, size, size, 0, 360*64);
					fillellipse(screen, addpt(screen->r.min, 
						Pt(x, y)), 
						size, size, colors[pixel], ZP);
				}

			}
		}
		else
		{
			free_projectile (p);
		}

		if (p->primary && p->fuse <= 0)
		{
			int j = (random () % scatter) + (scatter/2);
			while (j--)
				shrapnel (p);
			resort = 1;
		}
	}

	if (! (rand() % frequency))
	{
		real_xlim = Dx(screen->r);
		real_ylim = Dy(screen->r);
		xlim = real_xlim * 1000;
		ylim = real_ylim * 1000;
		launch (xlim, ylim, g);
		resort = 1;
	}

	/* being sorted lets us avoid changing the GC's foreground color as often. */
	if (resort)
		sort_by_pixel (how_many);
}


void
screenhack (void)
{

	while (1)
	{
		pyro ();

		flushimage(display, 1);
		frames++;
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);
}

void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "pyro") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);
	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(i<<2));
	}

	eresized(0);
	init_pyro();
	screenhack();
}
