/* ripples, Copyright (c) 1999 Ian McConnell <ian@emit.demon.co.uk>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */

/*
 * "Water" ripples that can cross and interfere with each other.
 *
 * I can't remember where I got this idea from, but it's been around for a
 * while in various demos. Some inspiration from
 *      water.txt by Tom Hammersley,tomh@globalnet.co.uk
 *
 */

/* Version history:
 * 13 Oct 1999: Initial hack
 * 30 Oct 1999: Speeded up graphics with dirty buffer. Returned to using
 *              putpixel for greater portability
 *              Added a variety of methods for splashing screen.
 * 31 Oct 1999: Added in lighting hack
 * 13 Nov 1999: Speed up tweaks
 *              Adjust "light" for different bits per colour (-water only)
 *
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 09/03
 */

/* plan9-related stuff */
#include "plan9.h"

typedef enum {
	ripple_drop, ripple_blob, ripple_box, ripple_stir} ripple_mode;

static int ctab[256];
static int ncolors;
static int light;

static int width, height; /* ripple size */
static int bigwidth, bigheight; /* screen size */

static Bool transparent;
static short *bufferA, *bufferB, *temp;
static char *dirty_buffer;

#define TABLE 256
static double cos_tab[TABLE];

/* Distribution of drops: many little ones and a few big ones. */
static double drop_dist[] =
{
	0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.2, 0.6};

static void (*draw_transparent)(short *src);

/* How hard to hit the water */
#define SPLASH 512
#undef  MIN
#define MIN(x, y) ((x) < (y) ? (x) : (y))
#undef  MAX
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#undef  DIRTY
#define DIRTY 3 /* dirty >= 2, 1 = restore original pixel, 0 = leave alone */

/* From fortune(6) */
/* -- really weird C code to count the number of bits in a word */
#define BITCOUNT(x)	(((BX_(x)+(BX_(x)>>4)) & 0x0F0F0F0F) % 255)
#define BX_(x)		((x) - (((x)>>1)&0x77777777) \
                             - (((x)>>2)&0x33333333) \
                             - (((x)>>3)&0x11111111))


/*      -------------------------------------------             */


static int
map_color(int grey)
{
	/* Clip it */
	grey = ncolors * abs(grey) / (SPLASH/4);
	if (grey >= ncolors)
		grey = ncolors - 1;


	/* Display it */
	return ctab[grey];
}


static void
draw_ripple(short *src)
{
	int across, down;
	char *dirty = dirty_buffer;

	for (down = 0; down < height - 1; down++, src ++, dirty += 1)
		for (across = 0; across < width - 1; across++, src++, dirty++) {
			int v1, v2, v3, v4;
			v1 = (int)*src;
			v2 = (int)*(src + 1);
			v3 = (int)*(src + width);
			v4 = (int)*(src + width + 1);
			if ((v1 == 0 && v2 == 0 && v3 == 0 && v4 == 0)) {
				if (*dirty > 0)
					(*dirty)--;
			} else
				*dirty = DIRTY;

			if (*dirty > 0) {
				//XPutPixel(buffer_map,(across<<1),  (down<<1),  map_color(dx + v1));
				buf[down*width + across] = map_color(v1);

				//XPutPixel(buffer_map,(across<<1)+1,(down<<1),  map_color(dx + ((v1 + v2) >> 1)));
				buf[down*width + across+1] = map_color((v1 + v2) >> 1);

				//XPutPixel(buffer_map,(across<<1),  (down<<1)+1,map_color(dx + ((v1 + v3) >> 1)));
				buf[(down+1)*width + across] = map_color((v1 + v3) >> 1);

				//XPutPixel(buffer_map,(across<<1)+1,(down<<1)+1,map_color(dx + ((v1 + v4) >> 1)));
				buf[(down+1)*width + across+1] = map_color((v1 + v4) >> 1);
			}
		}

	loadimage(img, Rpt(Pt(0, 0), Pt(Dx(screen->r), Dy(screen->r))), buf, width*height);
	draw(screen, screen->r, img, nil, ZP);

}



/*      -------------------------------------------             */



static void
DisplayImage(void)
{

	//XSync(display,False);
	flushimage(display, 1);
	if(ecanmouse()) {
		m = emouse();
		if(m.buttons&4) {
			if(emenuhit(3, &m, &menu) == 0)
				threadexitsall(0);
		}
	}
}


/* Interpolate the ripple colours between the background colour and
   foreground colour */
static void
init_linear_colors(void)
{
	int i;
	for (i = 0; i < 256; i++) {
		ctab[i] = rgb2cmap(i, 0, 0);
	}
}




/*      -------------------------------------------             */


static void
init_cos_tab(void)
{
	int i;
	for (i = 0; i < TABLE; i++)
		cos_tab[i] = cos(i * M_PI/2 / TABLE);
}


/* Shape of drop to add */
static double
sinc(double x)
{
	/* cosine hump */
	int i;
	i = (int)(x * TABLE + 0.5);
	if (i >= TABLE) i = (TABLE-1) - (i-(TABLE-1));
	return cos_tab[i];
}


static void
add_circle_drop(int x, int y, int radius, int dheight)
{
	int i, r2, cx, cy;
	short *buf = (random()&1) ? bufferA : bufferB;

	i = y * width + x;
	r2 = radius * radius;

	for (cy = -radius; cy <= radius; cy++)
		for (cx = -radius; cx <= radius; cx++) {
			int r = cx*cx + cy*cy;
			if (r <= r2) {
				buf[i + cx + cy*width] =
					(short)(dheight * sinc(sqrt(r)/radius));
			}
		}
}


static void
add_drop(ripple_mode mode, int drop)
{
	int newx, newy, dheight;
	int radius = MIN(width, height) / 50;
	/* Don't put drops too near the edge of the screen or they get stuck */
	int border = 8;

	switch (mode) {
	default:
	case ripple_drop:
		{
			int x;

			dheight = 1 + (random() % drop);
			newx = border + (random() % (width - 2*border));
			newy = border + (random() % (height - 2*border));
			x = newy * width + newx;
			bufferA[x + 1] = bufferA[x] = bufferA[x + width] = bufferA[x + width + 1] =
				bufferB[x + 1] = bufferB[x] = bufferB[x + width] = bufferB[x + width + 1] =
				dheight;
		}
		break;
	case ripple_blob:
		{
			double power;

			power = drop_dist[random() % (sizeof(drop_dist)/sizeof(drop_dist[0]))]; /* clumsy */
			dheight = (int)(drop * (power + 0.01));
			newx = radius + border + (random() % (int)(width - 2*border - 2*radius*power));
			newy = radius + border + (random() % (int)(height - 2*border - 2*radius*power));
			add_circle_drop(newx, newy, radius, dheight);
		}
		break;
		/* Adding too many boxes too quickly (-box 1) doesn't give the waves time
     to disperse and the waves build up (and overflow) */
	case ripple_box:
		{
			int x;
			int cx, cy;
			short *buf = (random()&1) ? bufferA : bufferB;

			radius = (1 + (random() % 5)) * (1 + (random() % 5));
			dheight = drop / 128;
			if (random() & 1) dheight = -dheight;
			newx = radius + border + (random() % (width - 2*border - 2*radius));
			newy = radius + border + (random() % (height - 2*border - 2*radius));
			x = newy * width + newx;
			for (cy = -radius; cy <= radius; cy++)
				for (cx = -radius; cx <= radius; cx++)
					buf[x + cx + cy*width] = (short)(dheight);
		}
		break;
	case ripple_stir:
		{
			static double ang = 0;
			border += radius;
			newx = border + (int)((width-2*border) * (1+cos(3*ang)) / 2);
			newy = border + (int)((height-2*border) * (1+sin(2*ang)) / 2);
			add_circle_drop(newx, newy, radius, drop / 10);
			ang += 0.02;
			if (ang > 12*M_PI) ang = 0;
		}
		break;
	}
}


static void
init_ripples(int ndrops, int splash)
{
	int i;

	bufferA = (short *)calloc(width * height, sizeof(*bufferA));
	bufferB = (short *)calloc(width * height, sizeof(*bufferB));
	temp = (short *)calloc(width * height, sizeof(*temp));

	dirty_buffer = (char *)calloc(width * height, sizeof(*dirty_buffer));

	for (i = 0; i < ndrops; i++)
		add_drop(ripple_blob, splash);

	draw(screen, screen->r, display->black, nil, ZP);

	DisplayImage();
}


/*
 Explanation from hq_water.zip (Arturo R Montesinos (ARM) arami@fi.upm.es)

 Differential equation is:  u  = a ( u  + u  )
                             tt       xx   yy

 Where a = tension * gravity / surface_density.

 Approximating second derivatives by central differences:

  [ u(t+1)-2u(t)+u(t-1) ] / dt = a [ u(x+1)+u(x-1)+u(y+1)+u(y-1)-4u ] / h

 where dt = time step squared, h = dx*dy = mesh resolution squared.

 From where u(t+1) may be calculated as:

            dt  |   1   |                   dt
 u(t+1) = a --  | 1 0 1 |u - u(t-1) + (2-4a --)u
            h   |   1   |                    h

 When a*dt/h = 1/2 last term vanishes, giving:

                 1 |   1   |
        u(t+1) = - | 1 0 1 |u - u(t-1)
                 2 |   1   |

 (note that u(t-1,x,y) is only used to calculate u(t+1,x,y) so
  we can use the same array for both t-1 and t+1, needing only
  two arrays, U[0] and U[1])

 Dampening is simulated by subtracting 1/2^n of result.
 n=4 gives best-looking result
 n<4 (eg 2 or 3) thicker consistency, waves die out immediately
 n>4 (eg 8 or 12) more fluid, waves die out slowly
 */

static void
ripple(int fluidity)
{
	int across, down, pixel;
	static int toggle;
	static int count;
	short *src, *dest;

	if (toggle == 0) {
		src = bufferA;
		dest = bufferB;
		toggle = 1;
	} else {
		src = bufferB;
		dest = bufferA;
		toggle = 0;
	}

	switch (count) {
	case 0:
	case 1:
		pixel = 1 * width + 1;
		for (down = 1; down < height - 1; down++, pixel += 2 * 1)
			for (across = 1; across < width - 1; across++, pixel++) {
				temp[pixel] =
					(((src[pixel - 1] + src[pixel + 1] +
					src[pixel - width] + src[pixel + width]) / 2)) - dest[pixel];
			}

		/* Smooth the output */
		pixel = 1 * width + 1;
		for (down = 1; down < height - 1; down++, pixel += 2 * 1)
			for (across = 1; across < width - 1; across++, pixel++) {
				if (temp[pixel] != 0) { /* Close enough for government work */
					int damp =
					(temp[pixel - 1] + temp[pixel + 1] +
						temp[pixel - width] + temp[pixel + width] +
						temp[pixel - width - 1] + temp[pixel - width + 1] +
						temp[pixel + width - 1] + temp[pixel + width + 1] +
						temp[pixel]) / 9;
					dest[pixel] = damp - (damp >> fluidity);
				} else
					dest[pixel] = 0;
			}
		break;
	case 2:
	case 3:
		pixel = 1 * width + 1;
		for (down = 1; down < height - 1; down++, pixel += 2 * 1)
			for (across = 1; across < width - 1; across++, pixel++) {
				int damp =
				(((src[pixel - 1] + src[pixel + 1] +
					src[pixel - width] + src[pixel + width]) / 2)) - dest[pixel];
				dest[pixel] = damp - (damp >> fluidity);
			}
		break;
	}
	if (++count > 3) count = 0;

	draw_ripple(dest);
}


/*      -------------------------------------------             */


char *progclass = "Ripples";


void screenhack(void)
{
	int iterations = 0;
	int delay = 0; //50000 
	int rate = 5;
	int box = 0;
	int stir = 0;
	int fluidity = 6;
	transparent = 0;
	light = 0;

	width = Dx(screen->r);
	height = Dy(screen->r);

	if (fluidity <= 1) fluidity = 1;
	if (fluidity > 16) fluidity = 16; /* 16 = sizeof(short) */
	if (light < 0) light = 0;

	init_cos_tab();

	ncolors = 256;

	init_linear_colors();

	init_ripples(0, -SPLASH); /* Start off without any drops */

	/* Main drawing loop */
	while (1) {
		if (rate > 0 && (iterations % rate) == 0)
			add_drop(ripple_blob, -SPLASH);
		if (stir)
			add_drop(ripple_stir, -SPLASH);
		if (box > 0 && (random() % box) == 0)
			add_drop(ripple_box, -SPLASH);

		ripple(fluidity);
		DisplayImage();

		if (delay)
			sleep(delay);

		iterations++;
		frames++;
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	fprint(2, "resize not supported...\n");
	threadexitsall("resize not supported");

}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "ripples") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);

	buf = malloc(Dx(screen->r)*Dy(screen->r));
	if(buf == nil)
		sysfatal("cannot malloc buffer: %r\n");

	img = allocimage(display, Rpt(Pt(0, 0), Pt(Dx(screen->r), Dy(screen->r))), CMAP8, 1, DCyan);
	if(img == nil)
		sysfatal("cannot allocate image: %r\n");

	screenhack();
}
