/* sierpinski --- Sierpinski's triangle fractal */


/*-
 * Copyright (c) 1996 by Desmond Daignault
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Dots initially appear where they "should not".  Later they get
 * "focused".  This is correct behavior.
 *
 * Revision History:
 * 01-Nov-2000: Allocation checks
 * 18-Sep-1997: 3D version Antti Kuntsi <kuntsi@iki.fi>.
 * 20-May-1997: Changed the name tri to sierpinski for more compatiblity
 * 10-May-1997: Jamie Zawinski <jwz@jwz.org> compatible with xscreensaver
 * 05-Sep-1996: Desmond Daignault Datatimes Incorporated
 *            <tekdd@dtol.datatimes.com> .
 */

/*
  * ported to Plan 9 by andrey@lanl.gov, 08/02
  */

/* plan9-related stuff */
#include "plan9.h"
#define CYCLES 100
#define COUNT 2000

#define BRIGHT_COLORS


#define MAXCORNERS 4

typedef struct {
	int         width, height;
	int         time;
	int         px, py;
	int         total_npoints;
	int         corners;
	int         npoints[MAXCORNERS];
	unsigned long colors[MAXCORNERS];
	XPoint     *pointBuffer[MAXCORNERS];
	XPoint      vertex[MAXCORNERS];
} sierpinskistruct;

static sierpinskistruct *tris = (sierpinskistruct *) NULL;

static void
startover(void)
{
	int         j;
	sierpinskistruct *sp = &tris[0];

	if (sp->corners == 3) {
		sp->colors[0] = (NRAND(255));
		sp->colors[1] = (sp->colors[0] + 255 / 7 +
			NRAND(2 * 255 / 7 + 1)) % 255;
		sp->colors[2] = (sp->colors[0] + 4 * 255 / 7 +
			NRAND(2 * 255 / 7 + 1)) % 255;
	}
	else if (sp->corners == 4) {
		sp->colors[0] = (NRAND(255));
		sp->colors[1] = (sp->colors[0] + 255 / 7 +
			NRAND(255/ 7 + 1)) % 255;
		sp->colors[2] = (sp->colors[0] + 3 * 255 / 7 +
			NRAND(255/ 7 + 1)) % 255;
		sp->colors[3] = (sp->colors[0] + 5 * 255 / 7 +
			NRAND(255 / 7 + 1)) % 255;
	}
	else {
		fprint (2, "colors not set for %d corners\n", sp->corners);
	}
	for (j = 0; j < sp->corners; j++) {
		sp->vertex[j].x = NRAND(sp->width);
		sp->vertex[j].y = NRAND(sp->height);
	}
	sp->px = NRAND(sp->width);
	sp->py = NRAND(sp->height);
	sp->time = 0;

	draw(screen, screen->r, display->black, nil, ZP);
	flushimage(display, 1);
}

static void
free_sierpinski(sierpinskistruct *sp)
{
	int corner;

	for (corner = 0; corner < MAXCORNERS; corner++)
		if (sp->pointBuffer[corner] != NULL) {
			free((void *) sp->pointBuffer[corner]);
			sp->pointBuffer[corner] = (XPoint *) NULL;
		}
}

void
init_sierpinski(void)
{
	int         i;
	sierpinskistruct *sp;

	if (tris == NULL) {
		if ((tris = (sierpinskistruct *) calloc(1,
			sizeof (sierpinskistruct))) == NULL)
			return;
	}
	sp = &tris[0];

	sp->width = Dx(screen->r);
	sp->height = Dy(screen->r);

	sp->total_npoints = COUNT;
	sp->corners = (int) (LRAND() & 1) + 3;
	for (i = 0; i < sp->corners; i++) {
		if (!sp->pointBuffer[i])
			if ((sp->pointBuffer[i] = (XPoint *) malloc(sp->total_npoints *
				sizeof (XPoint))) == NULL) {
				free_sierpinski(sp);
				return;
			}
	}
	startover();
}

void
draw_sierpinski(void)
{
	XPoint     *xp[MAXCORNERS];
	int         i, i2, v;
	sierpinskistruct *sp;

	if (tris == NULL)
		return;
	sp = &tris[0];
	if (sp->pointBuffer[0] == NULL)
		return;

	for (i = 0; i < sp->corners; i++)
		xp[i] = sp->pointBuffer[i];
	for (i = 0; i < sp->total_npoints; i++) {
		v = NRAND(sp->corners);
		sp->px = (sp->px + sp->vertex[v].x) / 2;
		sp->py = (sp->py + sp->vertex[v].y) / 2;
		xp[v]->x = sp->px;
		xp[v]->y = sp->py;
		xp[v]++;
		sp->npoints[v]++;
	}
	for (i = 0; i < sp->corners; i++) {
		//		if (MI_NPIXELS(mi) > 2)
		//			XSetForeground(display, gc, MI_PIXEL(mi, sp->colors[i]));

		//		XDrawPoints(display, MI_WINDOW(mi), gc, sp->pointBuffer[i], sp->npoints[i], CoordModeOrigin);
		for(i2 = 0; i2 < sp->npoints[i]; i2++) {
			Rectangle r;
			Point p = screen->r.min;

			p = addpt(p, sp->pointBuffer[i][i2]);
			r = Rpt(p, addpt(p, Pt(1,1)));
			draw(screen, r, colors[sp->colors[i]], nil, ZP);

		}
		flushimage(display, 1);


		sp->npoints[i] = 0;
	}
	if (++sp->time >= CYCLES)
		startover();
}

void
release_sierpinski(void)
{
	if (tris != NULL) {
		free_sierpinski(&tris[0]);
		free((void *) tris);
		tris = (sierpinskistruct *) NULL;
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	if(tris != nil)
		release_sierpinski();
	init_sierpinski();
}

void 
threadmain(int argc, char **argv)
{
	Mouse m;
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));

	if(initdraw(nil, nil, "sierpinski") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(nrand(65535)));
	}

	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_sierpinski();
		frames++;
	}
}
