/* xscreensaver, Copyright (c) 1997, 1998 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */


/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 1/04
 */

#include "plan9.h"
#include "spline.h"
#include "smooth.h"

static int ncolors;
static int fg_index;
static int blob_p;
static int cycle_p;
static int resized = 0;

#define SCALE        1000	/* fixed-point math, for sub-pixel motion */


#define RAND(n) ((long) ((random() & 0x7fffffff) % ((long) (n))))
#define RANDSIGN() ((random() & 1) ? 1 : -1)

enum starfish_mode {
	pulse,
	zoom
};


struct starfish {
	enum starfish_mode mode;
	Bool blob_p;
	int skip;
	long x, y;		/* position of midpoint */
	double th;		/* angle of rotation */
	double rotv;		/* rotational velocity */
	double rota;		/* rotational acceleration */
	long elasticity;	/* how fast it deforms: radial velocity */
	double rot_max;
	long min_r, max_r;	/* radius range */
	int npoints;		/* control points */
	long *r;		/* radii */
	spline *spline;
	XPoint *prev;
	int n_prev;
};

static struct starfish *
make_starfish (int maxx, int maxy, int size)
{
	struct starfish *s = (struct starfish *) calloc(1, sizeof(*s));
	int i;

	s->blob_p = nrand(2);
	blob_p = 0;

	s->elasticity = RAND(5*SCALE) + RAND(5*SCALE) + RAND(5*SCALE);
	s->rotv = frand()*4.0 + frand()*4.0+ frand()*4.0;

	s->rotv /= 360;  /* convert degrees to ratio */

	if (s->blob_p)
	{
		s->elasticity *= 3;
		s->rotv *= 3;
	}

	s->rot_max = s->rotv * 2;
	s->rota = 0.0004 + frand()*0.0002;


	if (! (random() % 20))
		size *= frand()*0.35 + frand()*0.35 + 0.3;

	{
		static char skips[] = {
			2, 2, 2, 2,
			3, 3, 3,
			6, 6,
			12 				};
		s->skip = skips[rand() % sizeof(skips)];
	}

	if (! (rand() % (s->skip == 2 ? 3 : 12)))
		s->mode = zoom;
	else
		s->mode = pulse;

	maxx *= SCALE;
	maxy *= SCALE;
	size *= SCALE;

	s->max_r = size;
	s->min_r = 0;

	if (s->min_r < (5*SCALE)) s->min_r = (5*SCALE);

	s->x = maxx/2;
	s->y = maxy/2;

	s->th = frand()*(M_PI+M_PI) * RANDSIGN();

	{
		static char sizes[] = {
			3, 3, 3, 3, 3,
			4, 4, 4, 4,
			5, 5, 5, 5, 5, 5,
			8, 8, 8,
			10,
			35 				};
		int nsizes = 20;
		if (s->skip > 3)
			nsizes -= 4;
		s->npoints = s->skip * sizes[rand() % nsizes];
	}

	s->spline = make_spline (s->npoints);
	s->r = (long *) malloc (sizeof(*s->r) * s->npoints);

	for (i = 0; i < s->npoints; i++)
		s->r[i] = ((i % s->skip) == 0) ? 0 : size;

	return s;
}


static void
free_starfish (struct starfish *s)
{
	if (s->r) free (s->r);
	if (s->prev) free (s->prev);
	if (s->spline)
	{
		if (s->spline->control_x) free (s->spline->control_x);
		if (s->spline->control_y) free (s->spline->control_y);
		if (s->spline->points) free (s->spline->points);
		free (s->spline);
	}
	free (s);
}


static void 
throb_starfish (struct starfish *s)
{
	int i;
	double frac = ((M_PI+M_PI) / s->npoints);

	for (i = 0; i < s->npoints; i++)
	{
		long r ;
		long ra;
		double th;
		long x, y;
		long elasticity;

		r = s->r[i];
		ra = (r > 0 ? r : -r);
		th = (s->th > 0 ? s->th : -s->th);
		elasticity = s->elasticity;

		/* place control points evenly around perimiter, shifted by theta */
		x = s->x + ra * cos (i * frac + th);
		y = s->y + ra * sin (i * frac + th);

		s->spline->control_x[i] = x / SCALE;
		s->spline->control_y[i] = y / SCALE;
		if (s->mode == zoom && ((i % s->skip) == 0))
			continue;

		/* Slow down near the end points: move fastest in the middle. */
		{
			double ratio = (double)ra / (double)(s->max_r - s->min_r);
			if (ratio > 0.5) ratio = 1-ratio;	/* flip */
			ratio *= 2;				/* normalize */
			ratio = (ratio * 0.9) + 0.1;		/* fudge */
			elasticity = (long)(((float)elasticity)*ratio);
		}


		/* Increase/decrease radius by elasticity */
		ra += (r >= 0 ? elasticity : -elasticity);
		if ((i % s->skip) == 0)
			ra += (elasticity / 2);

		r = ra * (r >= 0 ? 1 : -1);

		/* If we've reached the end (too long or too short) reverse direction. */
		if ((ra > s->max_r && r >= 0) ||
			(ra < s->min_r && r < 0))
			r = -r;

		s->r[i] = r;
	}
}


static void
spin_starfish (struct starfish *s)
{
	double th = s->th;
	if (th < 0)
		th = -(th + s->rotv);
	else
		th += s->rotv;

	if (th > (M_PI+M_PI))
		th -= (M_PI+M_PI);
	else if (th < 0)
		th += (M_PI+M_PI);

	s->th = (s->th > 0 ? th : -th);

	s->rotv += s->rota;

	if (s->rotv > s->rot_max || 
		s->rotv < -s->rot_max)
	{
		s->rota = -s->rota;
	}
	/* If it stops, start it going in the other direction. */
	else if (s->rotv < 0)
	{
		if (rand() & 1)
		{
			/* keep going in the same direction */
			s->rotv = 0;
			if (s->rota < 0)
				s->rota = -s->rota;
		}
		else
		{
			/* reverse gears */
			s->rotv = -s->rotv;
			s->rota = -s->rota;
			s->th = -s->th;
		}
	}


	/* Alter direction of rotational acceleration randomly. */
	if (! (rand() % 120))
		s->rota = -s->rota;

	/* Change acceleration very occasionally. */
	if (! (rand() % 200))
	{
		if (rand() & 1)
			s->rota *= 1.2;
		else
			s->rota *= 0.8;
	}
}


static void
draw_starfish (struct starfish *s)
{
	compute_closed_spline (s->spline);
	if (s->prev)
	{
		int n;
		XPoint *points = (XPoint *)
			malloc (sizeof(XPoint) * (s->n_prev + s->spline->n_points + 2));
		int i = s->spline->n_points;
		int j = s->n_prev;
		memcpy (points, s->spline->points, (i * sizeof(*points)));
		memcpy (points+i, s->prev, (j * sizeof(*points)));

		if (s->blob_p)
			draw(screen, screen->r, display->black, nil, ZP);

		for(n = 0; n < i+j; n++)
			points[n] = addpt(points[n], screen->r.min);
		//XFillPolygon (dpy, drawable, gc, points, i+j, Complex, CoordModeOrigin);
		fillpoly(screen, points, i+j, 0, colors[fg_index], ZP);


		free (points);

		free (s->prev);
		s->prev = 0;
	}

	s->prev = (XPoint *) malloc (s->spline->n_points * sizeof(XPoint));
	memcpy (s->prev, s->spline->points, s->spline->n_points * sizeof(XPoint));
	s->n_prev = s->spline->n_points;

	/*
*	{ 
*		int i;
*		for (i = 0; i < s->npoints; i++) {
*			line (screen, addpt(screen->r.min, Pt(s->x/SCALE, s->y/SCALE)),
*			 	addpt(screen->r.min, 
*					Pt(s->spline->control_x[i], s->spline->control_y[i])), 
*				Endsquare, Endsquare, 0, display->white, ZP);
*		}
*	}
*/
}


static struct starfish *
make_window_starfish (void)
{
	int size;
	size = (Dx(screen->r) < Dy(screen->r) ? Dx(screen->r) : Dy(screen->r));
	if (blob_p) size /= 2;
	else size *= 1.3;
	return make_starfish (Dx(screen->r),Dy(screen->r), size);
}


static struct starfish *
init_starfish (void)
{
	static Bool first_time = True;


	ncolors = 256;

	if (blob_p)
		cycle_p = False;

	smooth(colors);

	fg_index = 0;

	if (first_time)
	{
		draw(screen, screen->r, colors[fg_index], nil, ZP);
		first_time = False;
	}


	return make_window_starfish ();
}



static void
run_starfish (struct starfish *s)
{
	throb_starfish (s);
	spin_starfish (s);
	draw_starfish (s);


	fg_index = (fg_index + 1) % ncolors;
}


void
screenhack (void)
{
	struct starfish *s = init_starfish ();
	int delay = 10;
	int delay2 = 5000;
	int cycle_delay = 100000;
	int duration = 30;
	long start = time (0);
	long now;
	int direction = (rand() % 1) ? 1 : -1;

	if (blob_p)
		delay *= 3;

	while (1)
	{
		if(resized) {
			s = init_starfish();
			resized = 0;
		}
		run_starfish (s);
		flushimage(display, 1);
		frames++;

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(0);
		}
		if (cycle_p && cycle_delay)
		{
			if (cycle_delay <= delay)
			{
				int i = 0;
				while (i < delay)
				{
					sleep(cycle_delay);
					i += cycle_delay;
				}
			}
			else
			{
				static long tick = 0;
				if (tick >= cycle_delay)
				{
					tick = 0;
				}
				if (delay)
					sleep(delay);
				tick += delay;
			}

			if (! (rand() % 1000))
				direction = -direction;
		}
		else if (delay)
			sleep (delay);

		if (duration > 0)
		{
			now = time (0);
			if (start + duration < now)
			{
				start = now;

				free_starfish (s);

				if (delay2 && !blob_p && cycle_p)
				{
					int i = 0;
					while (i < delay2)
					{
						if(ecanmouse()) {
							m = emouse();
							if(m.buttons&4)
								if(emenuhit(3, &m, &menu) == 0)
/* code folded from here */
	threadexitsall(0);
/* unfolding */
						}
						sleep(cycle_delay);
						i += cycle_delay;
					}
				}

				if (! (rand () % 10))
					s = init_starfish ();
				else
					s = make_window_starfish();
			}
		}
	}
}


void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	resized = 1;
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0)*getpid());


	if(initdraw(nil, nil, "starfish") < 0)
		sysfatal("initdraw failed: %r");


	einit(Emouse);
	eresized(0);

	screenhack();
}
