/* Copyright (c) 1994 M.Dobie <mrd@ecs.soton.ac.uk>
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * 13-May-97: jwz@jwz.org: turned into a standalone program.
 * 21-Apr-95: improved startup time for TrueColour displays
 *            (limited to 16bpp to save memory) S.Early <sde1000@cam.ac.uk>
 * 09-Jan-95: fixed colour maps (more colourful) and the image now spirals
 *            outwards from the centre with a fixed number of points drawn
 *            every iteration. Thanks to M.Dobie <mrd@ecs.soton.ac.uk>.
 * 1994:      written.   Copyright (c) 1994 M.Dobie <mrd@ecs.soton.ac.uk>
 *            based on original code by R.Taylor
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/03
 */

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"

/****************************************************************/

#define MASS            4	/* maximum mass of a knot */
#define MIN_RES         5	/* minimim resolution (>= MIN_RES) */
#define MAX_RES         1	/* maximum resolution (>0) */
#define TWO_PLANE_PCNT  30	/* probability for two plane mode (0-100) */
#define RESTART         2500	/* number of cycles before restart */
#define BATCH_DRAW      100	/* points to draw per iteration */

/* knot types */
typedef enum {
	NONE = 0,
	ORBIT = (1 << 0),
	WHEEL = (1 << 1),
	PICASSO = (1 << 2),
	RAY = (1 << 3),
	HOOK = (1 << 4),
	ALL = (1 << 5)
} KNOT_T;

/* a knot */
typedef struct Knot {
	int         x, y;	/* position */
	int         m;		/* mass */
	KNOT_T      t;		/* type in the first (or only) plane */
	KNOT_T      T;		/* type in second plane if there is one */
	int         M;		/* mass in second plane if there is one */
} KNOT     , *KNOT_P;

/* a colour specification */
typedef struct Colour {
	unsigned short r, g, b;
} COLOUR   , *COLOUR_P;

/* drawing direction */
typedef enum {
	DRAW_RIGHT, DRAW_DOWN, DRAW_LEFT, DRAW_UP
} DIR_T;

/****************************************************************/

/* data associated with a swirl window */
typedef struct swirl_data {
	/* window paramaters */
	int         width, height;	/* window size */
	int         depth;	/* depth */
	int         rdepth;	/* real depth (for XImage) */

	/* swirl drawing parameters */
	int         n_knots;	/* number of knots */
	KNOT_P      knots;	/* knot details */
	KNOT_T      knot_type;	/* general type of knots */
	int         resolution;	/* drawing resolution, 1..5 */
	int         max_resolution;	/* maximum resolution, MAX_RES */
	int         r;		/* pixel step */
	Bool        two_plane;	/* two plane mode? */
	Bool        first_plane;	/* doing first plane? */
	int         start_again;	/* when to restart */

	/* spiral drawing parameters */
	int         x, y;	/* current point */
	DIR_T       direction;	/* current direction */
	int         dir_todo, dir_done;		/* how many points in current direction? */
	int         batch_todo, batch_done;	/* how many points in this batch */
	Bool        started, drawing;	/* are we drawing? */

	/* image stuff */
	unsigned char *image;	/* image data */

	/* colours stuff */
	int         colours;	/* how many colours possible */
	int         dcolours;	/* how many colours for shading */
	Bool        monochrome;	/* monochrome? */
	int         current_map;	/* current colour map, 0..dcolours-1 */
	int         shift;	/* colourmap shift */
	int         dshift;	/* colourmap shift while drawing */
	Bool       off_screen;
} SWIRL    , *SWIRL_P;


/* an array of swirls for each screen */
static SWIRL_P swirls = NULL;

/* 
   random_no

   Return a random integer between 0 and n inclusive

   -      n is the maximum number

   Returns a random integer */

static int
random_no(unsigned int n)
{
	return nrand(n+1);
}

/****************************************************************/

/* 
   initialise_swirl

   Initialise all the swirl data

   -      swirl is the swirl data */

static void
initialise_swirl(SWIRL_P swirl)
{

	swirl->width = 0;	/* width and height of window */
	swirl->height = 0;
	swirl->depth = 1;
	swirl->rdepth = 1;
	swirl->resolution = MIN_RES + 1;	/* current resolution */
	swirl->max_resolution = MAX_RES;	/* maximum resolution */
	swirl->n_knots = 0;	/* number of knots */
	swirl->knots = nil;
	swirl->knot_type = ALL;	/* general type of knots */
	swirl->two_plane = False;	/* two plane mode? */
	swirl->first_plane = False;	/* doing first plane? */
	swirl->start_again = -1;	/* restart counter */

	/* drawing parameters */
	swirl->x = 0;
	swirl->y = 0;
	swirl->started = False;
	swirl->drawing = False;

	/* image stuff */
	swirl->image = NULL;	/* image data */

	/* colours stuff */
	swirl->colours = 0;	/* how many colours possible */
	swirl->dcolours = 0;	/* how many colours for shading */
	swirl->current_map = 0;	/* current colour map, 0..dcolours-1 */

}


/****************************************************************/

/* 
 * create_knots
 *
 * Initialise the array of knot
 *
 * swirl is the swirl data
 */
static void
create_knots(SWIRL_P swirl)
{
	int         k;
	Bool        orbit, wheel, picasso, ray, hook;
	KNOT_P      knot;

	/* create array for knots */
	if (swirl->knots)
		free((void *) swirl->knots);
	swirl->knots = (KNOT_P) calloc(swirl->n_knots, sizeof (KNOT));

	/* no knots yet */
	orbit = wheel = picasso = ray = hook = False;

	/* what types do we have? */
	if ((int) swirl->knot_type & (int) ALL) {
		orbit = wheel = ray = hook = True;
	} else {
		if ((int) swirl->knot_type & (int) ORBIT)
			orbit = True;
		if ((int) swirl->knot_type & (int) WHEEL)
			wheel = True;
		if ((int) swirl->knot_type & (int) PICASSO)
			picasso = True;
		if ((int) swirl->knot_type & (int) RAY)
			ray = True;
		if ((int) swirl->knot_type & (int) HOOK)
			hook = True;
	}

	/* initialise each knot */
	knot = swirl->knots;
	for (k = 0; k < swirl->n_knots; k++) {
		/* position */
		knot->x = random_no((unsigned int) swirl->width);
		knot->y = random_no((unsigned int) swirl->height);

		/* mass */
		knot->m = random_no(MASS) + 1;

		/* can be negative */
		if (random_no(100) > 50)
			knot->m *= -1;

		/* type */
		knot->t = NONE;
		while (knot->t == NONE) {
			/* choose a random one from the types available */
			switch (random_no(4)) {
			case 0:
				if (orbit)
					knot->t = ORBIT;
				break;
			case 1:
				if (wheel)
					knot->t = WHEEL;
				break;
			case 2:
				if (picasso)
					knot->t = PICASSO;
				break;
			case 3:
				if (ray)
					knot->t = RAY;
				break;
			case 4:
				if (hook)
					knot->t = HOOK;
				break;
			}
		}

		/* if two planes, do same for second plane */
		if (swirl->two_plane) {
			knot->T = NONE;
			while (knot->T == NONE || knot->T == knot->t) {
				/* choose a different type */
				switch (random_no(4)) {
				case 0:
					if (orbit)
						knot->T = ORBIT;
					break;
				case 1:
					if (wheel)
						knot->T = WHEEL;
					break;
				case 2:
					if (picasso)
						knot->T = PICASSO;
					break;
				case 3:
					if (ray)
						knot->T = RAY;
					break;
				case 4:
					if (hook)
						knot->T = HOOK;
					break;
				}
			}
		}
		/* next knot */
		knot++;
	}
}

/****************************************************************/

/* 
 * do_point
 *
 * Work out the pixel value at i, j. Ensure it does not clash with BlackPixel
 * or WhitePixel.
 *
 * -      swirl is the swirl data
 * -      i, j is the point to calculate
 *
 * Returns the value of the point
 */
static unsigned long
do_point(SWIRL_P swirl, int i, int j)
{
	int         tT, k, value, add;
	double      dx, dy, theta, dist;
	int         dcolours, qcolours;
	double      rads;
	KNOT_P      knot;

	/* how many colours? */
	dcolours = swirl->dcolours;
	qcolours = dcolours / 4;

	/* colour step round a circle */
	rads = (double) dcolours / (2.0 * M_PI);

	/* start at zero */
	value = 0;

	/* go through all the knots */
	knot = swirl->knots;
	for (k = 0; k < swirl->n_knots; k++) {
		dx = i - knot->x;
		dy = j - knot->y;

		/* in two_plane mode get the appropriate knot type */
		if (swirl->two_plane)
			tT = (int) ((swirl->first_plane) ? knot->t : knot->T);
		else
			tT = (int) knot->t;

		/* distance from knot */
		dist = sqrt(dx * dx + dy * dy);

		/* nothing to add at first */
		add = 0;

		/* work out the contribution (if close enough) */
		if (dist > 0.1)
			switch (tT) {
			case ORBIT:
				add = (int) (dcolours / (1.0 + 0.01 * abs(knot->m) * dist));
				break;
			case WHEEL:
				/* Avoid atan2: DOMAIN error message */
				if (dy == 0.0 && dx == 0.0)
					theta = 1.0;
				else
					theta = (atan2(dy, dx) + M_PI) / M_PI;
				if (theta < 1.0)
					add = (int) (dcolours * theta +
						sin(0.1 * knot->m * dist) *
						qcolours * exp(-0.01 * dist));
				else
					add = (int) (dcolours * (theta - 1.0) +
						sin(0.1 * knot->m * dist) *
						qcolours * exp(-0.01 * dist));
				break;
			case PICASSO:
				add = (int) (dcolours *
					fabs(cos(0.002 * knot->m * dist)));
				break;
			case RAY:
				/* Avoid atan2: DOMAIN error message */
				if (dy == 0.0 && dx == 0.0)
					add = 0;
				else
					add = (int) (dcolours * fabs(sin(2.0 * atan2(dy, dx))));

				break;
			case HOOK:
				/* Avoid atan2: DOMAIN error message */
				if (dy == 0.0 && dx == 0.0)
					add = (int) (0.05 * (abs(knot->m) - 1) * dist);
				else
					add = (int) (rads * atan2(dy, dx) +
						0.05 * (abs(knot->m) - 1) * dist);
				break;
			}
		/* for a +ve mass add on the contribution else take it off */
		if (knot->m > 0)
			value += add;
		else
			value -= add;

		/* next knot */
		knot++;
	}

	/* toggle plane */
	swirl->first_plane = (!swirl->first_plane);

	/* make sure we handle -ve values properly */
	if (value >= 0)
		value = (value % dcolours) + 2;
	else
		value = dcolours - (abs(value) % (dcolours - 1));


	/* definitely make sure it is in range */
	value = value % swirl->colours;


	/* return it */
	return ((unsigned long) value);
}

/****************************************************************/

/* 
 * draw_block
 *
 * Draw a square block of points with the same value.
 *
 * -      x, y is the top left corner
 * -      s is the length of each side
 * -      v is the value
 */
static void
draw_block(int x, int y, int s, unsigned long v)
{
	Rectangle r;

	r = rectaddpt(Rect(x, y, x+s, y+s), screen->r.min);
	draw(screen, r, colors[v], nil, ZP);
}

/****************************************************************/

/* 
 * draw_point  Draw the current point in a swirl pattern onto the XImage
 *
 * -    swirl is the swirl
 * -    win is the window to update
 */
static void
draw_point(SWIRL_P swirl)
{
	int         r;
	int         x, y;

	/* get current point coordinates and resolution */
	x = swirl->x;
	y = swirl->y;
	r = swirl->r;

	/* check we are within the window */
	if ((x < 0) || (x > swirl->width - r) || (y < 0) || (y > swirl->height - r))
		return;

	/* what style are we drawing? */
	if (swirl->two_plane) {
		int         r2;

		/* halve the block size */
		r2 = r / 2;

		/* interleave blocks at half r */
		draw_block(x, y, r2, do_point(swirl, x, y));
		draw_block(x + r2, y, r2, do_point(swirl, x + r2, y));
		draw_block(x + r2, y + r2, r2, do_point(swirl,
			x + r2, y + r2));
		draw_block(x, y + r2, r2, do_point(swirl, x, y + r2));
	} else
		draw_block(x, y, r, do_point(swirl, x, y));

	flushimage(display, 1);
}

/****************************************************************/

/* 
 * next_point  Move to the next point in the spiral pattern
 *  -    swirl is the swirl
 *  -    win is the window to update
 */
static void
next_point(SWIRL_P swirl)
{
	/* more to do in this direction? */
	if (swirl->dir_done < swirl->dir_todo) {
		/* move in the current direction */
		switch (swirl->direction) {
		case DRAW_RIGHT:
			swirl->x += swirl->r;
			break;
		case DRAW_DOWN:
			swirl->y += swirl->r;
			break;
		case DRAW_LEFT:
			swirl->x -= swirl->r;
			break;
		case DRAW_UP:
			swirl->y -= swirl->r;
			break;
		}

		/* done another point */
		swirl->dir_done++;
	} else {
		/* none drawn yet */
		swirl->dir_done = 0;

		/* change direction - check and record if off screen */
		switch (swirl->direction) {
		case DRAW_RIGHT:
			swirl->direction = DRAW_DOWN;
			if (swirl->x > swirl->width - swirl->r) {
				/* skip these points */
				swirl->dir_done = swirl->dir_todo;
				swirl->y += (swirl->dir_todo * swirl->r);

				/* check for finish */
				if (swirl->off_screen)
					swirl->drawing = False;
				swirl->off_screen = True;
			} else
				swirl->off_screen = False;
			break;
		case DRAW_DOWN:
			swirl->direction = DRAW_LEFT;
			swirl->dir_todo++;
			if (swirl->y > swirl->height - swirl->r) {
				/* skip these points */
				swirl->dir_done = swirl->dir_todo;
				swirl->x -= (swirl->dir_todo * swirl->r);

				/* check for finish */
				if (swirl->off_screen)
					swirl->drawing = False;
				swirl->off_screen = True;
			} else
				swirl->off_screen = False;
			break;
		case DRAW_LEFT:
			swirl->direction = DRAW_UP;
			if (swirl->x < 0) {
				/* skip these points */
				swirl->dir_done = swirl->dir_todo;
				swirl->y -= (swirl->dir_todo * swirl->r);

				/* check for finish */
				if (swirl->off_screen)
					swirl->drawing = False;
				swirl->off_screen = True;
			} else
				swirl->off_screen = False;
			break;
		case DRAW_UP:
			swirl->direction = DRAW_RIGHT;
			swirl->dir_todo++;
			if (swirl->y < 0) {
				/* skip these points */
				swirl->dir_done = swirl->dir_todo;
				swirl->x += (swirl->dir_todo * swirl->r);

				/* check for finish */
				if (swirl->off_screen)
					swirl->drawing = False;
				swirl->off_screen = True;
			} else
				swirl->off_screen = False;
			break;
		}
	}
}

/****************************************************************/

/* 
 * init_swirl
 *
 * Initialise things for swirling
 *
 * -      win is the window to draw in
 */
void
init_swirl(void)
{
	SWIRL_P     swirl;

	/* does the swirls array exist? */
	if (swirls == NULL) {

		/* allocate an array, one entry for each screen */
		swirls = (SWIRL_P) calloc(1, sizeof (SWIRL));

		initialise_swirl(&swirls[0]);
	}
	/* get a pointer to this swirl */
	swirl = &(swirls[0]);

	/* get window parameters */
	swirl->width = Dx(screen->r);
	swirl->height = Dy(screen->r);
	swirl->depth = 8;
	swirl->rdepth = 8;


	draw(screen, screen->r, display->black, nil, ZP);

	swirl->colours = 256;
	swirl->dcolours = swirl->colours;
	smooth(colors);

	/* resolution starts off chunky */
	swirl->resolution = MIN_RES + 1;

	/* calculate the pixel step for this resulution */
	swirl->r = (1 << (swirl->resolution - 1));

	/* how many knots? */
	swirl->n_knots = random_no(10) +
		10 + 1;

	/* what type of knots? */
	swirl->knot_type = ALL;	/* for now */

	/* use two_plane mode occaisionally */
	if (random_no(100) <= TWO_PLANE_PCNT) {
		swirl->two_plane = swirl->first_plane = True;
		swirl->max_resolution = 2;
	} else
		swirl->two_plane = False;

	/* fix the knot values */
	create_knots(swirl);

	/* we are off */
	swirl->started = True;
	swirl->drawing = False;
}

/****************************************************************/

/* 
 * draw_swirl
 *
 * Draw one iteration of swirling
 *
 * -      win is the window to draw in
 */
void
draw_swirl(void)
{
	SWIRL_P     swirl = &(swirls[0]);

	/* are we going? */
	if (swirl->started) {
		/* in the middle of drawing? */
		if (swirl->drawing) {

			/* draw a batch of points */
			swirl->batch_todo = BATCH_DRAW;
			while ((swirl->batch_todo > 0) && swirl->drawing) {
				/* draw a point */
				draw_point(swirl);

				/* move to the next point */
				next_point(swirl);

				/* done a point */
				swirl->batch_todo--;
			}
		} else {

			/* time for a higher resolution? */
			if (swirl->resolution > swirl->max_resolution) {
				/* move to higher resolution */
				swirl->resolution--;

				/* calculate the pixel step for this resulution */
				swirl->r = (1 << (swirl->resolution - 1));

				/* start drawing again */
				swirl->drawing = True;

				/* start in the middle of the screen */
				swirl->x = (swirl->width - swirl->r) / 2;
				swirl->y = (swirl->height - swirl->r) / 2;

				/* initialise spiral drawing parameters */
				swirl->direction = DRAW_RIGHT;
				swirl->dir_todo = 1;
				swirl->dir_done = 0;
			} else {
				/* all done, decide when to restart */
				if (swirl->start_again == -1) {
					/* start the counter */
					swirl->start_again = RESTART;
				} else if (swirl->start_again == 0) {
					/* reset the counter */
					swirl->start_again = -1;

					/* Pick a new colormap! */
					//XClearWindow (MI_DISPLAY(mi), MI_WINDOW(mi));
					draw(screen, screen->r, display->black, nil, ZP);

					/* start again */
					init_swirl();
				} else
					/* decrement the counter */
					swirl->start_again--;
			}
		}
	}
}

/****************************************************************/

void
release_swirl(void)
{
	/* does the swirls array exist? */
	if (swirls != NULL) {

		/* free them all */
		SWIRL_P     swirl = &(swirls[0]);

		if (swirl->knots)
			free((void *) swirl->knots);

		/* deallocate an array, one entry for each screen */
		free((void *) swirls);
		swirls = NULL;
	}
}

/****************************************************************/

void
refresh_swirl(void)
{
	SWIRL_P     swirl = &(swirls[0]);

	if (swirl->started) {
		if (swirl->drawing)
			swirl->resolution = swirl->resolution + 1;
		swirl->drawing = False;
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	if(swirls != nil)
		release_swirl();
	init_swirl();
}


void 
threadmain(int argc, char **argv)
{
	Mouse m;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "swirl") < 0)
		sysfatal("initdraw failed: %r");


	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_swirl();
		frames++;
	}
}
