/* xscreensaver, Copyright (c) 1992, 1995, 1996, 1997, 1998, 2004
 *  Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* wormhole:
 * Animation of moving through a wormhole. Based on my own code written
 * a few years ago.
 * author: Jon Rafkind <jon@rafkind.com>
 * date: 1/19/04
 */

/*
 * Ported to Plan 9 by mirtchovski@gmail.com, 06/04
 */

#include "plan9.h"
#include "smooth.h"

int col = 0;
int resized = 0;
QLock ql;		/* buffering synchronization */

int SCREEN_X, SCREEN_Y;
int z_speed;
int make_stars;

typedef struct STAR{
	int x, y;
	int calc_x, calc_y;
	int Z;
	int center_x, center_y;
} star;

typedef struct STARLINE{
	star begin, end;
} starline;

typedef struct COLORCHANGER{
	Image * shade;
	int min, max;
	int shade_use;
	int shade_max;
	int min_want;
} color_changer;

typedef struct WORMHOLE{
	int diameter;
	int diameter_change;
	int actualx, actualy;
	double virtualx, virtualy;
	double speed;
	int ang;
	int want_ang;
	int want_x, want_y;
	int max_Z;
	int addStar;
	int spiral;
	color_changer changer;
	starline ** stars;
	int num_stars; /* top of array we are using */
	Image * black;
	//	Pixmap work;
} wormhole;

static int gang( int x1, int y1, int x2, int y2 ){

	int tang;
	if ( y1 == y2 ) {

		if ( x1 < x2 )
			tang = 0;
		else
			tang = 180;

	} else
		tang = (int)(0.5+atan2( -(y2-y1), x2 - x1 ) * 180.0 / M_PI );

	while ( tang < 0 )
		tang += 360;
	return tang % 360;

}

void blend_palette( XColor * pal, int max, XColor * sc, XColor * ec ) {

	int q;

	int sc_r = sc->red;
	int sc_g = sc->green;
	int sc_b = sc->blue;

	int ec_r = ec->red;
	int ec_g = ec->green;
	int ec_b = ec->blue;

	for ( q = 0; q < max; q++ ) {
		float j = (float)( q ) / (float)( max );
		int f_r = (int)( 0.5 + (float)( sc_r ) + (float)( ec_r-sc_r ) * j );
		int f_g = (int)( 0.5 + (float)( sc_g ) + (float)( ec_g-sc_g ) * j );
		int f_b = (int)( 0.5 + (float)( sc_b ) + (float)( ec_b-sc_b ) * j );
		/* pal[q] = makecol( f_r, f_g, f_b ); */
		pal[q].red = f_r;
		pal[q].blue = f_b;
		pal[q].green = f_g;
	}

}


static void initWormhole( wormhole * worm ){

	int i;

	SCREEN_X = Dx(screen->r);
	SCREEN_Y = Dy(screen->r);

	worm->diameter = nrand( 10 ) + 15;
	worm->diameter_change = nrand( 10 ) + 15;
	worm->actualx = SCREEN_X / 2;
	worm->actualy = SCREEN_Y / 2;
	worm->virtualx = worm->actualx;
	worm->virtualy = worm->actualy;
	worm->speed = (float)SCREEN_X / 180.0;
	worm->spiral = 0;
	worm->addStar = make_stars;
	worm->want_x = nrand( SCREEN_X - 50 ) + 25;
	worm->want_y = nrand( SCREEN_Y - 50 ) + 25;
	worm->want_ang = gang( worm->actualx, worm->actualy, worm->want_x, worm->want_y );
	worm->ang = worm->want_ang;
	worm->max_Z = 600;
	worm->black = display->black;

	worm->num_stars = 64;
	worm->stars = (starline **)malloc( sizeof(starline *) * worm->num_stars );
	for ( i = 0; i < worm->num_stars; i++ )
		worm->stars[i] = NULL;

}

static void destroyWormhole( wormhole * worm ){
	free( worm->stars );
}

static double Cos( int a ){
	return cos( a * 180.0 / M_PI );
}

static double Sine( int a ){
	return sin( a * 180.0 / M_PI );
}



static void calcStar( star * st ){
	if ( st->center_x == 0 || st->center_y == 0 ){
		st->Z = 0;
		return;
	}
	if ( st->Z <= 0 ){
		st->calc_x = (st->x << 10) / st->center_x;
		st->calc_y = (st->y << 10) / st->center_y;
	} else {
		st->calc_x = (st->x << 10 ) / st->Z + st->center_x;
		st->calc_y = (st->y << 10 ) / st->Z + st->center_y;
	}
}

static void initStar( star * st, int Z, int ang, wormhole * worm ){

	st->x = Cos( ang ) * worm->diameter;
	st->y = Sine( ang ) * worm->diameter;
	st->center_x = worm->actualx;
	st->center_y = worm->actualy;
	st->Z = Z;
	calcStar( st );

}

static void addStar( wormhole * worm ){

	starline * star_new;
	starline ** xstars;
	int old_stars;
	int q;
	int ang;
	star_new = (starline *)malloc( sizeof( starline ) );
	ang = nrand( 360 );
	initStar( &star_new->begin, worm->max_Z, ang, worm );
	initStar( &star_new->end, worm->max_Z+nrand(6)+4, ang, worm );

	for ( q = 0; q < worm->num_stars; q++ ){
		if ( worm->stars[q] == NULL ){
			worm->stars[q] = star_new;
			return;
		}
	}

	old_stars = worm->num_stars;
	worm->num_stars = worm->num_stars << 1;
	xstars = (starline **)malloc( sizeof(starline *) * worm->num_stars );
	for ( q = 0; q < worm->num_stars; q++ )
		xstars[q] = NULL;

	for ( q = 0; q < old_stars; q++ )
		if ( worm->stars[q] != NULL ) xstars[q] = worm->stars[q];
	free( worm->stars );
	worm->stars = xstars;

	worm->stars[ old_stars ] = star_new;

}

static int moveStar( starline * stl ){

	stl->begin.Z -= z_speed;
	stl->end.Z -= z_speed;

	calcStar( &stl->begin );
	calcStar( &stl->end );

	return ( stl->begin.Z <= 0 || stl->end.Z <= 0 );

}

static int dist( int x1, int y1, int x2, int y2 ){
	int xs, ys;
	xs = x1-x2;
	ys = y1-y2;
	return (int)sqrt( xs*xs + ys*ys );
}

static void moveWormhole( wormhole * worm ){

	int q;
	double dx, dy;
	/* int x1, y1, x2, y2; */
	int min_dist = 100;
	int find = 0;
	dx = Cos( worm->ang ) * worm->speed;
	dy = Sine( worm->ang ) * worm->speed;

	worm->virtualx += dx;
	worm->virtualy += dy;
	worm->actualx = (int)worm->virtualx;
	worm->actualy = (int)worm->virtualy;

	if ( worm->spiral ){

		if ( worm->spiral % 5 == 0 )
			worm->ang = (worm->ang + 1 ) % 360;
		worm->spiral--;
		if ( worm->spiral <= 0 ) find = 1;

	} else {

		if ( dist( worm->actualx, worm->actualy, worm->want_x, worm->want_y ) < 20 )
			find = 1;

		if ( nrand( 20 ) == nrand( 20 ) )
			find = 1;

		if ( worm->actualx < min_dist ){
			worm->actualx = min_dist;
			worm->virtualx = worm->actualx;
			find = 1;
		}
		if ( worm->actualy < min_dist ){
			worm->actualy = min_dist;
			worm->virtualy = worm->actualy;
			find = 1;
		}
		if ( worm->actualx > SCREEN_X - min_dist ){
			worm->actualx = SCREEN_X - min_dist;
			worm->virtualx = worm->actualx;
			find = 1;
		}
		if ( worm->actualy > SCREEN_Y - min_dist ){
			worm->actualy = SCREEN_Y - min_dist;
			worm->virtualy = worm->actualy;
			find = 1;
		}

		if ( nrand( 500 ) == nrand( 500 ) ) worm->spiral = nrand( 30 ) + 50;
	}

	if ( find ){
		worm->want_x = nrand( SCREEN_X - min_dist * 2 ) + min_dist;
		worm->want_y = nrand( SCREEN_Y - min_dist * 2 ) + min_dist;
		worm->ang = gang( worm->actualx, worm->actualy, worm->want_x, worm->want_y );
	}
	for ( q = 0; q < worm->num_stars; q++ ){
		if ( worm->stars[q] != NULL ){
			if ( moveStar( worm->stars[q] ) ){
				free( worm->stars[q] );
				worm->stars[q] = NULL;
			}
		}
	}


	if ( worm->diameter < worm->diameter_change )
		worm->diameter++;
	if ( worm->diameter > worm->diameter_change )
		worm->diameter--;
	if ( nrand( 30 ) == nrand( 30 ) )
		worm->diameter_change = nrand( 35 ) + 5;

	for ( q = 0; q < worm->addStar; q++ )
		addStar( worm );

}

static void drawWormhole( wormhole * worm ){

	int i;
	starline * current;
	Point p1, p2;


	draw(img, img->r, display->black, nil, ZP);
	for ( i = 0; i < worm->num_stars; i++ )
		if ( worm->stars[i] != NULL ){

			current = worm->stars[i];

			p1 = Pt(current->begin.calc_x, current->begin.calc_y);
			p2 = Pt(current->end.calc_x, current->end.calc_y);
			line(img, p1, p2, Endsquare, Endsquare, 0, colors[ col ], ZP);

		}
	draw(screen, screen->r, img, nil, img->r.min);

}



void screenhack (void) {

	wormhole worm;


	int delay = 10;
	make_stars = 20;
	z_speed = 10;

	initWormhole( &worm );

	while (1) {

		if(resized) {
			resized = 0;
			destroyWormhole( &worm );
			initWormhole( &worm );
		}

		moveWormhole( &worm );
		drawWormhole( &worm );

		flushimage(display, 1);
		frames++;

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(nil);
		}
		if (delay) sleep (delay);
		col = ++col%256;

	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	resized = 1;

	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

	smooth(colors);
	col = 0;

}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "wormhole") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);

	eresized(0);
	resized = 0;
	screenhack();

}
